/*
 *    X2  IRC Operator and Channel services
 *    by Rubin (rubin@afternet.org)
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 1, or (at your option)
 * any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <ctype.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <unistd.h>
#ifndef MSDOS
#include <strings.h>
#endif
#include <time.h>
#include <signal.h>

#include "config.h"
#include "chan.h"
#include "init.h"
#include "user.h"
#include "misc.h"
#include "sock.h"
#include "server.h"
#include "commands.h"
#include "send.h"
#include "match.h"
#include "route.h"
#include "parse.h"
#include "accounting.h"
#include "help.h"
#include "set.h"

int FILTER = 0;
char LISTTYPE = '\0';

unsigned long int TotalCommands = 0, TotalLines = 0, TotalUserCount = 0;
char *chan_to;

extern int REMOTE_AUTHENTICATED;


/* *INDENT-OFF* */
#define NONE     0
#define REQED 1
#define REGED    2
#define T 1
#define F 0

/* User levels: 
 OW = 300  Owner 
 CO = 290  Co Owner 
 MA = 200  Manager 
 OP = 100  Op 
 PE = 10   Peon 
*/
struct
{
        char  *name;
        int  (*function)(C_PARAM);

        int   ChanStatus;   /* 0/NONE = Not required, 1/REQED = Required but not registered
                               2/REGED = Must be registered. */
        int   ChanLvl;      /* Minimum channel access required to get access */
        int   ChanAuth;     /* Do they need to be authenticated? */
        int   HelpOverride; /* Can +h people do this without access? */
        int   IRCOp;        /* Is this an IRCOP only command? */
        int   PublicOK;     /* Can this command be done in public? */

} on_msg_commands[] =
/*    Command      function        userlvl   na    public   Chan  InChan  ircopOverride */
/***************************************************************************/
/*    Command      c_function  ChanStatus:  Lvl, Auth, +h.   IRCOp, PubOK  */
{
    { "JUMP",      c_jump,           NONE,   0,     0,  0,    T,     T },
    { "UPLINKS",   c_uplinks,        NONE,   0,     0,  0,    T,     T }, 
    { "CLUSTER",   c_cluster,        NONE,   0,     0,  0,    F,     T }, 
    { "NEWS",      c_news,           NONE,   0,     0,  0,    F,     T },
    { "REHASH",    c_rehash,         NONE,   0,     0,  0,    T,     T },
    { "SEARCHCHAN",c_searchchan,     NONE,   0,     0,  0,    F,     T },
    { "TOPIC",     c_topic,         REGED,  OP,     T,  F,    F,     T },
    { "MODE",      c_mode,          REGED,  OP,     T,  F,    F,     T },
    { "UPTIME",    c_uptime,         NONE,   0,     0,  0,    F,     T },
    { "HELP",      c_help,           NONE,   0,     0,  0,    F,     T },
    { "IRCOPS",    c_ircops,         NONE,   0,     0,  0,    F,     T },
    { "SUSPEND",   c_suspend,       REGED,   0,     0,  0,    T,     T },
    { "UNSUSPEND", c_unsuspend,     REGED,   0,     0,  0,    T,     T },
    { "DISCONNECT",c_disconnect,     NONE,   0,     0,  0,    T,     F },
    { "NETSTAT",   c_netstat,        NONE,   0,     0,  0,    F,     T },
    { "NETSTATS",  c_netstat,        NONE,   0,     0,  0,    F,     T },
/*  not safe to show this anymore :(
    { "SERVERS",   c_servers,        NONE,   0,     0,  0,    F,     T },
    { "LINKS",     c_servers,        NONE,   0,     0,  0,    F,     T },
    { "MAP",       c_map,            NONE,   0,     0,  0,    F,     T },
*/
    { "RMAP",      c_rmap,           NONE,   0,     0,  0,    T,     T },
    { "RLIST",     c_rlist,          NONE,   0,     0,  0,    T,     T },
    { "ADDR",      c_addr,           NONE,   0,     0,  0,    T,     T },
    { "RLOAD",     c_rload,          NONE,   0,     0,  0,    T,     T },
    { "REROUTE",   c_reroute,        NONE,   0,     0,  0,    T,     T },
    { "SETTINGS",  c_settings,       NONE,   0,     0,  0,    T,     T },
    { "OPEN",      c_open,          REGED,  MA,     T,  T,    F,     T },
    { "INVITE",    c_invite,        REGED,  MA,     T,  F,    F,     T }, 
    { "INVITEME",  c_inviteme,      REGED,  OP,     T,  T,    F,     T },
/** New userlist commands go here. **/
    { "ADDUSER",   c_adduser,       REGED,  MA,     T,  T,    F,     T },
    { "ADDMASK",   c_addmask,       REGED,  PE,     T,  T,    F,     T },
    { "ADDHOST",   c_addmask,       REGED,  PE,     T,  T,    F,     T },
    { "DELUSER",   c_deluser,       REGED,  MA,     T,  T,    F,     T },
    { "DELMASK",   c_delmask,       REGED,  PE,     T,  T,    F,     T },
    { "DELHOST",   c_delmask,       REGED,  PE,     T,  T,    F,     T },

/** Userlist commands here. Prolly will keep these. **/
    { "USERS",     c_userlist,      REGED,   0,     0,  0,    F,     T },
    { "USERLIST",  c_userlist,      REGED,   0,     0,  0,    F,     T },
    { "LAMERLIST", c_llist,         REGED,  PE,     T,  T,    F,     T },
    { "LAMELIST",  c_llist,         REGED,  PE,     T,  T,    F,     T },
    { "LAMERS",    c_llist,         REGED,  PE,     T,  T,    F,     T },
    { "SHITLIST",  c_llist,         REGED,  PE,     T,  T,    F,     T },
    { "PLIST",     c_plist,         REGED,   0,     T,  T,    F,     T },
    { "OLIST",     c_olist,         REGED,   0,     T,  T,    F,     T },
    { "MLIST",     c_mlist,         REGED,   0,     0,  0,    F,     T },
    { "CLIST",     c_clist,         REGED,   0,     0,  0,    F,     T }, 
    { "WLIST",     c_wlist,         REGED,   0,     0,  0,    F,     T },
    { "BLIST",     c_llist,         REGED,  PE,     T,  T,    F,     T },
    { "LLIST",     c_llist,         REGED,  PE,     T,  T,    F,     T },
    { "CHANSTAT",  c_chanstat,      REQED,   0,     0,  0,    F,     T },
    { "CHANSTATS", c_chanstat,      REQED,   0,     0,  0,    F,     T },
    { "LAST",      c_last,          REGED,  CO,     T,  T,    F,     T },
    { "WHO",       c_who,           REGED,  PE,     T,  T,    F,     T },
/* Auth says 'please use LOGIN' */
    { "AUTH",      c_auth,           NONE,   0,     0,  0,    F,     T },
    { "PASS",      c_pass,          REGED,  PE,     F,  F,    F,     F },
    { "LOGIN",     c_login,         REGED,   0,     F,  F,    F,     F },
    { "MLOGIN",    c_mlogin,         NONE,   0,     F,  F,    F,     F },
    { "TOGOP",     c_togop,         REGED,  PE,     T,  F,    F,     T },
    { "UNREG",     c_unreg,         REGED,   0,     0,  0,    T,     F },
    { "EXPIRECHANS",c_expirechans,   NONE,   0,     0,  0,    T,     T },
/***************************************************************************/
/*    Command      c_function  ChanStatus:  Lvl, Auth, +h.   IRCOp, PubOK  */
    { "SPOOFSERV", c_spoofserv,      NONE,   0,     0,  0,    T,     T },
    { "RESETPASS", c_resetpass,     REGED,  MA,     T,  T,    F,     F }, 
    { "OP",        c_op,            REGED,  OP,     T,  F,    F,     T },
    { "VOICE",     c_voice,         REGED,  OP,     T,  F,    F,     T },
    { "V",         c_voice,         REGED,  OP,     T,  F,    F,     T },
    { "DEVOICE",   c_devoice,       REGED,  OP,     T,  F,    F,     T },
    { "OPCHAN",    c_opchan,        REQED,   0,     0,  0,    T,     T },
    { "DEOPCHAN",  c_deopchan,      REQED,   0,     0,  0,    T,     T },
    { "DOWN",      c_down,          REGED,   0,     0,  0,    F,     T },
    { "UP",        c_up,            REGED,  PE,     T,  F,    F,     T },
    { "DOME",      c_up,            REGED,  PE,     T,  F,    F,     T },
    { "KICK",      c_kick,          REGED,  OP,     T,  F,    F,     T },
    { "K",         c_kick,          REGED,  OP,     T,  F,    F,     T },
    { "DEOP",      c_deop,          REGED,  OP,     T,  F,    F,     T },
    { "SET",       c_set,           REGED,  OP,     T,  T,    F,     T },
    { "XSET",      c_xset,           NONE,   0,     0,  0,    T,     T },
    { "GLOBAL",    c_global,         NONE,   0,     0,  0,    T,     T },
    { "UNVISITED", c_unvisited,      NONE,   0,     0,  0,    T,     T },
    { "SETTIME",   c_settime,        NONE,   0,     0,  0,    T,     T },
    { "GLINE",     c_gline,          NONE,   0,     0,  0,    T,     T },
    { "SGLINE",    c_sgline,         NONE,   0,     0,  0,    T,     T },
    { "SPAMMER",   c_spammer,        NONE,   0,     0,  0,    T,     T },
    { "UNGLINE",   c_ungline,        NONE,   0,     0,  0,    T,     T },
    { "UNSGLINE",  c_unsgline,       NONE,   0,     0,  0,    T,     T },
    { "UNGLINEALL",c_unglineall,     NONE,   0,     0,  0,    T,     T },
    { "GLINES",    c_glines,         NONE,   0,     0,  0,    T,     T },
    { "SGLINES",   c_sglines,        NONE,   0,     0,  0,    T,     T },
    { "SETGLINES", c_setglines,      NONE,   0,     0,  0,    T,     T },
    { "LOADCHAN",  c_loadchan,       NONE,   0,     0,  0,    T,     T },
    { "REGCHAN",   c_regchan,        NONE,   0,     0,  T,    T,     T },
    { "SCRIPTREG", c_scriptreg,      NONE,   0,     0,  T,    T,     T },
    { "SAVECHAN",  c_savechan,      REGED,   0,     0,  0,    T,     T },
    { "LEVELS",    c_levels,         NONE,   0,     0,  0,    0,     T },
    { "ADDLAMER",  c_addlamer,      REGED,  MA,     T,  T,    F,     T },
    { "ADDPEON",   c_addpeon,       REGED,  MA,     T,  T,    F,     T },
    { "ADDOP",     c_addop,         REGED,  MA,     T,  T,    F,     T },
    { "ADDMANAGER",c_addmanager,    REGED,  CO,     T,  T,    F,     T },
    { "ADDCOOWNER",c_addcoowner,    REGED,  OW,     T,  T,    F,     T },
    { "ADDCO",     c_addcoowner,    REGED,  OW,     T,  T,    F,     T },
    { "ADDOWNER",  c_addowner,      REGED,   0,     0,  0,    T,     T },
    { "CLVL",      c_clvl,          REGED,  MA,     T,  T,    F,     T },    
    { "ACCESS",    c_access,        REGED,   0,     0,  0,    0,     T },
    { "COMMANDS",  c_commands,      NONE,    0,     0,  0,    0,     T },
/***************************************************************************/
/*    Command      c_function  ChanStatus:  Lvl, Auth, +h.   IRCOp, PubOK  */
    { "SEEN",      c_seen,          REGED,   0,     0,  0,    0,     T },
    { "DELLAMER",  c_dellamer,      REGED,  MA,     T,  T,    F,     T },
    { "SETHANDLE", c_sethandle,     REGED,  MA,     T,  T,    F,     T },
    { "SETINFO",   c_setinfo,       REGED,  PE,     T,  F,    F,     T },
    { "QUIT",      c_quit,           NONE,   0,     0,  0,    T,     F },
    { "DIE",       c_quit,           NONE,   0,     0,  0,    T,     F },
    { "D",         c_roll,           NONE,   0,     0,  0,    F,     T },
    { "DICE",      c_roll,           NONE,   0,     0,  0,    F,     T },
    { "ROLL",      c_roll,           NONE,   0,     0,  0,    F,     T },
    { "8BALL",     c_8ball,          NONE,   0,     0,  0,    F,     T },
    { "8",         c_8ball,          NONE,   0,     0,  0,    F,     T },
    { "CALC",      c_calc,           NONE,   0,     0,  0,    F,     T },
    { "WHOIS",     c_whois,          NONE,   0,     0,  0,    T,     T },
    { "WHOAMI",    c_whoami,         NONE,   0,     0,  0,    F,     T },
    { "UBA",       c_uba,           REGED,  OP,     T,  T,    F,     T },
    { "UB",        c_unban,         REGED,  OP,     T,  T,    F,     T },
    { "UNBAN",     c_unban,         REGED,  OP,     T,  T,    F,     T },
    { "UBM",       c_ubm,           REGED,  OP,     T,  T,    F,     T },
    { "B",         c_ban,           REGED,  OP,     T,  T,    F,     T },
    { "BAN",       c_ban,           REGED,  OP,     T,  T,    F,     T },
    { "KB",        c_kickban,       REGED,  OP,     T,  T,    F,     T },
    { "KICKBAN",   c_kickban,       REGED,  OP,     T,  T,    F,     T },
    { "VERSION",   c_version,        NONE,   0,     0,  0,    F,     T },
    { "AUTHORS",   c_authors,        NONE,   0,     0,  0,    F,     T },
    { "MOTD",      c_motd,           NONE,   0,     0,  0,    F,     T },
    { "SUBSCAN",   c_subscan,        NONE,   0,     0,  0,    T,     T },
#ifdef debugon
    { "DUMP",      c_raw,            NONE,   0,     0,  0,    T,     T },  
    { "RAW",       c_raw,            NONE,   0,     0,  0,    T,     T },
#endif
    { "GOD",       c_god,            NONE,   0,     0,  0,    T,     T },
    { "BACKUP",    c_backup,         NONE,   0,     0,  0,    T,     T },
/* ... */
   { NULL,        null(int(*)()),       0,   0,     0,  0,    0,     0 }
/***************************************************************************/
/*    Command      c_function  ChanStatus:  Lvl, Auth, +h.   IRCOp, PubOK  */
};

/* *INDENT-ON* */

void on_msg(UserListType uptr, char* channel, char* channel_done_in, boolean public, char* command, char* rest)
{
  ChannelListType cptr;
  HandleListType hptr = NULL;
  ChannelListType said_inptr;
  int found = 0, HelpOverride = 0;
  char errmsg[MAXLEN] = "";
  char restbackup[MAXLEN];
  char* chanstr = NULL;
  int i;
  
  /* Fix the trailing spaces crash. */
  if(rest)
     for(;*rest==' ';rest++);

  Debug(DBGNOTICE, "on_msg(uptr:%s, channel:%s, channel_done_in:%s, public:%d, command:%s, rest:%s",
                         uptr->Nick, channel, channel_done_in, public, command, rest);

  cptr = GetChannel(channel);
  said_inptr = GetChannel(channel_done_in);

  if(!command)
  {
     Debug(DBGWARNING, "on_msg(): Null Command");
     return;
  }
  if(!uptr)
  {
     Debug(DBGWARNING, "on_msg(): Null uptr");
     return;
  }
  if(!rest)
     rest = "";

  for(i=0;(on_msg_commands[i].name != NULL); i++)
  {
      if(!strcasecmp(on_msg_commands[i].name, command))
      {
          found++;
          break;
      }
  }
  if(found < 1)
  {
     char *TmpPtr;
     int valid = 1;

     for (TmpPtr = command; *TmpPtr && *TmpPtr != ' '; TmpPtr++)
     {				/* This checks for '<WeirdKev> ...' */
        if(!isalnum(*TmpPtr))
		{
           valid = 0;
           break;
		}
	 }
     if (valid)
        send_to_user(uptr, "\002%s\002: Unknown command name.  Try \002HELP\002", command);
     return;
  }

/* on_msg_commands[i].name  .ChanStatus Level Auth Help .IRCOp .PublicOK */
  if(on_msg_commands[i].IRCOp == T)
  {
     if(!IsOper(uptr))
     {
        send_to_user(uptr, "\002%s\002: Access denied.  IRC Operator status required.", on_msg_commands[i].name);
        return;
     }
  }
  if(on_msg_commands[i].PublicOK == F)
  {
      if(public)
      {
         send_to_user(uptr, "\002%s\002: Command must be done in private message. (/msg)", on_msg_commands[i].name);
         return;
      }
  }
  hptr = FindBestUserMatch(cptr, uptr);
  switch(on_msg_commands[i].ChanStatus)
  {
      case NONE: /* cptr can be NULL, the command doesnt REQUIRE a channel pointer */
           break;
      case REQED: /* This command requires a channel pointer, but it doesnt have to be registered */
           if(!cptr)
           {
             send_to_user(uptr, "\002%s\002: Existing channel name required.",
                                 on_msg_commands[i].name);
             return;
           }
           break;
      case REGED: /* This command requires a registered channel. */
           if(!cptr)
           {
               send_to_user(uptr, "\002%s\002: Registered channel name required.", 
                                   on_msg_commands[i].name);
               return;
           }
           if(!(cptr->Regged))
           {
               send_to_user(uptr, "\002%s\002: Registered channel name required. %s is not registered.",
                                   on_msg_commands[i].name, cptr->Name);
               return;
           }
           if(on_msg_commands[i].ChanLvl > 0)
           {
                if(hptr == NULL)
                {
                    sprintf(errmsg, "\002%s\002: Access denied. Do you need to LOGIN?", on_msg_commands[i].name); 
                }
                else if(userlevelbychar(hptr->Level) < on_msg_commands[i].ChanLvl)
                {
                    sprintf(errmsg, "\002%s\002: Access denied.", on_msg_commands[i].name);
                }
                if(*errmsg)
                {
                     if(on_msg_commands[i].HelpOverride &&  IsGodOper(uptr))
                     {
                          strcpy(errmsg, "");
                          HelpOverride++; 
                     }
                }
            }
            break;
      default: Debug(DBGWARNING, "Unknown status in command table for %s", on_msg_commands[i].name);
      return;
  }
 /* after switch() */
  if(*errmsg)
  {
     send_to_user(uptr, "%s", errmsg);
     return;
  }
  if(on_msg_commands[i].IRCOp && on_msg_commands[i].HelpOverride)
  {
      /* This is somewhat of a hack to get the #channel paramiter
       * though to regchan and scriptreg. */
      chanstr = channel;
  }
  strcpy(restbackup, rest);
  TotalCommands++;
  if(on_msg_commands[i].function(C_PARAM_CALL))
  {
     LogCommand(cptr, uptr, hptr, on_msg_commands[i].name, restbackup, (on_msg_commands[i].ChanStatus == REGED)?T:F, HelpOverride);
  }
}

/* Begin command descriptions */

int c_ban(C_PARAM)
{
  char nuh[MAXLEN];
  UserListType UserNode;
  char *nick, *user, *host;
  char tmpbuf[MAXBANLEN];

  rest = (char *) strtok(rest, " ");
  if (!rest)
  {
    send_to_user(uptr, "\002Usage: BAN <user | hostmask>\002");
    return(FALSE);
  }
  if (!cptr)
  {
    send_to_user(uptr, "\002Warning: Error getting channel! Internal error.\002");
    return(FALSE);
  }
  if(cptr->settings[INCHAN] == FALSE)
  {
    send_to_user(uptr, "\002%s must be INCHAN to perform a BAN.\002", IrcNick);
  }
  if ((UserNode = FindNick(rest, UsersN)))
  {
    if (IsModeK(UserNode))
    {
      send_to_user(uptr, "\002%s is a Service.\002", rest);
      return(FALSE);
    }
    sprintf(nuh, "%s!%s@%s", UserNode->Nick, ActiveUsername(UserNode), ActiveHost(UserNode));

    if (!banok(cptr, uptr, nuh))
    {
      send_to_user(uptr, "\002Sorry, that user is protected.\002");
      return(FALSE);
    }
    ban_user(cptr, UserNode);
    send_to_user(uptr, "\x2%s banned on %s.\x2", nuh, cptr->Name);
    send_wallchops(cptr, "\002This ban (%s) brought to you by %s/%s.\002", nuh, uptr->Nick, cptr->Name);
    return(TRUE);
  }
  else
  {
    if (!strchr(rest, '@') || !strchr(rest, '!'))
    {
      send_to_user(uptr, "\002Usage: BAN <nick | hostmask>\002");
      return(FALSE);
    }
    if(strlen(rest) > MAXBANLEN)
    {
      send_to_user(uptr, "Ban too long");
      return(FALSE);
    }
    safecopy(tmpbuf, rest, MAXBANLEN-1);
    tmpbuf[MAXLEN - 1] = '\0';
    nick = strtok(tmpbuf, "!");
    user = strtok(NULL, "@");
    host = strtok(NULL, " ");
    if (!nick || !user || !host || !*nick || !*user || !*host)
    {
      send_to_user(uptr, "\002A ban must contain a value for the nick, user and host (* is valid).\002");
      return(FALSE);
    }
    if (!banok(cptr, uptr, rest))
    {
      send_to_user(uptr, "\002Sorry, that user is protected.\002");
      return(FALSE);
    }
    else
    {
      ban_mask(cptr, rest);
      send_to_user(uptr, "\x2%s banned on %s.\x2", rest, cptr->Name);
      send_wallchops(cptr, "\002This ban (%s) brought to you by %s/%s.\002", rest, uptr->Nick, cptr->Name);
      return(TRUE);
    }
  }
  return(FALSE);
}

int c_jump(C_PARAM)
{
  UpLinkList *UpLinkPtr;
  char *server, *portstr;
  int port = 0;

  server = (char *) strtok(rest, " ");
  portstr = (char *) strtok(NULL, " ");
  if(!portstr)
     portstr = "6667";

  if(!server)
  {
     send_to_user(uptr, "\002Usage: JUMP <server> [port]\002");
     return(FALSE);
  }
  for(UpLinkPtr = UpLinks;UpLinkPtr;UpLinkPtr = UpLinkPtr->Next)
  {
    if(!strcasecmp(server, UpLinkPtr->ServName))
    {
      if(portstr)
        port = atol(portstr);
      if(port < 0)
        port = UpLinks->Port;
      
      send_to_user(uptr, "Attempting to switch uplink to %s...", UpLinkPtr->ServName);
      send_wallops("Changing to server: %s", UpLinkPtr->ServName);
      send_quit("Changing servers.");
      /* Disconnecting... */
      send_squit(Uplink, "Changing to server: %s", UpLinkPtr->ServName);
      DiscTime = time(NULL);
      close(Sock);
      ReadSocket(-1, NULL);
      Debug(DBGINFO, "Changing uplink servers...");
      ServerConnected = FALSE;
      REMOTE_AUTHENTICATED = FALSE;
      NETBURSTING = TRUE;
      SentBurst = FALSE;
      Authenticated = FALSE;
      KillServerList();
      KillUserList();
      OperCount = 0;
      CurUpLink = UpLinkPtr;
      if(TryToConnect(UpLinkPtr))
      {
          Authenticate(CurUpLink);
          Authenticated = TRUE;
          SentBurst = TRUE;
      }
      else
      {
         Debug(DBGINFO, "Uplink server change failed!  Trying to reconnect elsewhere...");
         SockConnected = FALSE;
         return(TRUE);
      }
      return(TRUE);
    }
  }
  send_to_user(uptr, "Server %s not found in uplinks list.", server);
  return(FALSE);
}

int c_uplinks(C_PARAM)
{
   UpLinkList *UpLinkPtr;

   send_to_user(uptr, "=----- Valid servers to link to: -------=");
   for(UpLinkPtr = UpLinks;UpLinkPtr;UpLinkPtr = UpLinkPtr->Next)
   {
     send_to_user(uptr, "%40s[%16s] %4d", UpLinkPtr->ServName, UpLinkPtr->ServIp, UpLinkPtr->Port);
   }
   send_to_user(uptr, "=---------------- End ------------------=");
   return(TRUE);
}

int c_togop(C_PARAM)
{
  char *phrase;

  if (cptr == NULL)
  {
    send_to_user(uptr, "Channel %s is not registered.", cptr->Name);
    return(FALSE);
  }
  rest = strtok(rest, " ");
  if(IsChanPeon(hptr))
     phrase = "Voice";
  else
     phrase = "Op";
  if (!rest)
  {
    if (IsChanAOp(hptr))
    {
      send_to_user(uptr, "X2 will now no longer auto-%s you when you join %s.", phrase, cptr->Name);
      SetChanAOp(hptr, FALSE);
      WriteChanOpFile(cptr);
      return(TRUE);
    }
    else
    {
      send_to_user(uptr, "X2 will now auto-%s you when you join %s.", phrase, cptr->Name);
      SetChanAOp(hptr, TRUE);
      WriteChanOpFile(cptr);
      return(TRUE);
    }
  }
  else if(!strcasecmp(rest, "ON") || !strcasecmp(rest, "TRUE"))
  {
    if (IsChanAOp(hptr))
      send_to_user(uptr, "Your auto-%s status on %s is already on.", phrase, cptr->Name);
    else
      send_to_user(uptr, "X2 will now auto-%s you when you join %s.", phrase, cptr->Name);
    SetChanAOp(hptr, TRUE);
    WriteChanOpFile(cptr);
    return(TRUE);
  }
  else if(!strcasecmp(rest, "OFF") || !strcasecmp(rest, "FALSE"))
  {
    if (!IsChanAOp(hptr))
      send_to_user(uptr, "Your auto-%s status on %s is already off.", phrase, cptr);
    else
      send_to_user(uptr, "X2 will now no longer auto-%s you when you join %s.", phrase, cptr->Name);
    SetChanAOp(hptr, FALSE);
    WriteChanOpFile(cptr);
    return(TRUE);
  }
  send_to_user(uptr, "Syntax: Please use ON or OFF.");
  return(FALSE);
}

int c_kickban(C_PARAM)
{
  char nuh[MAXLEN];
  char nick[MAXLEN], reason[MAXLEN], *buff;
  UserListType target;

  buff = (char *) mysep(&rest, " ");
  if (buff)
    safecopy(nick, buff, MAXLEN);
  else
  {
    send_to_user(uptr, "\002Usage: BAN <nick | hostmask>\002");
    return(FALSE);
  }
  buff = (char *) mysep(&rest, "\0");
  if (buff)
    sprintf(reason, "\002<%s> %s\002", uptr->Nick, buff);
  else
    sprintf(reason, "\002<%s> Goodbye %s.\002", uptr->Nick, nick);
  if (cptr == NULL)
  {
    Debug(DBGWARNING, "c_kickban(): NULL channel");
    return(FALSE);
  }
  if((target = FindNick(nick, UsersN)))
  {
    if (IsModeK(target))
    {
      send_to_user(uptr, "\002Sorry, you cannot kick or deop channel services.\002");
      return(FALSE);
    }
    sprintf(nuh, "%s!%s@%s", target->Nick, ActiveUsername(target), ActiveHost(target));

    if ((!kickok(cptr, uptr, target)) || (!banok(cptr, uptr, nuh)))
    {
      send_to_user(uptr, "\x2Sorry, that user is protected.\x2");
      return(FALSE);
    }
    ban_user(cptr, target);
    send_kick(cptr, target, "%s", reason);
    send_to_user(uptr, "\x2%s kick-banned on %s.\x2", nuh, cptr->Name);
  }
  else
    send_to_user(uptr, "\002Unknown user: %s.\002", nick);
  return(TRUE);
}

int c_ubm(C_PARAM)
{
  char nuh[MAXLEN];
  char nuh2[MAXLEN];

  if(uptr == NULL)
  {
    Debug(DBGWARNING, "c_ubm(): Null uptr");
    return(FALSE);
  }
  sprintf(nuh, "%s!%s@%s", uptr->Nick, ActiveUsername(uptr), ActiveHost(uptr));
  sprintf(nuh2, "%s!%s@%d.%d.%d.%d", uptr->Nick, uptr->UserID,
	  uptr->IP / 65536 / 256 % 256, uptr->IP / 65536 % 256, uptr->IP / 256 % 256, uptr->IP % 256);

  if (channel_unban(cptr, nuh) || channel_unban(cptr, nuh2))
  {
      send_to_user(uptr, "\x2You have been unbanned from %s.\x2", cptr->Name);
      return(TRUE);
  }
  else
  {
      send_to_user(uptr, "\002No bans matching %s found.\002", nuh);
      return(FALSE);
  }
  return(FALSE);
}

int c_uba(C_PARAM)
{
  if(channel_massunban(cptr))
  {
    send_to_user(uptr, "\002Unbanned everyone on %s.\002", cptr->Name);
    send_wallchops(cptr, "\002%s unbanned everyone on %s.\002", uptr->Nick, cptr->Name);
    return(TRUE);
  }
  else
  {
    send_to_user(uptr, "\002No bans in %s.\002", cptr->Name);
    return(FALSE);
  }
}

int c_unban(C_PARAM)
{
  char nuh[MAXLEN];
  UserListType target;

  if (!(rest = (char *) strtok(rest, " ")))
  {
    send_to_user(uptr, "\002Usage: UB <nick | hostmask>\002");
    return(FALSE);
  }
  if((target = FindNick(rest, UsersN)))
  {
    sprintf(nuh, "%s!%s@%s", target->Nick, ActiveUsername(target), ActiveHost(target));
    if (channel_unban(cptr, nuh))
    {
      send_to_user(uptr, "\002%s has been unbanned from %s.\002", nuh, cptr->Name);
      send_wallchops(cptr, "\02This ban removal (%s) was brought to you by %s/%s.\002", nuh, uptr->Nick, cptr->Name);
    }
    else
    {
      send_to_user(uptr, "\002No matching bans.\002");
      return(FALSE);
    }
  }
  else
  {
    if (channel_unban(cptr, rest))
    {
      send_to_user(uptr, "\002%s unbanned on %s.\002", rest, cptr->Name);
      send_wallchops(cptr, "\002This ban removal (%s) was brought to you by %s/%s.\002", rest, uptr->Nick, cptr->Name);
    }
    else
    {
      send_to_user(uptr, "\002No matching bans.\002");
      return(FALSE);
    }
  }
  return(TRUE);
}

int c_news(C_PARAM)
{
    SendNetMessage(uptr);
    return(true);
}

/* (kk) Note: if you type .cluster blah@blah.net in a channel, X2 returns *!**@*.
 * The cluster code needs to check for proper hostmasks.  -kk.
 */
int c_cluster(C_PARAM)
{
   if(!rest || !*rest)
   {
      send_to_user(uptr, "\002Usage: cluster <nick | hostmask>\002");
      send_to_user(uptr, "This command is just a quick access for testing the proper responce of the X2 hostmask guessing functions.");
      return(FALSE);
   }
   if(strchr(rest, '!') || strchr(rest, '@'))
     send_to_user(uptr, "That string clusters to: %s", cluster(rest));
   else
   {
     UserListType target;

     if (!(target = FindNick(rest, UsersN)))
        send_to_user(uptr, "Cannot find %s on IRC.", rest);
     else
     {
        char str[MAXLEN];
        snprintf(str, MAXLEN, "%s!%s@%s", target->Nick, ActiveUsername(target), ActiveHost(target));
        send_to_user(uptr, "That user (%s) clusters to: %s", str, cluster(str));
     }
   }
   return(FALSE);
}

int c_addmaster(C_PARAM)
{
   send_to_user(uptr, "The MASTER level has been changed to MANAGER, to avoid confusion.  Please see the LEVELS command for a description of all the levels available.  Please use MANAGER in all places you previously used MASTER.  Thanks! -%s CService Staff.",NETWORK); 
   return(FALSE);
}

int c_test(C_PARAM)
{
  return(FALSE);
}

/* TODO: I don't want functions in here changing complex data types such as malloc'ed strings
 * directly. It should call a function SetTopic() or such from chan.c
 * So the todo is, write these things and modify this to call them instead.
 * make sure they return success etc so the user can be told if it worked.
 */
int c_topic(C_PARAM)
{
  if(cptr == NULL)
  {
     Debug(DBGWARNING, "c_topic(): NULL channel");
     return(FALSE);
  }
  if(rest && *rest)
  {
     if((!IsGodOper(uptr)) && userlevel(hptr) < MA)
     {
        send_to_user(uptr, "You don't have access to change the default topic in %s, sorry.", cptr->Name);
        return(FALSE);
     }
     if(!strcmp(rest, "*"))
     {
        send_to_user(uptr, "Default topic cleared in %s.", cptr->Name);
        strcpy(cptr->Topic, "*");
        return(TRUE);
     }
     else
     {
       free(cptr->Topic);
       if((cptr->Topic = (char *) malloc(strlen(rest) + 10)))
       { 
         safecopy(cptr->Topic, rest, MAXTOPICLEN);
         send_to_user(uptr, "Default topic in %s changed to: %s", cptr->Name, cptr->Topic);
	 WriteChanOpFile(cptr);
       }
       else
       {
           Debug(DBGSEVERE, "Out of memory allocating topic in c_topic()");
           return(FALSE);
       }
     }
  }
  if (cptr->settings[INCHAN] && (strcmp(cptr->Topic, "*")))
    send_topic(cptr, cptr->Topic);
  else
  {
    send_to_user(uptr, "\002There is no default topic for %s.\002", cptr->Name);
    return(FALSE);
  }
  return(TRUE);
}

int c_god(C_PARAM)
{
  if(uptr != NULL)
  {
    if (!strcasecmp(rest, "ON"))
    {
      uptr->Modes = (uptr->Modes | UMODEgod);
      send_to_user(uptr, "\002You now have authority to override channel security.  Use it wisely.\002");
    }
    else if (!strcasecmp(rest, "OFF"))
    {
      uptr->Modes = (uptr->Modes & ~UMODEgod);
      send_to_user(uptr, "\002You have given up authority to override channel security.\002");
      return(TRUE);
    }
    else
    {
      send_to_user(uptr, "\002Usage: GOD <ON | OFF>\002");
      return(FALSE);
    }
  }
  else
  {
    Debug(DBGWARNING, "c_god(): null user");
    return(FALSE);
  }
  return(FALSE);
}

int c_whoami(C_PARAM)
{
  char buff[MAXLEN];
  strcpy(buff, uptr->Nick);
  rest = buff;
  return(c_whois(C_PARAM_CALL));
}

int c_whois(C_PARAM)
{
  UserListType target;
  char modetmp[20] = "";
  AuthHandleListType ahptr;
  JoinedChansListType jptr;

  /* The ownerlist is really for what chans were in,
   * but it used to only keep for owners.. the name is misleading
   * its for anyone whos LAMER/OP/MANAGER/(CO)OWNER etc */

  if (!rest)
  {
    send_to_user(uptr, "\002Usage: whois <nick>\002");
    return(FALSE);
  }
  if((target = (FindNick(rest, UsersN))))
  {
    send_to_user(uptr, "\002=------%s------=\002", rest);
    send_to_user(uptr, "\002  NumNick : %s\002", target->Numnick);
    send_to_user(uptr, "\002  Nick    : %s\002", target->Nick);
    send_to_user(uptr, "\002  UserID  : %s\002", target->UserID);
    send_to_user(uptr, "\002  Address : %s\002", target->Address);
    send_to_user(uptr, "\002  Realname: %s\002", target->RealName);
    send_to_user(uptr, "\002  Account : %s\002", target->Account);
    send_to_user(uptr, "\002  fakehost: %s\002", target->Fakehost);
    send_to_user(uptr, "\002  sethost : %s\002", target->Sethost);
    send_to_user(uptr, "\002  IP Addr : %s\002", unsigned32ToDotQuad(target->IP));
    send_to_user(uptr, "\002  Server  : %s\002", target->Serv);
    send_to_user(uptr, "\002  Clones  : %d\002", target->Clones);
    if (target->Modes & UMODEd)
      strcat(modetmp, "d");
    if (target->Modes & UMODEi)
      strcat(modetmp, "i");
    if (target->Modes & UMODEo)
      strcat(modetmp, "o");
    if (target->Modes & UMODEw)
      strcat(modetmp, "w");
    if (target->Modes & UMODEk)
      strcat(modetmp, "k");
    if (target->Modes & UMODEg)
      strcat(modetmp, "g");
    if (target->Modes & UMODEh)
      strcat(modetmp, "h");
    if (target->Modes & UMODEx)
      strcat(modetmp, "x");
    if (target->Modes & UMODEB)
      strcat(modetmp, "B");
    if (target->Modes & UMODER)
      strcat(modetmp, "R");
    if (target->Modes & UMODEr)
      strcat(modetmp, "r");
    if (target->Modes & UMODEf)
      strcat(modetmp, "f");
    if (target->Modes & UMODEgod)
      strcat(modetmp, "(GOD)");    
    send_to_user(uptr, "\002  Modes  : +%s\002", modetmp);
    
    for(ahptr = target->AuthHandles;ahptr;ahptr = ahptr->Next)
    {
        send_to_user(uptr, "\002  %s\002 on %s.", ahptr->hptr->Handle, ahptr->cptr->Name);
    }
    for(jptr = target->Chans;jptr;jptr = jptr->Next)
    {
        send_to_user(uptr, "\002 In channel %s", jptr->cptr->Name);
    }

    send_to_user(uptr, "\002=----------------=\002");
    return(TRUE);
  }
  else
  {
    send_to_user(uptr, "\002No such user: %s.\002", rest);
    return(FALSE);
  }
}

int c_roll(C_PARAM)
{
  int my_rand;
  int size = 1;
  char buff[MAXLEN];

  rest = strtok(rest, " ");
  if (rest && *rest)
  {
    size = atoi(rest);
  }
  else
    size = 6;
  if (size < 2)
  {
    send_to_user(uptr, "\002Each die must have at least 2 sides.\002");
    return(FALSE);
  }
  my_rand = (rand() % size);
  my_rand++;
  sprintf(buff, "A \002%d\002 shows on the %d-sided die.", my_rand, size);

  if(PublicReplyAllowed(hptr) && said_inptr)
  {
      send_to_channel(said_inptr, "\002%s:\002 %s.", uptr->Nick, buff);
  }
  else
  {
      send_to_user(uptr, "%s", buff);
  }
  return(TRUE);
}

void eightball(char *outcome, int method)
{
   int answer = 0;

#define NUMOFCOLORS 17
   char ballcolors[40][17] = {"blue", "red", "green", "yellow",
        "white", "black", "grey", "brown",
        "yellow", "pink", "purple", "orange", "teal", "burgandy",
        "fuchsia","turquoise","magenta"};
#define NUMOFLOCATIONS 45
   char balllocations[50][50] = { "Locke's house", "Oregon", "California", "Indiana", "Canada",
        "Russia", "Japan", "Florida", "the Bahamas", "Hiroshima",
        "the Caribbean", "the Everglades", "your head", "your pants", "your school",
        "the Statue of Liberty", "Mt. Fugi", "your mother's house", "IRC", "OSU",
        "Locke's cat", "the closet", "the washroom", "a lake", "Spain"
        "a bathtub", "a toilet", "a sewer", "a horse", "Jupiter",
        "Uranus", "Pluto", "a dark place", "your undies", "your shirt",
        "your bra", "your hair", "bed", "the couch", "the wall",
        "your skin", "your brain", "Italy", "the Netherlands", "Germany"};
#define NUMOFPREPS 10
   char ballpreps[50][50] = { "Near", "In", "On", "Under", "Next to", 
        "above", "In place of", "Beside", "Far away from", "North of"};
#define NUMOFNUMS 10
   char ballnums[50][50] = { "Ten", "A hundred", "A few", "Approximately 7",
        "About 9", "-42", "4", "8", "3", "2"};
#define NUMOFMULTS 4
   char ballmults[4][50] = { " million", "", " thousand", ""};

   /* Method:
    * 0: normal
    * 1: color
    * 2: where is
    * 3: how many
    */
 
    if(method == 0)
    {
      answer = (rand() % 11); /* Make sure this is the # of entries */
      switch(answer)
      {
        case 0: strcpy(outcome, "Are you out of your MIND?");
                return;
        case 1: strcpy(outcome, "It won't happen, not a chance, definitely no.");
                return;
        case 2: strcpy(outcome, "Outlook seems bleak.");
                return;
        case 3: strcpy(outcome, "My sources say no.");
                return;
        case 4: strcpy(outcome, "You bet!");
                return;
        case 5: strcpy(outcome, "It is decidedly so.");
                return;
        case 6: strcpy(outcome, "It's hard to be sure.");
                return;
        case 7: strcpy(outcome, "Most definitely.");
                return;
        case 8: strcpy(outcome, "In your dreams...");
                return;
        case 9: strcpy(outcome, "If the prophets wish it...");
                return;
        case 10: strcpy(outcome, "Forecast hazy, try again later.");
                return;
        case 11: strcpy(outcome, "I don't know!");
                return;
        default: strcpy(outcome, "An invalid random number was generated.");
                return;
      }
    }
    else if (method == 1) /* A Color */
    {
      char tmp[MAXLEN];

      answer = (rand() % 5); /* Make sure this is the # of entries */
      switch(answer)
      {
        case 0: strcpy(tmp, "Very bright %s, I'd say.");
                break;
        case 1: strcpy(tmp, "Sort of a light %s color.");
                break;
        case 2: strcpy(tmp, "Dark and dreary %s.");
                break;
        case 3: strcpy(tmp, "Quite a pale shade of %s.");
                break;
        case 4: strcpy(tmp, "A gross kind of mucky %s.");
                break;
        default: strcpy(outcome, "An invalid random number was generated.");
                return;
      }
      sprintf(outcome, tmp, ballcolors[rand() % NUMOFCOLORS]);
      return;
    }
    else if (method == 2)  /* Location */
    {
       sprintf(outcome, "%s %s.", ballpreps[rand() % NUMOFPREPS], balllocations[rand() % NUMOFPREPS]);
    }
    else if (method == 3)  /* Number of ___ */
    {
       sprintf(outcome, "%s%s.", ballnums[rand() % NUMOFNUMS], ballmults[rand() % NUMOFMULTS]);
    }
    else
    {
      Debug(DBGWARNING, "Error in 8ball.");
    }
    return;
}

int c_8ball(C_PARAM)
{
  char *word1, *word2, *word3;
  static char eb[MAXLEN];
  
  if(!rest || !*rest)
  {
    send_to_user(uptr, "\002Usage:\002 8BALL [how many...?|where is...?|what color...?|yes or no question]");
    return(FALSE);
  }
  word1 = (char *) strtok(rest, " ,.\"_[]()");
  word2 = (char *) strtok(NULL, " ,.\"_[]()");
  word3 = (char *) strtok(NULL, " ,.\"_[]()");

  if(!word1)
	  word1 = "";

/*** COLOR *****/
  if((word2) && strcasecmp(word1, "what") == 0 && strcasecmp(word2, "color") == 0)
     eightball(eb, 1);
  else if((word3) && strcasecmp(word1, "what's") == 0 && strcasecmp(word2, "the") == 0 && strcasecmp(word3, "color") == 0)
     eightball(eb, 1);
  else if((word3) && strcasecmp(word1, "whats") == 0 && strcasecmp(word2, "the") == 0 && strcasecmp(word3, "color") == 0)
     eightball(eb, 1);
/*** LOCATION *****/
  else if(
           (
             word2 && 
             (
                (strcasecmp(word1, "where") == 0) && 
                (strcasecmp(word2, "is") == 0)
             )
           ) || 
             (
                   strcasecmp(word1, "where's") == 0
             )
         )
     eightball(eb, 2);
/*** NUMBER *****/
  else if((word2) && strcasecmp(word1, "how") == 0 && strcasecmp(word2, "many") == 0)
     eightball(eb, 3);
/*** GENERIC *****/
  else
     eightball(eb, 0);
  if(said_inptr && PublicReplyAllowed(hptr))
  {
     send_to_channel(said_inptr, "\002%s:\002 %s", uptr->Nick, eb);
  }
  else
  {
     send_to_user(uptr, "%s", eb);
  }
return(TRUE);
}

int c_calc(C_PARAM)
{
  char buff[MAXLEN];
  char newrest[MAXLEN];
  char *ptr;
  boolean lastNumber = false;
  boolean lastBracket = false;
  boolean failed;

  if(!rest || !*rest)
  {
      send_to_user(uptr, "\002Usage: CALC <expression>\002");
      send_to_user(uptr, "For example: CALC 5 + 3 / 6");
      return(TRUE);
  }

  /* This is a HACK to insert * before ( so multiplication takes place
   * in things such as 3(3) instead of ignoring the first 3. Submitted
   * by Osiris.  
   * I spent some time tlooking at calc.c and I didnt see an easy way
   * to do this there... and that file gives me a headache.. so.. -Rubin
   */
  ptr = newrest;
  while(*rest && ptr < newrest+MAXLEN-1) 
  {
      switch(*rest) 
      {
        case '1': case '2':
        case '3': case '4':
        case '5': case '6':
        case '7': case '8':
        case '9': case '0':
          lastNumber = true;
          if(lastBracket == true) 
          {
              *ptr = '*';
              ptr++;
          }
          *ptr = *rest;
          lastBracket = false;
          break;
        case ')': case ']': /* Support sin[12](3) also */
          lastNumber = true;
          lastBracket = true;
          *ptr = *rest;
          break;
        case '(': case '[':
          if (lastNumber == true) 
          {
            *ptr = '*';
            ptr++;
          }
          *ptr = *rest;
          lastNumber = false;
          lastBracket = false;
          break;
        default:
          if(isalpha(*rest))
          {
            if(lastNumber == true) 
            {
              *ptr = '*';
              ptr++;
            }
          }
          *ptr = *rest;
          lastNumber = false;
          lastBracket = false;
          break;
      }
      rest++;
      ptr++;
    }
    *ptr = '\0';
    do_math(buff, newrest);
    if(said_inptr && PublicReplyAllowed(hptr))
    {
       send_to_channel(said_inptr, "\002%s:\002 %s", uptr->Nick, buff);
    }
    else
    {
       send_to_user(uptr, "%s", buff);
    }
    return(TRUE);
}

int c_inviteme(C_PARAM)
{
  send_invite(cptr,uptr);
  send_to_user(uptr, "\002You are invited to %s.\002", cptr->Name);
  return(TRUE);
}

int c_open(C_PARAM)
{
  send_chanmode(cptr, "-ilkmnrtps *");
  send_to_user(uptr, "Clearing %s of all modes.", cptr->Name);
  if(cptr->Regged)
  {
     channel_massunban(cptr);
  }
  return(TRUE);
}

int c_invite(C_PARAM)
{
  char* targetnick;
  UserListType target;
  time_t Now;
  targetnick = strtok(rest, " ");

  if(!cptr->settings[INCHAN] || cptr->Suspended)
  {
      send_to_user(uptr, "I need to be in the channel to invite persons");
      return(FALSE);
  }
  if(!targetnick || !*targetnick)
  {
     send_to_user(uptr, "\002Usage: INVITE <nick>\002");
     return(FALSE);
  }
  target = FindNick(targetnick, UsersN);
  if(!target)
  {
     send_to_user(uptr, "Cant find %s on IRC", targetnick);
     return(FALSE);
  }

  Now = time(NULL);
  if(uptr->LastInvite < Now)
  {
    uptr->LastInvite = Now;
  }
  if (uptr->LastInvite - Now > MAXINVITE)
  {
    send_to_user(uptr, "Please do not abuse invite, you are using it too much");
    return(FALSE);
  }
  uptr->LastInvite += DELTAINVITE;

  send_invite(cptr, target);
  send_to_user(target, "%s has invited you to %s.", uptr->Nick, cptr->Name);
  send_to_user(uptr, "Inviting %s to %s.", targetnick, cptr->Name);
  return(TRUE);
}

int c_resetpass(C_PARAM)
{
  HandleListType target;
  char* handle;
  char* newpass;
  char* cryptpass;

  handle = (char *) strtok(rest, " ");
  newpass = (char *) strtok(NULL, " ");
  if(!newpass)
  {
      send_to_user(uptr, "\002Usage: RESETPASS <handle> <newpass>\002");
      return(FALSE);
  }
  if(!(target = FindHandle(cptr, handle)))
  {
    send_to_user(uptr, "Sorry, I cannot find %s in the %s userlist.", handle, cptr->Name);
    return(FALSE);
  }
  if(!PassOk(newpass))
  {
      send_to_user(uptr, "The password you specified is not acceptable.  A password must be less than %d characters and cannot have control characters.", PASSWORDLEN);
      return(FALSE);
  }
  if((!IsGodOper(uptr)) && (userlevel(hptr) <= userlevel(target)) )
  {
      send_to_user(uptr, "You cannot modify users at access levels greater than or equal to your own.");
      return(FALSE);
  } 
  cryptpass = crypt(newpass, "$1$92737837$");
  if(target->Pass != NULL)
    free(target->Pass);
  target->Pass = (char *) malloc(strlen(cryptpass) + 1);
  strcpy(target->Pass, cryptpass);
  send_to_user(uptr, "Password for %s changed.", target->Handle);
  WriteChanOpFile(cptr);
  return(TRUE);
}

/* With this command the user logs in as a handle in a channel. */
int c_login(C_PARAM)
{
   char* handle;
   char* password;
   AuthHandleListType ahptr;

   if(!rest || !*rest)
   {
      /* It's important that commands that _must_ be in a message are listed as such. *
       * I hope I didn't miss any above... (kk). */
      send_to_user(uptr, "\002Usage: /msg %s %s LOGIN <handle> <password>\002", IrcNick, cptr->Name);
      return(FALSE);
   }
   handle = (char *) mysep(&rest, " ");
   password = (char *) mysep(&rest, " ");

   if(!password)
   {
      send_to_user(uptr, "Improper syntax.");
      send_to_user(uptr, "Usage: /msg %s %s LOGIN <handle> \002<password>\002", IrcNick, cptr->Name);
      return(FALSE);
   }
   if(!(hptr = FindHandle(cptr, handle)))
   {
      send_to_user(uptr, "Sorry, there is no handle `%s' in the %s userlist.", handle, cptr->Name);
      return(FALSE);
   }
   if(!strcmp(password, "*"))
   {
      send_to_user(uptr, "You cannot LOGIN if you don't have a password set.  Set a password with PASS.");
      send_to_user(uptr, "Try: /msg %s %s PASS * <new password>", IrcNick, cptr->Name);
      return(FALSE);
   }
   if(strcmp(hptr->Pass, crypt(password, hptr->Pass)))
   {
      send_to_user(uptr, "\002Login incorrect!\002  Invalid password.");
      return(FALSE);
   }
   for(ahptr = uptr->AuthHandles; ahptr; ahptr = ahptr->Next)
   {
       if(ahptr->cptr == cptr)
       {
           send_to_user(uptr, "You are already logged into %s as %s.", cptr->Name, ahptr->hptr->Handle);
           return(FALSE);
       }
   }
   /* All appears well, so add AuthHandle and AuthUser pointers etc */
   AddAuthUserPtr(uptr, hptr);
   AddAuthHandlePtr(hptr, uptr, cptr);
   send_to_user(uptr, "You are now logged into %s as %s.", cptr->Name, hptr->Handle);
   if (isinchan(cptr, uptr))
   {
      /* hptr->InChan = TRUE; done in setlastseen */
      SetLastSeen(cptr, hptr, JOIN);
       
   }

   /* If they are (co)owner, make sure the (co)owner last seen time gets reset. */
   if(IsChanOwner(hptr) || IsChanCoOwner(hptr))
   {
      /* cant do this because they may have been already counted from a mask.
       * cptr->OwnersOnline++; */

      /* If OwnersOnline is 0, there is currently an owner already counted in the channel,
       * so expiring wont be a problem. Otherwise, lets update the timestamp. */
      if(cptr->LastAccess > 0)
         cptr->LastAccess = (unsigned long) time(NULL);
      WriteChanOpFile(cptr);
   }
   return(TRUE);
}

/* Login to multiple channels at once
 * syntax: MLOGIN <handle> <password> <#channel1> [#channel2] [etc]
 */
int c_mlogin(C_PARAM)
{
   char* handle;
   char* password;
   char* channel;
   int OK;
   AuthHandleListType ahptr;

   if(!rest || !*rest)
   {
      send_to_user(uptr, "\002Usage: /msg %s MLOGIN <handle> <password> <channel> [channel channel ...]\002", IrcNick);
      return(FALSE);
   }
   handle = (char *) mysep(&rest, " ");
   password = (char *) mysep(&rest, " ");

   if(!handle)
   {
       send_to_user(uptr, "Incorrect syntax, please enter a handle.");
       send_to_user(uptr, "Usage: /msg %s MLOGIN \002<handle>\002 <password> <#channel 1> [#channel 2] [...]", IrcNick);
       return(FALSE);
   }
   if(!password)
   {
       send_to_user(uptr, "Incorrect syntax, please enter a password.");
       send_to_user(uptr, "Usage: /msg %s MLOGIN <handle> \002<password>\002 <#channel 1> [#channel 2] [...]", IrcNick);
       return(FALSE);
   }
   while (TRUE) {
      channel = (char *) mysep(&rest, " ,");
      if (!channel) {
          break;
      }
      cptr = GetChannel(channel);
      if (!cptr) {
          send_to_user(uptr, "The channel %s does not exist.", channel);
          continue;
      }
      if(!cptr->Regged) {
          send_to_user(uptr, "The channel %s has not been registered.", channel);
          continue;
      }
      if(!(hptr = FindHandle(cptr, handle)))
      {
         send_to_user(uptr, "There is no handle %s in the channel %s userlist.", handle, cptr->Name);
         continue;
      }
      if(!strcmp(hptr->Pass, crypt("*", hptr->Pass)))
      {
         send_to_user(uptr, "Your password was not set in the previous version of X2, for channel %s.", channel);
         send_to_user(uptr, "You have two options:");
         send_to_user(uptr, "  1.  Log on from a hostmask listed in the userlist, and use PASS to set your password (old password == *).");
         send_to_user(uptr, "  2.  Have someone higher up on the userlist change your password with RESETPASS."); 
         continue;
      }
      if(strcmp(hptr->Pass, crypt(password, hptr->Pass)))
      {
         send_to_user(uptr, "\002Login incorrect!\002  Incorrect password for channel %s.", channel);
         continue;
      }
      OK = 1;
      for(ahptr = uptr->AuthHandles; ahptr; ahptr = ahptr->Next)
      {
          if(ahptr->cptr == cptr)
          {
              send_to_user(uptr, "You are already logged into %s as %s.", cptr->Name, ahptr->hptr->Handle);
              /* we want to skip the AddAuthUserPtr stuff */
              OK = 0;
          }
      }
      if (OK == 1)
      {
         /* All appears well, so add AuthHandle and AuthUser pointers etc */
         AddAuthUserPtr(uptr, hptr);
         AddAuthHandlePtr(hptr, uptr, cptr);
         send_to_user(uptr, "You are now logged into %s as %s.", cptr->Name, hptr->Handle);
         if (isinchan(cptr, uptr))
             hptr->InChan = TRUE;
      }
   }
   return(TRUE);
}

int c_pass(C_PARAM)
{
  char *OldPass, *NewPass;
  char *CryptPass;

  if (!rest || !*rest)
  {
    send_to_user(uptr, "\002Usage: PASS <old password> <new password>\002");
    send_to_user(uptr, "\002Example: /msg %s #Operations PASS k48ad p83ad\002", IrcNick);
    return(FALSE);
  }
  OldPass = (char *) mysep(&rest, " ");
  NewPass = (char *) mysep(&rest, " "); 
  if(!NewPass)
  {
    send_to_user(uptr, "Incorrect syntax, please enter a new password.");
    send_to_user(uptr, "Usage: PASS <old password> \002<new password>\002");
    return(FALSE);
  }
  if(strcmp(hptr->Pass, crypt(OldPass, hptr->Pass)))
  {
     /* The old pass doesnt match. */
     send_to_user(uptr, "\002Password mismatch!\002  Old password does not match.");
     return(FALSE);
  }
  if(strlen(NewPass) > PASSWORDLEN)
  {
    send_to_user(uptr, "\002Your new password is too long.  It must be %d or fewer characters.\002", PASSWORDLEN);
    return(FALSE);
  }
  CryptPass = crypt(NewPass, "$1$92737837$");
  free(hptr->Pass);
  hptr->Pass = (char *) malloc(strlen(CryptPass) + 1);
  strcpy(hptr->Pass, CryptPass);
  send_to_user(uptr, "Your password in channel %s has been changed.", cptr->Name);
  WriteChanOpFile(cptr);
  return(TRUE);
}

int c_setinfo(C_PARAM)
{
  if (!rest || !rest[0])
  {
    send_to_user(uptr, "\002Usage: SETINFO <infoline>\002");
    send_to_user(uptr, "Set your info to \"*\" to disable it.");
    return(FALSE);
  }
  if (strlen(rest) >= MAXINFOLEN);
    rest[MAXINFOLEN] = '\0';
  SetInfo(cptr, hptr, rest);
  send_to_user(uptr, "Your info-line has been changed to: %s", rest);
  WriteChanOpFile(cptr);
  return(TRUE);
}

char* getscflags(char* rest, int* secret, int* nodelete, int* noinchan, int* suspended, int* noowner, int* total)
{
   char* mark = rest;
   char buff[MAXLEN];
   char* param = buff;


   for(mark = rest;*mark;mark++)
   {
       if(*mark == '-')
       {
           for(;*mark != ' ' && *mark != '\0';mark++)
           {
              *param = *mark;
              param++;
           }
           *param = '\0';
           param = buff;
           if(!strcmp(param, "--nodelete"))
              (*nodelete)++;
           else if(!strcmp(param, "--suspended"))
              (*suspended)++;
           else if(!strcmp(param, "--secret"))
              (*secret)++;
           else if(!strcmp(param, "--noinchan"))
              (*noinchan)++;
           else if(!strcmp(param, "--noowner"))
              (*noowner)++;
           else if(!strcmp(param, "-t"))
              (*total)++;
           else
           {
               Debug(DBGNOTICE, "getscflags(): Unknown paramiter");
           }
           if(*mark == ' ')
              continue;
           else  /* mark == \0 */
           {
              return(""); /* no mask given */
           }
       }
       else
       {
           return(mark);
       }
   }
   return(NULL);
}

int countowners(ChannelListType cptr)
{
  int i = 0;
  HandleListType hptr;

  for(hptr = cptr->Handles;hptr;hptr = hptr->Next)
  {
      if(hptr->Level == 'W')
         i++;
  }
  return(i);
}

/* TODO : This is not finished.. */
int c_searchchan(C_PARAM)
{
  int i, count = 0;
  ChannelListType TempChanPtr;
  int secret = 0, nodelete = 0, noinchan = 0, suspended = 0, total = 0, noowner = 0;
  char* topicsearch;
  char* chansearch;
  char buff[MAXLEN];

  if(!(rest && *rest))
  {
      send_to_user(uptr, "\002Usage: SEARCHCHAN [options] <name> [topic]\002");
      send_to_user(uptr, "The fields `name' and `topic' can have * and ? wildcards.");
      send_to_user(uptr, "Options (most of which aren't implemented yet):");
      send_to_user(uptr, "    --secret    : Show only secret files.");
      send_to_user(uptr, "    --nodelete  : Show only channels with the NODELETE flag.");
      send_to_user(uptr, "    --noinchan  : Show only channels X2 is not in (INCHAN == 0).");
      send_to_user(uptr, "    --suspended : Show only suspended channels.");
      send_to_user(uptr, "    --noowner   : Show only chans with no owner.");
      send_to_user(uptr, "    -t          : Provide a total at the end");
      send_to_user(uptr, "Options can be combined.");
      return(FALSE);
  }
  rest = getscflags(rest, &secret, &nodelete, &noinchan, &suspended, &noowner, &total);
/*  send_to_user(uptr, "secret=%d, nodelete=%d, noinchan=%d, suspended=%d, total=%d :%s",
 *                      secret,    nodelete,    noinchan,    suspended, total, rest);
 */
  count = 0;
  chansearch = (char *) strtok(rest, " ");
  if(!chansearch)
       chansearch = "*";
  topicsearch = (char *) strtok(NULL, "\0\r\n");
  if(!topicsearch)
       topicsearch="*";

  send_to_user(uptr, "Searching channels matching: %s %s", chansearch, topicsearch);
  send_to_user(uptr, "=--Name-----------------------=----Topic-----------=");
  for(i = 0; i< CHANSEGSIZE; i++)
  {
     if(Channels->Table[i])
     {
           TempChanPtr = Channels->Table[i];
           for(TempChanPtr = Channels->Table[i];TempChanPtr;TempChanPtr = TempChanPtr->Next)
           {
              if(TempChanPtr->Regged &&
                 !matches(chansearch, TempChanPtr->Name) &&
                 !matches(topicsearch, TempChanPtr->Topic?TempChanPtr->Topic:""))
              {
                 if(nodelete && (TempChanPtr->settings[NODELETE] == 0))
                    continue;
                 if(noinchan && (TempChanPtr->settings[INCHAN] > 0))
                    continue;
                 if(suspended && (TempChanPtr->Suspended == NULL))
                    continue;
                 if(noowner && (countowners(TempChanPtr) > 0))
                    continue;
                 if(((TempChanPtr->PlusModes & CMODEs) || (TempChanPtr->PlusModes & CMODEp)))
                 {
                   if(IsOper(uptr))
                   {
                      send_to_user(uptr, "\002(SECRET)\002 %20s : %s", TempChanPtr->Name, TempChanPtr->Suspended?TempChanPtr->Suspended:(TempChanPtr->Topic?TempChanPtr->Topic:""));
                      count++;
                   }
                 }
                 else
                 {
                   if(secret == 0)
                   {
                       if (TempChanPtr->Suspended)
                       {
                         send_to_user(uptr, "%29s : %s %s", TempChanPtr->Name, t2s(buff, TempChanPtr->SuspendedTime), TempChanPtr->Suspended);
                       }
                       else
                       {
                         send_to_user(uptr, "%29s : %s", TempChanPtr->Name, TempChanPtr->Topic?TempChanPtr->Topic:"");
                       }
                       count++;
                   }
                 }
              }
           } 
     }
  }
  if(!count)
    send_to_user(uptr, "No matching channels found.");
  send_to_user(uptr, "=-----------------------------=--------------------=");
  if(total)
    send_to_user(uptr, "Total matching channels: %d", count);

  return(TRUE);
}

int c_rehash(C_PARAM)
{
  struct stat StatStruct;

  send_wallops("\002Rehash started by %s.\002", uptr->Nick);
  send_to_user(uptr, "\002Rescanning the news file...\002");
  stat(MESSAGEFILENAME, &StatStruct);
  NetMessage.Timestamp = StatStruct.st_mtime;
  send_to_user(uptr, "   NetMessage timestamp is %lu.", NetMessage.Timestamp);
  send_to_user(uptr, "\002Rehash completed.\002");
  return(TRUE);
}

int c_quit(C_PARAM)
{
  if (rest && *rest)
  {
      send_quit("\002Quit\002 by %s: \002(%s)\002", uptr->Nick, rest);
  }
  else
  {
      send_quit("\002Quit\002 by %s.", uptr->Nick);
  }
  send_squit(Uplink, "Quitting as requested by %s.", uptr->Nick);
  KillServerList();
  unlink("./x2.pid");
  exit(0);
}

int c_suspend(C_PARAM)
{
  char Buf[MAXLEN];
  char Buf2[MAXLEN];
  char* Arg1;
  char* Arg2;
  char* Arg3;
  char* Arg4;
  int i;
  int len;
  int Value;

  if (cptr->Suspended)
  {
    send_to_user(uptr, "Channel %s is already marked as suspended: %s", cptr->Name, cptr->Suspended);
    return(FALSE);
  }

  Arg1 = strtok(rest, " ");
  Arg2 = strtok(NULL, "");
  
  if (!Arg1 || !Arg2) 
  {
    send_to_user(uptr, "Usage: SUSPEND <#channel> <time>:<unit> <reason>");
    return(FALSE);
  }

  if (!strchr(Arg1, ':'))
  {
    send_to_user(uptr, "\002Illegal time argument.\002");
    return(FALSE);
  }
  Arg3 = (char *) strtok(Arg1, ":");
  Arg4 = (char *) strtok(NULL, " ");
  if (!Arg3 || !Arg4)
  {
    send_to_user(uptr, "\002Invalid time argument.\002");
    return(FALSE);
  }

  len = strlen(Arg3);
  if (len > 5)
  {
    send_to_user(uptr, "\002Illegal time value.\002");
    return(FALSE);
  }

  for (i = 0; i < len; i++)
  {
    if (!isdigit(Arg3[i]))
    {
      send_to_user(uptr, "\002Illegal time value.\002");
      return(FALSE);
    }
  }
  Value = atoi(Arg3);

  if ((strlen(Arg4) > 4))
  {
    send_to_user(uptr, "\002Illegal time period.\002");
    return(FALSE);
  }

  if (!strcasecmp(Arg4, "MIN"))
    cptr->SuspendedTime = time(NULL) + Value * 60;
  else if (!strcasecmp(Arg4, "M"))
    cptr->SuspendedTime = time(NULL) + Value * 60;
  else if (!strcasecmp(Arg4, "HOUR"))
    cptr->SuspendedTime = time(NULL) + Value * 3600;
  else if (!strcasecmp(Arg4, "H"))
    cptr->SuspendedTime = time(NULL) + Value * 3600;
  else if (!strcasecmp(Arg4, "D"))
    cptr->SuspendedTime = time(NULL) + Value * 86400;  
  else if (!strcasecmp(Arg4, "DAY"))
    cptr->SuspendedTime = time(NULL) + Value * 86400;
  else
  {
    send_to_user(uptr, "\002Illegal time period.\002  Use <#>:MIN/HOUR/DAY");
    return(FALSE);
  }

  sprintf(Buf2, "<%s> %s [%s]", uptr->Nick, Arg2, t2s(Buf, time(NULL)));
  cptr->Suspended = malloc(strlen(Buf2) + 1);
  strcpy(cptr->Suspended, Buf2);
  send_to_user(uptr, "Channel %s is marked as suspended: %s", cptr->Name, Arg2);
  if (cptr->settings[INCHAN] != 0)
    send_part(cptr, "Suspended by %s.", uptr->Nick);
  cptr->settings[INCHAN] = 0;
  WriteChanOpFile(cptr);
  return(TRUE);
}

int c_unsuspend(C_PARAM)
{
  char ModeString[LINEBUFFSIZE];

  if (!cptr->Suspended)
  {
    send_to_user(uptr, "Channel %s is not suspended.", cptr->Name);
    return(FALSE);
  }
  free(cptr->Suspended);
  cptr->Suspended = NULL;
  cptr->settings[INCHAN] = 1;
  send_to_user(uptr, "Unsuspending channel %s.", cptr->Name);
  send_join(cptr);
  send_opchan(cptr);
  send_chanmode(cptr, "%s", GetModeString(ModeString, cptr->PlusModes, 
                            cptr->MinModes, cptr->Key, cptr->Limit));
  if (strcmp(cptr->Topic, "*") && cptr->settings[TOPICFREQ])
      send_topic(cptr, cptr->Topic);
  WriteChanOpFile(cptr);
  return(TRUE);
}

int c_disconnect(C_PARAM)
{
  send_quit("\002Disconnected\002 by %s (port: %d).", uptr->Nick, SERVERPORT);
  send_squit(Uplink, "\002Disconnected\002 by %s (port: %d).", uptr->Nick, SERVERPORT);
  AUTOCONNECT = 0;
  SockConnected = FALSE;
  return(TRUE);
}

/* This is replaced with c_login */
int c_auth(C_PARAM)
{
  send_to_user(uptr, "The command \"AUTH\" is no longer in use.  Please use the new command, \002LOGIN\002.");
  send_to_user(uptr, "See: /msg %s HELP LOGIN", IrcNick);
  return(FALSE);
}

int c_seen(C_PARAM)
{
  time_t difference;
  char buf[MAXLEN] = "", tmp[MAXLEN] = "";
  char sendbuff[MAXLEN];
  HandleListType Handle;

  int days, hours, minutes, seconds;

  rest = strtok(rest, " ");
  if (!rest)
  {
    send_to_user(uptr, "\002Usage: SEEN <handle>\002");
    send_to_user(uptr, "`handle' is the handle in the channel's userlist.");
    return(FALSE);
  }
  if (!(Handle = FindHandle(cptr, rest)))
  {
    sprintf(sendbuff, "Sorry, nobody on the %s userlist has a handle of %s.", cptr->Name, rest);
  }
  else if (IsInChan(Handle))
  {
    sprintf(sendbuff, "%s is in %s right now.", rest, cptr->Name);
  }
  else if (Handle->LastSeen == 0)
  {
     sprintf(sendbuff, "%s has not yet been seen.", rest);
  }
  else
  {
     difference = time(NULL) - Handle->LastSeen;
     days = difference / 86400;
     difference = difference - (days * 86400);
     hours = difference / 3600;
     difference = difference - (hours * 3600);
     minutes = difference / 60;
     difference = difference - (minutes * 60);
     seconds = difference;
     /* put our time together */
     if (days == 1)
       strcat(buf, "1 day, ");
     else if (days > 1)
     {
       sprintf(tmp, "%d days, ", days);
       strcat(buf, tmp);
     }
     if (hours == 1)
       strcat(buf, "1 hour, ");
     else if (hours > 1)
     {
       sprintf(tmp, "%d hours, ", hours);
       strcat(buf, tmp);
     }
     if (!days)	/* why show minutes if it's been over a day? */
     {
       if (minutes == 1)
         sprintf(buf + strlen(buf), "1 minute%s", (seconds != 0 ? ", " : " "));
       else if (minutes > 1)
       {
         sprintf(tmp, "%d minutes%s", minutes, (seconds != 0 ? ", " : " "));
         strcat(buf, tmp);
       }
       if (!hours)			/* no reason to display seconds if it has been over an hour */
       {
         if (seconds == 1)
           strcat(buf, "1 second");
         else if (seconds > 1)
         {
            sprintf(tmp, "%d seconds", seconds);
            strcat(buf, tmp);
         }
       }
     }
     sprintf(sendbuff, "I last saw %s on %s \002%s\002 ago", rest, cptr->Name, buf);
  }
  if( said_inptr && PublicReplyAllowed(hptr))
  {
      send_to_channel(said_inptr, "\002%s:\002 %s", uptr->Nick, sendbuff);
  }
  else
  {
      send_to_user(uptr, "%s", sendbuff);
  }
  return(TRUE);
}

int c_access(C_PARAM)
{
  UserListType target;

  if (!strcmp(rest, "")) {
    target = uptr;
    hptr = FindBestUserMatch(cptr, target);
  }
  else {
    if (!(target = FindNick(rest, UsersN)))
    {
      hptr = FindHandle(cptr, rest);
      if (!hptr) {
        send_to_user(uptr, "\002Unknown user!\002  (%s)", rest);
        return(FALSE);
      }
    }
    else
      hptr = FindBestUserMatch(cptr, target);
  }
  if (hptr && userlevelbychar(hptr->Level) > 0)
  {
    if (target)
      send_to_user(uptr, "%s : \002%s\002 on %s, under handle: \002%s\002.",
	target->Nick, levelnamefromchar(hptr->Level), cptr->Name, hptr->Handle);
    else
      send_to_user(uptr, "%s : \002%s\002 on %s.", hptr->Handle, levelnamefromchar(hptr->Level),
        cptr->Name);
    if (strcmp(hptr->Info, "*"))
      send_to_user(uptr, "[%s] %s", target ? target->Nick : hptr->Handle, hptr->Info);
    if (IsChanAOp(hptr))
      send_to_user(uptr, "%s has auto-op turned \002ON\002.", target ? target->Nick : hptr->Handle);
    else
      send_to_user(uptr, "%s has auto-op turned \002OFF\002.", target ? target->Nick : hptr->Handle);
  }
  else if (target && IsChanLamer(cptr, target))
    send_to_user(uptr, "%s is a lamer from %s!", target->Nick, cptr->Name);
  else
    send_to_user(uptr, "%s has not authenticated to %s", target->Nick, cptr->Name);

  if (!strcasecmp(IrcNick, rest))
    send_to_user(uptr, "\002%s\002 is the channel services bot.  Type: /msg %s HELP for info on how to use it.", IrcNick, IrcNick);
  if (target && IsOper(target))
    send_to_user(uptr, "%s is an \002IRC Operator\002.", target->Nick);
  if (target && IsGodOper(target))
    send_to_user(uptr, "%s has \002security override (GOD)\002 enabled.", target->Nick);
  if (target && IsModeK(target))
    send_to_user(uptr, "%s is a \002no-kick service (+k)\002.", target->Nick);
  return(TRUE);
}

int c_commands(C_PARAM)
{
   int i, length = 0, level;
   char buff[MAXLEN] = "";
   char *levelstr;

   rest = (char *) mysep(&rest, " ");
   /* If they gave us a param, show that. otherwise, use their level */
   if(rest && *rest)
   {
        if(!strcasecmp(rest, "OWNER"))
           level = userlevelbychar('W');
        else if(!strcasecmp(rest, "CO") || !strcasecmp(rest, "COOWNER"))
           level = userlevelbychar('C');
        else if(!strcasecmp(rest, "MASTER"))
           level = userlevelbychar('M');
        else if(!strcasecmp(rest, "OP"))
           level = userlevelbychar('O');
        else if(!strcasecmp(rest, "PEON"))
           level = userlevelbychar('P');
        else if(!strcasecmp(rest, "NONE"))
           level = 0;
        else if(!strcasecmp(rest, "IRCOP"))
           level = 999;
        else
        {
           send_to_user(uptr, "Unknown level name: \"%s\".", rest);
           return(FALSE);
        }
        levelstr = rest;
   }
   else
   {
      if(hptr)
      {
         level = userlevelbychar(hptr->Level);
         levelstr = levelnamefromchar(hptr->Level);
      }
      else
      {
         level = 0;
         levelstr = "NONE";
      }
   }
   send_to_user(uptr, "=----- Commands available to %s -----=", levelstr);
   for(i=0;(on_msg_commands[i].name != NULL); i++)
   {
       if(on_msg_commands[i].ChanLvl > level)
          continue;
       if(on_msg_commands[i].IRCOp && level < 900)
          continue;
       strcat(buff, " ");
       strcat(buff, on_msg_commands[i].name);
       length += strlen(on_msg_commands[i].name) + 1;
       if(length > 40)
       {
          send_to_user(uptr, "%s", buff);
          strcpy(buff, "");
          length = 0;
       }
   }
   send_to_user(uptr, "%s", buff);
   send_to_user(uptr, "=----- End of list -----=");
   return(TRUE);
}

int c_settime(C_PARAM)
{
  time_t now, l;
  char buf[MAXLEN];
  char *param;
  char *ammount;

  now = time(NULL);

  param = (char *) strtok(rest, " ");
  ammount = (char *) strtok(NULL, " ");
  if(ammount)
  {
    if(!strcmp(param, "POWER"))
    {
       if(atol(ammount) != 0)
       {
         send_wallops("%s is bringing the timestamp up SLOWLY.", uptr->Nick);
         for(l=atol(ammount);l<now;l+=999999)
         {
              send_settime(l);
         }
       }
    }
    return(TRUE);
  }
  send_settime(now);
  send_to_user(uptr, "\002Time set to %lu: %s.\002", now, t2s(buf, now));
  NetOffset = 0;
  return(TRUE);
}

int c_gline(C_PARAM)
{
  int i;
  int Value, len;
  char *Arg1;
  char *Arg2;
  char *Arg3;
  char *Arg4;
  char *Address;
  char *UpCasePer;
  time_t gtime;
  char buff[MAXLEN];

  Address = strtok(rest, " ");
  Arg1 = strtok(NULL, " ");
  Arg2 = strtok(NULL, "\0");
  Address = strtok(Address, "$");
  if (!Arg1 || !Arg2 || !Address)
  {
    send_to_user(uptr, "\002Usage: GLINE <address> <time>:<unit> <reason>\002");
    return(FALSE);
  }
  if (!GlineAddyOk(Address))
  {
    send_to_user(uptr, "\002Illegal mask.\002");
    return(FALSE);
  }
  if (!strchr(Arg1, ':'))
  {
    send_to_user(uptr, "\002Illegal time argument.\002");
    return(FALSE);
  }
  Arg3 = (char *) strtok(Arg1, ":");
  Arg4 = (char *) strtok(NULL, " \0");
  if (!Arg3 || !Arg4)
  {
    send_to_user(uptr, "\002Invalid time argument.\002");
    return(FALSE);
  }
  len = strlen(Arg3);

  if (len > 5)
  {
    send_to_user(uptr, "\002Illegal time value.\002");
    return(FALSE);
  }

  for (i = 0; i < len; i++)
  {
    if (!isdigit(Arg3[i]))
    {
      send_to_user(uptr, "\002Illegal time value.\002");
      return(FALSE);
    }
  }
  Value = atoi(Arg3);

  if ((strlen(Arg4) > 4))
  {
    send_to_user(uptr, "\002Illegal time period.\002");
    return(FALSE);
  }
  UpCasePer = UpCase(Arg4);

  /* 11-21-97 GLINE <TS> is now GLINE <target (* for all)> ...
   * to be compatable with ircu2.10.  */
  if (!strcmp(UpCasePer, "MIN"))
    gtime = Value * 60;
  else if (!strcmp(UpCasePer, "M"))
    gtime = Value * 60;
  else if (!strcmp(UpCasePer, "HOUR"))
    gtime = Value * 3600;
  else if (!strcmp(UpCasePer, "H"))
    gtime = Value * 3600;
  else if (!strcmp(UpCasePer, "D"))
    gtime = Value * 86400;  
  else if (!strcmp(UpCasePer, "DAY"))
    gtime = Value * 86400;
  else
  {
    send_to_user(uptr, "\002Illegal time period.\002  Use <#>:MIN/HOUR/DAY");
    free(UpCasePer);
    return(FALSE);
  }
  sprintf(buff, "<%s> %s", uptr->Nick, Arg2); /* for ProcessAddGline() */
  send_gline(Address, gtime, "%s", buff);
  ProcessAddGline(Address, gtime, buff);

  send_to_user(uptr, "Adding GLINE for %s :<%s> %s", Address, uptr->Nick, Arg2);
  WriteGlineDat();
  free(UpCasePer);
  return(TRUE);
}

int c_sgline(C_PARAM)
{
  char *Address;
  char buff[MAXLEN];

  Address = strtok(rest, " ");
  rest = strtok(NULL, "\0");
  if (!rest || !Address)
  {
    send_to_user(uptr, "\002Usage: SGLINE <address> <reason>\002");
    return(FALSE);
  }
  if (!GlineAddyOk(Address))
  {
    send_to_user(uptr, "\002Illegal mask.\002");
    return(FALSE);
  }

  sprintf(buff, "<%s> %s", uptr->Nick, rest);
  ProcessAddSGline(Address, time(NULL), buff);
  send_to_user(uptr, "Adding SuperGLINE for %s :<%s> %s", Address, uptr->Nick, rest);
  WriteSGlineDat();
  return(TRUE);
}

int c_spammer(C_PARAM)
{
    char* nick;
    char reason[MAXLEN];
    char* buff;
    char Address[MAXLEN];
    char* IP;
    UserListType target;
    FILE *SPAMFILE;

    nick = (char *) mysep(&rest, " ");
    if (!nick)
    {
        send_to_user(uptr, "\002Usage: SPAMMER <nick> <reason>\002");
        return(FALSE);
    }
    buff = (char *) mysep(&rest, "\0");
    if(!buff || !*buff)
        buff = "";
    sprintf(reason, "<SPAMMER> <%s> We do not allow advertising on AfterNET! (%s)", uptr->Nick, target->Nick);
    if((target = FindNick(nick, UsersN)))
    {
        if (IsModeK(target))
        {
            send_to_user(uptr, "\002Cannot spamassassinate a service!\002");
            return(FALSE);
        }
        IP = unsigned32ToDotQuad(target->IP);
        sprintf(Address, "*@%s", IP);
        send_gline(Address, 86400*2, "%s", reason);
        ProcessAddGline(Address, 86400*2, reason);
        if((SPAMFILE = fopen(SPAMFILENAME, "a")))
        {
            fprintf(SPAMFILE, "%s - %s (%s!%s@%s) - [%s] %s\n", time2str(time(NULL)), IP, target->Nick, target->UserID, target->Address, buff, reason);
            fclose(SPAMFILE);
        }
        return(TRUE);
    }
    else
    {
        send_to_user(uptr, "That user is not online. Check spelling and try again.");
	return(FALSE);
    }
}

int c_unglineall(C_PARAM)
{
   GlineListType *gptr, *gtmp;

  gptr = Glines;
  while(gptr)
  {
    gtmp = gptr->Next;
    send_ungline(gptr->address);
    free(gptr);
    gptr = gtmp;
  }
  Glines = NULL;
  WriteGlineDat();
  return(TRUE);
}

int c_ungline(C_PARAM)
{
  if (!GlineAddyOk(rest))
  {
    send_to_user(uptr, "Illegal mask!");
    return(FALSE);
  }
  if(!GetGline(rest))
  {
     send_to_user(uptr, "GLINE \"%s\" not found in memory.  Clearing from server anyway...", rest);
  }
  send_ungline(rest);
  ProcessRemoveGline(rest);
  send_to_user(uptr, "GLINE removed.");
  return(TRUE);
}

int c_unsgline(C_PARAM)
{
/* (kk) This was commented out when I got here, so I'm going to comment on why it's
 *      commented.  :P  Bah, it just makes it so you can clear out invalid masks that
 *      are already in the list.  :P
  if (!GlineAddyOk(rest))
  {
    send_to_user(uptr, "Illegal mask!");
    return(FALSE);
  }
*/
  if(!GetSuperGline(rest))
  {
     send_to_user(uptr, "No SuperGLINE: %s", rest);
     return(FALSE);
  }
  ProcessRemoveSGline(rest);
  send_to_user(uptr, "SuperGLINE %s removed.", rest);
  return(TRUE);
}

int c_setglines(C_PARAM)
{
  GlineListType *gptr;
  int FORCE = 0;

  gptr = Glines;
  if (!gptr)
  {
    send_to_user(uptr, "There are currently no GLINEs in my database.");
    return(FALSE);
  }
  if (!strcasecmp("HARD", rest))
  {
    FORCE = TRUE;
  }
  send_wallops("\002%s is resynchronizing the GLINEs...", uptr->Nick);
  while (gptr)
  {
    if (FORCE)
    {
      send_ungline(gptr->address);
    }
    send_gline(gptr->address, gptr->time - NetTime(), "%s", gptr->reason);
    gptr = gptr->Next;
  }
  return(TRUE);
}

int c_glines(C_PARAM) 
{
  GlineListType *gptr;
  char buff[MAXLEN];

  gptr = Glines;
  if (!gptr)
  {
    send_to_user(uptr, "There are no GLINEs in my database.");
    return(FALSE);
  }
  send_to_user(uptr, "GLINE List for: %s", rest);
  send_to_user(uptr, "=-------------Address--------------Time------------=");

  while (gptr)
  {
    if(!(*rest) || strstr(gptr->address, rest))
    {
       send_to_user(uptr, " %s  %s", gptr->address, t2s(buff, gptr->time));
       if(*rest)
         send_to_user(uptr, "      Reason : %s", gptr->reason);
    }
    gptr = gptr->Next;
  }
  send_to_user(uptr, "=----------------- End of GLINEs ----------------------=");
  return(TRUE);
}

int c_sglines(C_PARAM) 
{
  GlineListType *gptr;
  char buff[MAXLEN];

  gptr = SuperGlines;
  if (!gptr)
  {
    send_to_user(uptr, "There are no SuperGLINEs in my database.");
    return(FALSE);
  }
  send_to_user(uptr, "SuperGLINE List");
  send_to_user(uptr, "=-------------Address--------------Time added------=");

  while (gptr)
  {
    send_to_user(uptr, " %23s  %s", gptr->address, t2s(buff, gptr->time));
    send_to_user(uptr, "                  Reason : %s", gptr->reason);
    gptr = gptr->Next;
  }
  send_to_user(uptr, "=----------------- End of SuperGlines ---------------------=");
  return(TRUE);
}

/* TODO : Put time stuff from GLINE, unvisited etc.  All into
 * one function, maybe in misc.c.   Too much duplicated code.
 */
int c_unvisited(C_PARAM)
{
  int i;
  ChannelListType ChanPtr;
  int Value, len;
  char *Arg1;
  char *Arg2;
  int count = 0;
  char *UpCasePer;
  int w, c, m, o, p, b;

  if (!strchr(rest, ':'))
  {
    send_to_user(uptr, "\002Usage: UNVISITED <time>:<unit>\002");
    return(FALSE);
  }
  Arg1 = (char *) strtok(rest, ":");
  Arg2 = (char *) strtok(NULL, " \r\n");
  len = strlen(Arg1);
  if (!Arg2)
  {
    send_to_user(uptr, "\002Illegal time period.\002  Use DAY, HOUR, or MIN.");
    return(FALSE);
  }
  if (len > 5)
  {
    send_to_user(uptr, "\002Illegal time value.\002");
    return(FALSE);
  }

  for (i = 0; i < len; i++)
  {
    if (!isdigit(Arg1[i]))
    {
      send_to_user(uptr, "\002Illegal time value.\002");
      return(FALSE);
    }
  }
  Value = atoi(Arg1);

  if (strlen(Arg2) > 4)
  {
    send_to_user(uptr, "\002Illegal time period.\002");
    return(FALSE);
  }
  UpCasePer = UpCase(Arg2);

  if ((strcmp(UpCasePer, "MIN")) && (strcmp(UpCasePer, "HOUR")) && (strcmp(UpCasePer, "DAY")))
  {
    send_to_user(uptr, "\002Illegal time period.\002");
    free(UpCasePer);
    return(FALSE);
  }
  send_to_user(uptr, "-----------------------------------------------------");
  send_to_user(uptr, "Channels unvisited by an owner for %d %s(S) or more :", Value, UpCasePer);
  send_to_user(uptr, " ");

  for (i = 0; i < CHANSEGSIZE; i++)
  {
    ChanPtr = Channels->Table[i];
    while (ChanPtr)
    {
      if (!ChanPtr->Regged)
      {
	ChanPtr = ChanPtr->Next;
	break;
      }
      if (ChanPtr->LastAccess == 0)
      {
	ChanPtr = ChanPtr->Next;
	break;
      }
      if(ChanPtr->settings[NODELETE])
      {
        ChanPtr = ChanPtr->Next;
        break;
      }
      GetChanStats(ChanPtr, &w, &c, &m, &o, &p, &b);
      if (!strcmp(UpCasePer, "MIN"))
      {
	if (((time(NULL) - ChanPtr->LastAccess) / 60) >= Value)
	{
	  count++;
	  send_to_user(uptr, " %s  -   %d Min(s): [%dw %dc %dm %do %dp %db]", ChanPtr->Name, (time(NULL) - ChanPtr->LastAccess) / 60,
		       w, c, m, o, p, b);
	}
      }
      else if (!strcmp(UpCasePer, "HOUR"))
      {
	if (((time(NULL) - ChanPtr->LastAccess) / 3600) >= Value)
	{
	  count++;
	  send_to_user(uptr, " %s  -   %d Hour(s): [%dw %dc %dm %do %dp %db]", ChanPtr->Name, (time(NULL) - ChanPtr->LastAccess) / 3600,
		       w, c, m, o, p, b);
	}
      }
      else
      {
	if (((time(NULL) - ChanPtr->LastAccess) / 86400) >= Value)
	{
	  count++;
	  send_to_user(uptr, " %s  -   %d Day(s): [%dw %dc %dm %do %dp %db]", ChanPtr->Name, (time(NULL) - ChanPtr->LastAccess) / 86400,
		       w, c, m, o, p, b);
	}
      }
      ChanPtr = ChanPtr->Next;
    }
  }
  if (!count)
    send_to_user(uptr, " None");
  send_to_user(uptr, "-----------------------------------------------------");
  free(UpCasePer);
  return(TRUE);
}

int c_global(C_PARAM)
{
  int i;
  ChannelListType ChanPtr;

  if(!rest || !*rest)
  {
       send_to_user(uptr, "\002Usage:\002 GLOBAL <message>");
       return(FALSE);
  }
  for (i = 0; i < CHANSEGSIZE; i++)
  {
    ChanPtr = Channels->Table[i];
    while (ChanPtr)
    {
      if(ChanPtr->Regged == TRUE)
      {
         send_to_channel(ChanPtr, "\002Channel Service Announcement from <%s>:\002", uptr->Nick);
         send_to_channel(ChanPtr, "%s", rest);
      }
         ChanPtr = ChanPtr->Next;
    }
  }
  return(TRUE);
}

int c_who(C_PARAM)
{
   AuthUserListType auptr;

   send_to_user(uptr, "=---- Who's logged into %s ----=", cptr->Name);
   for(hptr = cptr->Handles;hptr;hptr = hptr->Next)
   {
      for(auptr = hptr->AuthUsers;auptr;auptr = auptr->Next)
      {
           send_to_user(uptr, "%s is logged in as %s.", auptr->uptr->Nick, hptr->Handle);
      }
   }
   send_to_user(uptr, "=-------- End of Who list --------=");
   return(TRUE);
}

int c_chanstat(C_PARAM)
{
  ChannelUsersListType cuptr;
  char ModeString[LINEBUFFSIZE];
  char timestring[LINEBUFFSIZE];
  time_t GoneTime;
  int OwnerNumber = 0, CoOwnerNumber = 0, ManagerNumber = 0, OpNumber = 0, PeonNumber = 0, LamerNumber = 0;

  if(cptr->Regged == FALSE)
  {
     send_to_user(uptr, "Unregistered channel %s has %d users.", cptr->Name, cptr->NumOfNicks);
     for(cuptr = cptr->Joined;cuptr;cuptr = cuptr->Next)
     {
         send_to_user(uptr, " %s is in %s", cuptr->User->Nick, cptr->Name);
     }
     send_to_user(uptr, "End of list.");
     return(TRUE);
  }
  GetChanStats(cptr, &OwnerNumber, &CoOwnerNumber, &ManagerNumber, 
      &OpNumber, &PeonNumber, &LamerNumber);
  strcpy(timestring, ctime(&cptr->CreationDate));
  timestring[strlen(timestring) - 1] = '\0';
  GoneTime = time(NULL) - cptr->LastAccess;
  send_to_user(uptr, "=---------------------------------=");
  send_to_user(uptr, "        Channel Statistics         ");
  send_to_user(uptr, "=---------------------------------=");
  send_to_user(uptr, " Channel Name  : \002%s\002 ", cptr->Name);
  send_to_user(uptr, " Created       : \002%s\002 by \002%s\002", timestring, cptr->Reggedby);

  if (!strcmp(cptr->Topic, "*"))
    send_to_user(uptr, " No default topic.");
  else
    send_to_user(uptr, " Default Topic : %s", cptr->Topic);
  if(IsChanOwner(hptr) || IsOper(uptr))
  {
    send_to_user(uptr, " Owner E-mail  : %s - Visible to owner and IRC Ops only.", cptr->Email);
  }
  if(strcmp(cptr->Description, "*"))
    send_to_user(uptr, " Description   : %s", cptr->Description);
  else
    send_to_user(uptr, " No Description");
  if(strcmp(cptr->URL, "*"))
    send_to_user(uptr, " URL           : %s", cptr->URL);
  else
    send_to_user(uptr, " No URL");
  send_to_user(uptr, " Default mode  : %s", GetModeString(ModeString, cptr->PlusModes, cptr->MinModes, "????", cptr->Limit));
  send_to_user(uptr, " ");
  send_to_user(uptr, " Num Of Owners  : %3d  | CoOwners : %3d", OwnerNumber, CoOwnerNumber);
  send_to_user(uptr, " Num Of Managers: %3d", ManagerNumber);
  send_to_user(uptr, " Num Of Ops     : %3d  | Peons    : %3d", OpNumber, PeonNumber);
  send_to_user(uptr, " Num Of Lamers  : %3d", LamerNumber);
  send_to_user(uptr, " ");
  if (OwnerNumber > 0)
  {
    if (cptr->LastAccess == 0)
      send_to_user(uptr, " A (Co)Owner or Manager is currently in the channel.", GoneTime);
    else if (GoneTime < 60)
      send_to_user(uptr, "\002 Last time visited by a (Co)Owner/Manager:\002 %d secs ago.", GoneTime);
    else if (GoneTime < 3600)
      send_to_user(uptr, "\002 Last time visited by a (Co)Owner/Manager:\002 %d mins ago.", GoneTime / 60);
    else if (GoneTime < 86400)
      send_to_user(uptr, "\002 Last time visited by a (Co)Owner/Manager:\002 %d hours ago.", GoneTime / 3600);
    else
      send_to_user(uptr, "\002 Last time visited by a (Co)Owner/Manager:\002 %d days ago.", GoneTime / 86400);
  }
  send_to_user(uptr, "\002Number of users currently in channel:\002 %d", cptr->NumOfNicks);
  if (cptr->Suspended)
  {
    send_to_user(uptr, "\002 This channel has been suspended:\002 %s", cptr->Suspended);
  }
  return(TRUE);
}


int c_last(C_PARAM)
{
   char *arg;
   int numoflines;

   arg = (char *) strtok(rest, " ");


   if(!arg)
   {
      send_to_user(uptr, "USAGE: last <n>");
      send_to_user(uptr, "where <n> is the number of lines you wish to see.");
      return(FALSE);
   }
   numoflines = atoi(rest);

   if(numoflines < 1 || numoflines > 200)
   {
       send_to_user(uptr, "Invalid argument.  must be 1-200");
       return(FALSE);
   }
   ShowLog(uptr, cptr, "*", "*", "*", "*", 0, time(NULL), numoflines);
   return(TRUE);
}

int c_mode(C_PARAM)
{
  char ModeString[LINEBUFFSIZE];

  send_chanmode(cptr, "-k %s", cptr->CurrentKey);
  send_chanmode(cptr, "%s", GetModeString(ModeString, cptr->PlusModes, cptr->MinModes, cptr->Key, cptr->Limit));
  return(TRUE);
}

int c_set(C_PARAM)
{
  return(do_set(C_PARAM_CALL));
}

int c_xset(C_PARAM)
{
  return(do_xset(C_PARAM_CALL));
}

int c_levels(C_PARAM)
{
  send_to_user(uptr, "\002OWNER\002   - Only one person per channel can have this access level.");
  send_to_user(uptr, "\002COOWNER\002 - Can do the same as Owners, except for modifying CoOwners.");
  send_to_user(uptr, "\002MANAGER\002 - Can add OP, PEON, and LAMER levels, but cannot set permanent modes.");
  send_to_user(uptr, "\002OP\002      - Cannot add users or lamers, but can use x2 to do remote commands such as kick and op.");
  send_to_user(uptr, "\002PEON\002    - Is known in the seen command, can be given +v, but no other access.");
  send_to_user(uptr, "\002LAMER\002   - Is kickbanned from the channel on sight.  Reason optionally given.");
  return(TRUE);
}

int c_addpeon(C_PARAM)
{
   return(easy_add(C_PARAM_CALL, "P"));
}

int c_addop(C_PARAM)
{
   return(easy_add(C_PARAM_CALL, "O"));
}

int c_addcoowner(C_PARAM)
{
   return(easy_add(C_PARAM_CALL, "C"));
}

int c_addmanager(C_PARAM)
{
   return(easy_add(C_PARAM_CALL, "M"));
}

int c_addowner(C_PARAM)
{
   return(easy_add(C_PARAM_CALL, "W"));
}

int easy_add(C_PARAM, char *level)
{
  char* nick;
  char* handle;
  char Pass[GENPASSLEN+1];
  char hbuff[MAXHANDLELEN+1];
  HandleListType tmphptr;
  UserListType target;
  int OwnerNumber = 0, CoOwnerNumber = 0, ManagerNumber = 0, OpNumber = 0, PeonNumber = 0, LamerNumber = 0;



  nick = (char *) mysep(&rest, " ");
  handle = (char *) mysep(&rest, " ");
 
  if(handle == NULL)
  {
    handle = hbuff;
    safecopy(handle, nick, MAXHANDLELEN+1);
  }
  if(nick == NULL)
  {
     send_to_user(uptr, "Usage: /msg %s ADD\002<level>\002 <nick> [handle]", IrcNick);
     return(FALSE);
  }
  if(strlen(handle) > MAXHANDLELEN)
  {
     send_to_user(uptr, "Sorry, handles are limited to %d characters.", MAXHANDLELEN);
     return(FALSE);
  }
  target = FindNick(nick, UsersN);
  if(target == NULL)
  {
     send_to_user(uptr, "I don't see anyone on IRC with the nick %s.", nick);
     return(FALSE);
  }
  if(!isjoined(cptr, target) && !IsOper(uptr))
  {
     send_to_user(uptr, "%s is not in %s.  They must join first.", nick, cptr->Name);
     return(FALSE);
  }
  if(FindHandle(cptr, handle))
  {
      send_to_user(uptr, "A %s is already on the list.  They do not need to be readded.", handle);
      send_to_user(uptr, "You can use RESETPASS to change their password for them, or CLVL to change what level they are.");
      return(FALSE);
  }
  GetChanStats(cptr, &OwnerNumber, &CoOwnerNumber, &ManagerNumber, 
      &OpNumber, &PeonNumber, &LamerNumber);
  if(*level == 'W' && OwnerNumber > 0)
  {
      send_to_user(uptr, "Sorry, only one owner per channel is allowed. Delete the old owner first, or add this person as a CoOwner instead.");
      return(FALSE);
  }

  genpass(Pass, GENPASSLEN);
  tmphptr = AddHandle(cptr, handle, crypt(Pass, "$1$92737837$"), "*", *level, 1, 0, 0);
  if(target->Account)
  {
    /* Add a mask of *!*@account.users.afternet.org automagically! */
    char buf[MAXLEN];

    sprintf(buf, "*!*@%s.Users.AfterNET.Org", target->Account);
    AddMask(tmphptr, buf);
  }
  WriteChanOpFile(cptr);
  send_to_user(uptr, "New user \"%s\" added to %s userlist, as %s.", handle, cptr->Name, levelnamefromchar(*level));

  send_privmsg(target->Numnick, "You have been added as %s to my %s userlist with the handle \"%s\" by %s.", 
                    levelnamefromchar(*level), cptr->Name, tmphptr->Handle, uptr->Nick);
  send_privmsg(target->Numnick, "I've generated a password for you: %s", Pass);
  send_privmsg(target->Numnick, "You can log in via: /msg %s %s LOGIN %s %s",
                      IrcNick, cptr->Name, tmphptr->Handle, Pass);
  send_privmsg(target->Numnick, "To change your password: /msg %s %s PASS %s <new password>",
                                 IrcNick, cptr->Name, Pass);
  send_privmsg(target->Numnick, "To learn %s's functions, please see, see /msg %s HELP", IrcNick, IrcNick);
  send_to_user(target, "\002Note:\002 Please check your /msg window (if necessary).  I sent you a new %s password.", cptr->Name);
  send_to_user(uptr, "Sent a generated password via /MSG.");

  return(TRUE);
}


/* These are the NEW userlist add/delete user commands. */
/* I put a TODO here, because I'm sure they can be made more spiffy
 * then the quick bit i'm putting now.
 */
int c_adduser(C_PARAM)
{
  char* handle;
  char* level;
  char levelch;
  char* pass;
  char Pass[GENPASSLEN+1];
  HandleListType tmphptr;
  int OwnerNumber = 0, CoOwnerNumber = 0, ManagerNumber = 0, OpNumber = 0, PeonNumber = 0, LamerNumber = 0;

  if(!(rest && *rest))
  {
      send_to_user(uptr, "\002Usage: ADDUSER [channel] <level> <handle> [password]\002");
/*    send_to_user(uptr, "If \002mask\002 is given it will be added as if addmask was also called.");
 *    (KK) I removed the above line, ass it does not seem to work.
 */
      return(FALSE);
  }
  level = (char *) mysep(&rest, " ");
  handle = (char *) mysep(&rest, " ");
  pass = (char *) mysep(&rest, " ");

  if(handle == NULL)
  {
     send_to_user(uptr, "Improper syntax; handle is required.");
     send_to_user(uptr, "Usage: ADDUSER [channel] <level> \002<handle>\002 [password]");
     return(FALSE);
  }
  if((levelch = levelcharfromname(level)) == '\0' || levelch == 'L')
  {
     send_to_user(uptr, "\002Error:\002 Improper level parameter.  /msg %s LEVELS for a list.", IrcNick);
     return(FALSE);
  }
  if(strlen(handle) > MAXHANDLELEN)
  {
      send_to_user(uptr, "\002Error:\002 Handles cannot be more than %d characters.", MAXHANDLELEN);
      return(FALSE);
  }
  if (!IsGodOper(uptr) && userlevelbychar(levelch) >= userlevelbychar(hptr->Level))
  {
      send_to_user(uptr, "\002Error:\002 You cannot add users with a level greater than or equal to yours.");
      return(FALSE);
  }
  if(FindHandle(cptr, handle))
  {
     send_to_user(uptr, "A user is already in the list with the handle \"%s\". Too add an additional mask see \002/msg %s HELP ADDMASK\002.", handle, IrcNick);
     return(FALSE);
  }
  GetChanStats(cptr, &OwnerNumber, &CoOwnerNumber, &ManagerNumber, 
      &OpNumber, &PeonNumber, &LamerNumber);
  if(levelch == 'W' && OwnerNumber > 0)
  {
       send_to_user(uptr, "Sorry, only one owner is allowed per channel. Either delete the old owner first, or add this person as a CoOwner instead.");
       return(FALSE);
  }
  if(!pass)
    genpass(Pass, GENPASSLEN);
  else
    safecopy(Pass, pass, GENPASSLEN);
  tmphptr = AddHandle(cptr, handle, crypt(Pass, "$1$92737837$"), "*", levelch, 1, 0, 0);
  send_to_user(uptr, "New user \"%s\" added to the %s userlist as %s.", handle, cptr->Name, level);
  send_to_user(uptr, "Their password has been set to %s.", Pass);
  WriteChanOpFile(cptr);
  return(TRUE);
}

int c_deluser(C_PARAM)
{
  char *handle;
  HandleListType handleptr;
  char localhandle[MAXLEN];

  if(!(rest && *rest))
  {
    send_to_user(uptr, "\002Usage:\002 DELUSER [channel] <handle>");
    return(FALSE);
  }

  handle = (char *) mysep(&rest, " ");
  if (!(handleptr = FindHandle(cptr, handle)))
  {
     send_to_user(uptr, "\002Error:\002 Could not find a user with handle \"%s\".", handle);
     return(FALSE);
  }
  if (!IsGodOper(uptr) && userlevelbychar(handleptr->Level) >= userlevelbychar(hptr->Level))
  {
      send_to_user(uptr, "\002Error:\002 You can't delete users with an equal or higher level than yours.");
      return(FALSE);
  }
  safecopy(localhandle, handleptr->Handle, MAXLEN-1);
  DelHandle(cptr, handleptr);
  /* cant use hptr anymore now..  */
  send_to_user(uptr, "User \"%s\" deleted from %s userlist.", localhandle, cptr->Name);
  WriteChanOpFile(cptr);
  return(TRUE);
}

int c_addmask(C_PARAM)
{
  char* Arg0;
  char* Arg1;

  char* handle = NULL;
  HandleListType desthptr;
  char* mask = NULL;
  MaskListType mptr;

  if(!rest || !*rest)
  {
     send_to_user(uptr, "USAGE: ADDMASK [channel] [handle] <mask>");
     return(FALSE);
  }
  Arg0 = (char *) mysep(&rest, " ");
  Arg1 = (char *) mysep(&rest, " ");
  if(Arg1)
  {
     handle = Arg0;
     mask = Arg1;
  }
  else
  {
     mask = Arg0;
  }
  if (!AddyOk(mask)) {
      send_to_user(uptr, "Incorrect mask.");
      return(FALSE);
  }
  if(handle)
  {
      desthptr = FindHandle(cptr, handle);
      if(!desthptr)
      {
           send_to_user(uptr, "\002Error:\002 Could not find handle %s in the userlist.", handle);
           return(FALSE);
      }
      if (!IsGodOper(uptr) && (hptr != desthptr) && ((userlevelbychar(desthptr->Level) >= userlevelbychar(hptr->Level)) ||
          (userlevelbychar(hptr->Level) < MA)))
      {
          send_to_user(uptr, "Sorry, you do not have the access level to do that.");
          return(FALSE);
      }
  }
  else
  {
      desthptr = hptr;
  }
  if(!desthptr)
  {
     send_to_user(uptr, "\002Error:\002 Could not find which handle to ADDMASK under.");
     return(FALSE);
  }
  if (strlen(mask) > MAXMASKLEN) 
  {
	  send_to_user(uptr, "\002Error:\002 Your hostmask is too long, it exceeds %d.", MAXMASKLEN);
	  return(FALSE);
  }
  if (FindMask(desthptr, mask))
  {
     send_to_user(uptr, "\002Error:\002 That mask already exists for user %s.", desthptr->Handle);
     return(FALSE);
  }
  mptr = AddMask(desthptr, mask);
  send_to_user(uptr, "New mask \"%s\" added for user \"%s\" in %s.", mptr->Mask, desthptr->Handle, cptr->Name);
  WriteChanOpFile(cptr);
  return(TRUE);
}

int c_delmask(C_PARAM)
{
  char* handle;
  HandleListType desthptr;
  char* mask;
  if(!(rest && *rest))
  {
     send_to_user(uptr, "\002Usage: DELMASK [channel] [handle] <mask>\002");
     return(FALSE);
  }
  handle = (char *) mysep(&rest, " ");
  mask = (char *) mysep(&rest, " ");

  if(!mask)
  {
     mask = handle;
     /* desthptr set to sender's */
     desthptr = hptr;
     if(!desthptr)
     {
        send_to_user(uptr, "\002Error:\002 You don't have a handle to delete the mask from.");
        return(FALSE);
     }
  }
  else
  {
     desthptr = FindHandle(cptr, handle);
     if(desthptr == NULL)
     {
        send_to_user(uptr, "\002Error:\002 Can't find the handle \"%s\" in %s userlist.", handle, cptr->Name);
        return(FALSE);
     }
     if (!IsGodOper(uptr) && hptr != desthptr && ((userlevelbychar(desthptr->Level) >= userlevelbychar(hptr->Level)) ||
         (userlevelbychar(hptr->Level) < MA))) 
     {
         send_to_user(uptr, "Sorry, you do not have access to do that.");
         return(FALSE);
     }
  }
  if(DelMask(desthptr, mask))
  {
     send_to_user(uptr, "Successfully removed mask \"%s\" from %s's account in %s.", mask, desthptr->Handle, cptr->Name);
     WriteChanOpFile(cptr);
     return(TRUE);
  }
  send_to_user(uptr, "Error: There is no \"%s\" listed under %s's account in %s.", mask, desthptr->Handle, cptr->Name);
  return(FALSE);
}

int c_llist(C_PARAM)
{
  LamerListType lptr;

  send_to_user(uptr, "=------ %s Lamer list -----=", cptr->Name);
  for(lptr = cptr->Lamers;lptr;lptr = lptr->Next)
  {
      send_to_user(uptr, "%s \002-\002 %s", lptr->Mask, lptr->Reason);
  }
  send_to_user(uptr, "=-- End of %s lamer list --=", cptr->Name);
  return(TRUE);
}

int c_plist(C_PARAM)
{
  ShowUserlist(C_PARAM_CALL, "P");
  return(TRUE);
}

int c_olist(C_PARAM)
{
  ShowUserlist(C_PARAM_CALL, "O");
  return(TRUE);
}

int c_mlist(C_PARAM)
{
  ShowUserlist(C_PARAM_CALL, "M");
  return(TRUE);
}

int c_wlist(C_PARAM)
{
  ShowUserlist(C_PARAM_CALL, "W");
  return(TRUE);
}

int c_clist(C_PARAM)
{
  ShowUserlist(C_PARAM_CALL, "C");
  return(TRUE);
}

int ShowUserlist(C_PARAM, char* filter)
{
   int count, days, hours;
   MaskListType mptr;
   char SeenStr[MAXLEN] = "";
   char tmp[MAXLEN];
   time_t seentime, difference;
   HandleListType savehptr = hptr;

   char* searchstr;

   if(rest && *rest)
   {
      searchstr = (char *) mysep(&rest, " ");
   }
   else
   {
      searchstr = NULL;
   }
   send_to_user(uptr, "=-Level---Aop----Handle--+---Trusted Mask(s)-------=");
   for(hptr = cptr->Handles; hptr; hptr = hptr->Next)
   {
        if(filter && *filter && *filter != hptr->Level)
           continue; /* filter didnt match */
        if(searchstr && matches(searchstr, hptr->Handle))
           continue; /* handle didnt match */
/**************************/
        strcpy(SeenStr, "");
        seentime = hptr->LastSeen;
        difference = time(NULL) - seentime;
        days = difference / 86400;
        difference = difference - (days * 86400);
        hours = difference / 3600;
        if(days > 0)
        {
           sprintf(tmp, "\002%d\002d", days);
           strcat(SeenStr, tmp);
        }
        else if(hours > 0)
        {
           sprintf(tmp, "\002%d\002h", hours);
           strcat(SeenStr, tmp);
        }
        if(!*SeenStr)
        {
           if(IsInChan(hptr))
              strcat(SeenStr, "In Chan");
           else
              strcat(SeenStr, "Just left");
        }
	if (!seentime) {
	   strcpy(SeenStr, "n/a");
	}
/**************************/
	/* Users who are not logged in don't need to see the masks. 
	   IRCOps might want to help users, and they won't abuse the masks (i hope (: ) */
        if(hptr->Masks && (savehptr || IsOper(uptr)))
        {
           send_to_user(uptr, "\002%7s\002 - [\002%s\002] %10s\002%c-\002 %s (%s)", 
               levelnamefromchar(hptr->Level), hptr->AutoOp?"@":" ", hptr->Handle, 
               hptr->Masks->Next?'+':'-', hptr->Masks->Mask, SeenStr);
           for(mptr = hptr->Masks->Next;mptr;mptr = mptr->Next)
           {
               send_to_user(uptr, "                        \002%c-\002 %s", 
                            mptr->Next?'|':'`', mptr->Mask);
               count++;
           }
        }
        else
        {
           send_to_user(uptr, "\002%7s\002 - [\002%s\002] %10s (%s)", 
               levelnamefromchar(hptr->Level), hptr->AutoOp?"@":" ", hptr->Handle, SeenStr);
        }
   }
   send_to_user(uptr, "=---End of Userlist------+-------------------------=");
   return(TRUE);
}

int c_userlist(C_PARAM)
{
    return(ShowUserlist(C_PARAM_CALL, ""));
}

int countregchans()
{
  int i,c=0;
  ChannelListType cptr;
  
  for(i=0;i<CHANSEGSIZE;i++)
  {
     cptr = Channels->Table[i];
     while(cptr)
     {
        if(cptr->Regged)
          c++;
        cptr = cptr->Next;
     }
  }
  return(c);
}

int c_netstat(C_PARAM)
{
  float ave = 0;

  ave = (float) OperCount / (float) ServerCount;
  send_to_user(uptr, "---------------------------------");
  send_to_user(uptr, "      N E T    S T A T S         ");
  send_to_user(uptr, "---------------------------------");
  send_to_user(uptr, " Number of Users:   %d", UserCount);
  send_to_user(uptr, " Number of Servers: %d", ServerCount);
  send_to_user(uptr, " Number of IRC Ops: %d", OperCount);
  send_to_user(uptr, " There are %d channels.", ChannelCount);
  send_to_user(uptr, " Avg Number of Opers per Server currently online: %1.1f", ave);
  send_to_user(uptr, " ");
  send_to_user(uptr, " Total number of registered channels: %d", countregchans());
  send_to_user(uptr, " Total number of Peons : %d ", PeonCount);
  send_to_user(uptr, " Total number of Ops : %d ", OpCount);
  send_to_user(uptr, " Total number of Managers : %d ", ManagerCount);
  send_to_user(uptr, " Total number of Co-Owners : %d ", CoOwnerCount);
  send_to_user(uptr, " Total number of Owners : %d ", OwnerCount);
  if (ChannelCount > 0)
    send_to_user(uptr, " Avg Number of Peons+Opers+Managers+(Co)Owners per channel: %d", ((PeonCount + ManagerCount + OpCount + OwnerCount) / ChannelCount));
  send_to_user(uptr, " End of stats");
  return(TRUE);
}

int c_uptime(C_PARAM)
{
  time_t UpTime, ConTime;
  char buf[MAXLEN];
  extern int NumOfGateSocks;

  UpTime = time(NULL) - StartTime;
  ConTime = time(NULL) - ConnectTime;
  send_to_user(uptr, "\002Uptime    %d day(s) (%02d:%02d)\002", UpTime / 86400, (UpTime / 3600) % 24, (UpTime / 60) % 60);
  send_to_user(uptr, "\002Connected %d day(s) (%02d:%02d)\002", ConTime / 86400, (ConTime / 3600) % 24,(ConTime / 60) % 60);
  send_to_user(uptr, "\002Local clock    : %s", t2s(buf, time(NULL)));
  send_to_user(uptr, "\002Network clock  : %s", t2s(buf, NetTime()));
  send_to_user(uptr, "\002%d commands parsed from %d lines processed.\002", TotalCommands, TotalLines);
  send_to_user(uptr, "\002%d users signed on.\002", TotalUserCount);
  if(GATECHECK)
     send_to_user(uptr, "\002There are %d socket(s) in use checking for Wingates.\002", NumOfGateSocks);
   send_to_user(uptr, "\002NETBURST mode is %s", NETBURSTING?"ON":"OFF");
  return(TRUE);
}

int c_settings(C_PARAM)
{
  ShowXSet(uptr);
#ifdef NOTDEF
  UpTime = time(NULL) - StartTime;
  send_to_user(uptr, " ");
  send_to_user(uptr, " X2 Uptime: Day %d (%02d:%02d)", UpTime / 86400, (UpTime / 3600) % 24, (UpTime / 60) % 60);
  send_to_user(uptr, " ");
  send_to_user(uptr, "\002 X2 NET SETTINGS:");
  send_to_user(uptr, " --------------------");
  send_to_user(uptr, "\002  PINGOUTCONNECT\002: %s", GetSetting(PINGOUTCONNECT));
  send_to_user(uptr, "\002  AUTOROUTING   \002: %s", GetSetting(AUTOROUTING));
  send_to_user(uptr, "\002  AUTOCONNECT   \002: %s", GetSetting(AUTOCONNECT));
  send_to_user(uptr, "\002  GATECHECK     \002: %s", GetWinGateMet(GATECHECK));
  send_to_user(uptr, "\002  MAXCLONES     \002: %d", MAXCLONES);
  send_to_user(uptr, "\002  CLONERES      \002: %s", GetCloneResMet(CLONERES));
  send_to_user(uptr, "\002  SUB7SCAN      \002: %d",  GetCloneResMet(SUB7SCAN));
  send_to_user(uptr, "\002  LITMUS        \002: %d", GetCloneResMet(LITMUS));
  send_to_user(uptr, "\002  SDBOT         \002: %d", GetCloneResMet(SDBOT));
  send_to_user(uptr, "\002  DEBUG         \002: %d", DEBUG);
  if (*DBTARGET)
    send_to_user(uptr, "\002  DBTARGET      \002: %s", DBTARGET);
  else
    send_to_user(uptr, "\002  DBTARGET      \002: NULL");
  send_to_user(uptr, " End of settings");
#endif
  return(TRUE);
}

/* TODO : Move any non c_ functions out of this file.. */
int NumDownLinks(char *Name)
{
  ServerListType ServPtr;
  int i, num = 0;

  for (i = 0; i < SERVERSEGSIZE; i++)
  {
    ServPtr = Servers->Table[i];
    while (ServPtr)
    {
      if (!(strcasecmp(ServPtr->LinkName, Name)))
	num++;
      ServPtr = ServPtr->NextA;
    }
  }
  return (num);
}

int ShowDownLinks(ChannelListType cptr, UserListType uptr, char *Name, char *prevpre, char *arrowchar)
{
  ServerListType ServPtr;
  int i, j;
  char pre[MAXLEN];
  char *nextpre;
  int num = 0;
  static int depth = 0;

  nextpre = malloc(MAXLEN);
  safecopy(pre, prevpre, MAXLEN);

  if((ServPtr = GetServer(Name)))
  {
    if (strcmp((ServPtr->Name) + strlen(ServPtr->Name) - strlen(".Services"), ".Services"))
    {
      if (ServPtr->NumOfUsers != 1)
	send_to_user(uptr, "%s%s-\002%s\002 [%d users]", pre, arrowchar, Name, ServPtr->NumOfUsers);
      else
	send_to_user(uptr, "%s%s-\002%s\002 [1 user]", pre, arrowchar, Name);
    }
    else
      send_to_user(uptr, "%s%s-\002%s\002 (service)", pre, arrowchar, Name);
  }
  else				/* its me */
    send_to_user(uptr, "\002%s\002 (me)", ServiceName);
  j = NumDownLinks(Name);
  for (i = 0; i < SERVERSEGSIZE; i++)
  {
    ServPtr = Servers->Table[i];
    while (ServPtr)
    {
      if (!(strcasecmp(ServPtr->LinkName, Name)))
      {
	safecopy(nextpre, pre, MAXLEN);
	if (depth++ > 0)
	{
	  if (arrowchar[0] == '`')
	    strcat(nextpre, "  ");
	  else
	    strcat(nextpre, "| ");
	}
	if (j > ++num)
	{
	  ShowDownLinks(cptr, uptr, ServPtr->Name, nextpre, "|");
	}
	else
	{
	  ShowDownLinks(cptr, uptr, ServPtr->Name, nextpre, "`");
	}
      }
      ServPtr = ServPtr->NextA;
    }
  }
  free(nextpre);
  return(FALSE);
}

int c_addr(C_PARAM)
{
  char* server;
  char* portstr;
  char* uplink;
  char* secondary;
  int port = 0;
  server = strtok(rest, " ");
  portstr = strtok(NULL, " ");
  uplink = strtok(NULL, " ");
  secondary = strtok(NULL, " ");
  if(!(server && portstr && port))
  {
     send_to_user(uptr, "\002Usage: ADDR <server> <port> [uplink] [secondary uplink]\002");
     return(FALSE);
  }
  port = atoi(portstr);
  AddRouteServer(Route, server, port, uplink, secondary);
  send_to_user(uptr, "Server added to routing table.");
  return(TRUE);
}

int c_reroute(C_PARAM)
{
    if(!rest || !*rest)
    {
       send_to_user(uptr, "\002Usage: REROUTE <N | C | T>\002");
       send_to_user(uptr, "  N = <\002N\002>ormal, C = <\002C\002>onnect, or T = <\002T\002>est.");
       return(FALSE);
    }
    if(Reroute(Route, uptr, toupper(rest[0])))
       return(TRUE);
     return(FALSE);
}

int c_rload(C_PARAM)
{
  if(LoadRouteFile(Route, "routing.dat"))
  {
     send_to_user(uptr, "Routing data file loaded.");
     return(TRUE);
  }
  send_to_user(uptr, "\002Error:\002 Failed to open route file.");
  return(FALSE);
}

int c_rlist(C_PARAM)
{
  ShowRouteList(Route, uptr);
  return(TRUE);
}

int c_rmap(C_PARAM)
{
  ShowRouteMap(Route, uptr);
  return(TRUE);
}

/* Uses the above 2 functions to draw the map. Should
 * move them to more appropreate locations.. 
 */
int c_map(C_PARAM)
{
  ShowDownLinks(cptr, uptr, ServiceName, "", "`");
  send_to_user(uptr, "End of MAP");
  return(TRUE);
}

int c_servers(C_PARAM)
{
  int i;
  ServerListType ServPtr;

  send_to_user(uptr, " ------------------------------");
  send_to_user(uptr, " Servers and their user counts:");
  send_to_user(uptr, " ------------------------------");
  for (i = 0; i < SERVERSEGSIZE; i++)
  {
    ServPtr = Servers->Table[i];
    while (ServPtr)
    {
      if((strcmp((ServPtr->Name) + strlen(ServPtr->Name) - strlen(".Services"), ".Services"))
	 || (strstr(rest, "DEBUG") )) /* filter out Services */
	send_to_user(uptr, "\002 %s\002(%d) has %d user(s) <-> \002%s\002", ServPtr->Name, 
			base64toint(ServPtr->numnick), ServPtr->NumOfUsers, ServPtr->LinkName);
      ServPtr = ServPtr->NextA;
    }
  }
  send_to_user(uptr, " =--------------END--------------=");
  return(TRUE);
}

int c_expirechans(C_PARAM)
{
   send_to_user(uptr, "Removing unvisited channels...");
   ExpireChannels(uptr);
   return(TRUE);
}

int c_unreg(C_PARAM)
{
  FILE *ChanFile;
  FILE *NewChanFile;
  char *UpCaseChan;
  char str[LINEBUFFSIZE];
  char ChanName[LINEBUFFSIZE];
  int s, line = 0;
  HandleListType tmphptr, nexthptr;
  LamerListType lptr, nextlptr;

  ChanFile = fopen(CHANFILENAME, "r");
  if (ChanFile == NULL)
  {
    send_to_user(uptr, "\002Error:\002 Op file doesn't not exist!");
    return(TRUE);
  }
  NewChanFile = fopen("tmp/.tmpchanfile.dat", "w");
  if(!NewChanFile)
  {
     Debug(DBGSEVERE, "Error opening temp file in unreg()!!");
     return(FALSE);
  }
  while ((s = fscanf(ChanFile, "%s\n", ChanName)) != EOF)
  {
    line++;
    if (s != 1)
    {
      Debug(DBGERROR, "Error with param count in chan.dat! s=%d line=%d", s, line);
      send_to_user(uptr, "\002Error:\002 Param count in chan.dat! (s=%d line=%d)\002", s, line);
      fclose(ChanFile);
      fclose(NewChanFile);
      return(FALSE);
    }
    if (strcasecmp(ChanName, cptr->Name))
      fprintf(NewChanFile, "%s\n", ChanName);
  }
  if(ferror(NewChanFile))
  {  
     Debug(DBGSEVERE, "\002Error:\002 Failure writing .tmpchanfile.dat (out of disk space?)!  Unregistration NOT saved.", cptr->Name);
     fclose(NewChanFile);
     sleep(4);
     exit(2);
  }
  fclose(ChanFile);
  fclose(NewChanFile);
  sprintf(str, "./%s", CHANFILENAME);
  unlink(str);
  rename("tmp/.tmpchanfile.dat", str);
  sprintf(str, "./CHANS/%s", UpCaseChan = UpCase(&cptr->Name[1]));
  unlink(str);
  free(UpCaseChan);
  if (cptr->settings[INCHAN] != 0)
    send_part(cptr, "Unregistered by %s.", uptr->Nick);
  send_wallops("Channel %s unregistered by %s.", cptr->Name, uptr->Nick);
/** If we were the only thing in the channel, delete it. **/
  if(cptr->NumOfNicks == 0)
    DelChan(cptr->Name);
  else
  {  /* Convert it back to an unreg'd channel.. */

     /* ModeString */
     if(cptr->Key)
     {
        free(cptr->Key);
        cptr->Key = NULL;
     }
     if(cptr->Limit)
     {
        free(cptr->Limit);
        cptr->Limit = NULL;
     }
     /* Topic */
     free(cptr->Topic);
     cptr->Topic = NULL;
     /* Email */
     free(cptr->Email);
     cptr->Email = NULL;
     /* Desc */
     free(cptr->Description);
     cptr->Description = NULL;
     /* URL... */
     free(cptr->URL);
     cptr->URL = NULL;
     /* ReggedBy... */
     free(cptr->Reggedby);
     cptr->Reggedby = NULL;
     /* Suspended... */
     if (cptr->Suspended)
     {
         free(cptr->Suspended);
         cptr->Suspended = NULL;
     }

/* Delete userlist, and lamerlist.. */
     for(tmphptr = cptr->Handles;tmphptr;tmphptr = nexthptr)
     {
        nexthptr = tmphptr->Next;
        DelHandle(cptr, tmphptr);
     }
     cptr->Handles = NULL;
     for(lptr = cptr->Lamers;lptr;lptr = nextlptr)
     {
        nextlptr = lptr->Next;
        free(lptr);
     }
     cptr->Lamers = NULL;

     /* Regged...*/
     cptr->Regged = FALSE;
    /* The channel is now Unregistered. */
  }
  return(TRUE);
}

int c_savechan(C_PARAM)
{
  WriteChanOpFile(cptr);
  send_to_user(uptr, "\002Wrote %s user data to disk.\002", cptr->Name);
  return(TRUE);
}

int c_loadchan(C_PARAM)
{
  FILE *OpFile;
  char linebuff[LINEBUFFSIZE];
  char filename[LINEBUFFSIZE], *sepptr;
  char *ptrbuff, *key, *lb;
  char FullChanName[LINEBUFFSIZE];
  int i,s,line2=0,fileversion=0;
  int UserAutoOp, UserPwdReq;
  int AutoOp;
  HandleListType LastHandle = NULL;
  char *Mask, *Reason;
  char Level;
  char *OpAddress, *Pass, *Info, *Handle;
  ChannelListType ChanPtr = NULL;
  time_t LastSeen, BanTime;
  int numofadds = 0;

  if(!rest || !*rest)
  {
     send_to_user(uptr, "\002Usage:\002 LOADCHAN <absolute path>");
     send_to_user(uptr, "Path must point to a channel data file with, at the very least, CHANNEL, VERSION, and USER lines.");
     send_to_user(uptr, "This command loads a file from the local file system into memory.");
     send_to_user(uptr, "It is for the X2 admin to quickly restore a single channel from backup files.");
     send_to_user(uptr, "If anyone uses this command who is not directly related to the account X2 runs from, you are breaking the law!");
     send_to_user(uptr, "Unauthorized access of a computer system in Oregon carries very real penalties.  HANDS OFF!");
     return(0);
  }
  sepptr = strtok(rest, " ");
  if(!sepptr)
     return(0);
  strcpy(filename, "./");
  strcat(filename, sepptr);
  if((OpFile = fopen(filename, "r")))
  {
     line2 = 0;
/* File reading stuff from main.c */
/**********************************/
    while (!feof(OpFile))
    {
      i++;
      s = fscanf(OpFile, "%[^\n]\n", linebuff);
      line2++;
      if (s == EOF)
        break;
      if (s != 1)
      {
         send_to_user(uptr,"\002Error:\002 Corruption in %s, line %d.", filename, line2);
         return(FALSE);
      }
      lb = linebuff;
      key = (char *) mysep(&lb, " ");
      if(!key || !*key)
      {
         send_to_user(uptr, "Can't really happen...\n");
         return(FALSE);
      }
      switch(key[0])
      {
         case 'V': /* VERSION   current version = 4*/
                ptrbuff = (char *) mysep(&lb, " ");
                if(!ptrbuff)
                {
                    send_to_user(uptr, "Invalid VERSION syntax in %s, line %d.", filename, line2);
                    return(0);
                }
                fileversion = strtol(ptrbuff, NULL, 10);
                if(fileversion > 4)
                {
                   send_to_user(uptr,"Warning!  Reading %s which is file version %d.", filename, fileversion);
                }
                continue;
         case 'C': /* Channel */
                ptrbuff = (char *) mysep(&lb, "\n");
		safecopy(FullChanName, ptrbuff, MAXLEN);
                ChanPtr = GetChannel(FullChanName);
                if(!ChanPtr)
                {
                   send_to_user(uptr, "You must pre-register %s first.", FullChanName);
                   return(FALSE);
                }
                continue;
         case 'T': /* Topic */
         case 'E': /* Email */
         case 'M': /* ModeString */
         case 'D': /* Description */
         case 'S': /* Settings */
         case 'L': /* LastVisted */
         case 'R': /* Regged by */
         case 'W': /* Website (URL) */
         case 'P': /* Suspended */
		continue;
         case 'H': /* HANDLE */
                if(ChanPtr == NULL)
                {
                     send_to_user(uptr, "HANDLE before channel fully defined (%s line %d).", filename, line2);
                     return(FALSE);
                }
                Handle = (char *) mysep(&lb, " ");
                ptrbuff = (char *) mysep(&lb, " ");
                Level = ptrbuff[0];
                ptrbuff = (char *) mysep(&lb, " ");
                AutoOp = strtol(ptrbuff, NULL, 10);
                ptrbuff = (char *) mysep(&lb, " ");
                LastSeen = strtol(ptrbuff, NULL, 10);
                Pass = (char *) mysep(&lb, " ");
                Info = (char *) mysep(&lb, "\n");
                /* ******************************************** */
                if(!Info)
                {
                    send_to_user(uptr, "Invalid HANDLE syntax in %s, line %d\n", filename, line2);
                    continue;
                }
                if(strlen(Pass) > MAXPASS)
                {
                    send_to_user(uptr,"Error in %s, line %d. Password too long.\n", filename, line2);
                    continue;
                }
                if(strlen(Handle) > 20)
                {
                    send_to_user(uptr,"Error in %s, line %d. Handle too long.\n", filename, line2);
                    continue;
                }
                if(strlen(Info) > 200)
                {
                    send_to_user(uptr,"Error in %s, line %d. Infoline too long.\n", filename, line2);
                    continue;
                }
                if(fileversion < 4)
                {
                     /* Convert to crypted. */
                     Pass = crypt(Pass, "$1$92737837$");
                }
                if(FindHandle(ChanPtr, Handle))
                {
                   LastHandle = FindHandle(ChanPtr, Handle);
                   continue; /* This one's already listed with us */
                }
                LastHandle = AddHandle(ChanPtr, Handle, Pass, Info, Level, AutoOp, LastSeen, 0);
                numofadds++;
                continue;
         case 'K': /* MASK */
                if(LastHandle == NULL)
                {
                    send_to_user(uptr,"Error in %s, line %d.  MASK line does not follow HANDLE.\n", filename, line2);
                    continue;
                }
                Mask = (char *) mysep(&lb, " ");
                if(FindMask(LastHandle, Mask))
                {
                      continue; /* duplicate */
                }
                AddMask(LastHandle, Mask);
                continue;
         case 'B': /* BAN (lamer lits) */
                
                ptrbuff = (char *) mysep(&lb, " ");
                BanTime = strtol(ptrbuff, NULL, 10);
                ptrbuff = (char *) mysep(&lb, " ");
                LastSeen = strtol(ptrbuff, NULL, 10);
                Mask = (char *) mysep(&lb, " ");
                Reason = (char *) mysep(&lb, "\n\0");
                AddLamer(ChanPtr, BanTime, LastSeen, Mask, Reason);
                continue;
         case 'U': /* USER */  /* Now defunct. convert to handle/mask only */
                if(ChanPtr == NULL)
                {
                    send_to_user(uptr,"USER before channel fully defined (%s line %d).", filename, line2);
                    return(FALSE);
                }
                OpAddress = (char *)mysep(&lb, " ");
                ptrbuff = (char *)mysep(&lb, " ");
                UserAutoOp = strtol(ptrbuff, NULL, 10);
                if(fileversion >= 2)
                {
                  ptrbuff = (char *)mysep(&lb, " ");
                  UserPwdReq = strtol(ptrbuff, NULL, 10);
                }
                else  /* Hack to be backward compatable w version 1 data file */
                {
                  UserPwdReq = 1;
                }
                ptrbuff = (char *)mysep(&lb, " ");
                LastSeen = strtol(ptrbuff, NULL, 10);
                Handle = (char *)mysep(&lb, " ");
                Pass = (char *)mysep(&lb, " ");
                Info = (char *)mysep(&lb, "\n");
                /* ******************************************** */
		if(!Info)
                {
                    send_to_user(uptr,"Invalid USER syntax in %s, line %d.", filename, line2);
                    continue;
                }
                if(strlen(Pass) > MAXPASS)
                {
                    send_to_user(uptr,"Error in %s, line %d.  Password too long.", filename, line2);
                    continue;
                }
                Pass = crypt(Pass, "$1$92737837$");
                if(strlen(Handle) > 20)
                {
                    send_to_user(uptr,"Error in %s, line %d.  Handle too long.\n", filename, line2);
                    continue;
                }
                if(strlen(Info) > 200)
                {
                    send_to_user(uptr,"Error in %s, line %d. Infoline too long.\n", filename, line2);
                    continue;
                }    
                if(OpAddress[0] == 'L')
                {
                  /* Add lamer list entry and return */
                  AddLamer(ChanPtr, time(NULL), LastSeen, &OpAddress[1], Info);
                  continue;
                }
                /* A hack to enforce the new Handle length limit set in config.h */
                if(strlen(Handle) > MAXHANDLELEN)
                       Handle[MAXHANDLELEN] = '\0';
                if(!(hptr = FindHandle(ChanPtr, Handle)))
                {
                      hptr = AddHandle(ChanPtr, Handle, Pass, Info, OpAddress[0], UserAutoOp, LastSeen, 0);
                      numofadds++;
                }
                if(!strcmp(crypt("*", "X2"), hptr->Pass))  /* If they didnt have a pass set, add all there masks.. */
                   AddMask(hptr, &OpAddress[1]);
		continue;
         default: send_to_user(uptr,"Unknown key %s in %s.", key, filename);
                  continue;
      }
     }
     WriteChanOpFile(ChanPtr);
     send_to_user(uptr, "%s loaded as %s.  %d users added.", filename, FullChanName, numofadds);
     return(TRUE);
  }
  else
  {
    send_to_user(uptr, "%s not found.", filename);
    return(FALSE);
  }
  send_to_user(uptr, "\002Error:\002 SHOULD NEVER SEE.");
  return(FALSE);
}

int c_scriptreg(C_PARAM)
{
   /* This is a front-end to regchan, topic, addowner etc */
   /* So a script can be more easily made to register channels */

   /* Syntax is: */
   /* SCRIPTREG <#channel> <OwnerHandle> <Password> <user@email> +<modes> [:topic] */
   char *ChanName, *OwnerHandle, *OwnerPassword, *Modes, *Topic, *tmpstr, *OwnerEmail;
   char buff[MAXLEN];

   ChanName = chanstr;
   OwnerHandle = (char *) strtok(rest, " ");
   OwnerPassword = (char *) strtok(NULL, " ");
   OwnerEmail = (char *) strtok(NULL, " ");
   Modes = (char *) strtok(NULL, ":");
   Topic = (char *) strtok(NULL, "\0");

   if(!Modes || !*Modes)
   {
     send_to_user(uptr, "\002Error:\002 Syntax error.\002");
     return(FALSE);
   }
   sprintf(buff, "%s", OwnerEmail);
   send_to_user(uptr, "\002=------- Beginning Register -------=\002");
   tmpstr = rest;
   rest = buff;
   /* Call RegisterChan.. C_PARAM_CALL is defined with C_PARAM, but since
    * it uses the variable "rest", we have to make rest the right variable for it. (above)
    */
   if(c_regchan(C_PARAM_CALL))
   {
      char Str[MAXLEN];
      ChannelListType cptr;
      HandleListType LastHandle = NULL;

      if(!(cptr = GetChannel(ChanName)))
      {
         send_to_user(uptr, "REGCHAN returned OK, however, I cant find %s in memory...", ChanName);
         send_to_user(uptr, "\026FAILED\026");
         return TRUE;
      }

      /* Add owner */
      if((LastHandle = AddHandle(cptr, OwnerHandle, crypt(OwnerPassword, "$1$92737837$"), "*", 'W', 1, 0, 0)))
      {
	 char buf[MAXLEN];

	 sprintf(buf, "*!*@%s.users.AfterNET.Org", LastHandle->Handle);
	 AddMask(LastHandle, buf);
         strcpy(Str, "Owner: \002SUCCESS\002");
      }
      else
      {
         strcpy(Str, "Owner: \026FAILED\026 ");
      }

      /* Set modes */
      ConvertModeString(Modes, &cptr->PlusModes, &cptr->MinModes, &cptr->Key, &cptr->Limit);
      send_chanmode(cptr, "%s", GetModeString(buff, cptr->PlusModes, cptr->MinModes,
			  cptr->Key, cptr->Limit));
      if (cptr->Key)
	safecopy(cptr->CurrentKey, cptr->Key, MAXCHANKEY);
      else
	strcpy(cptr->CurrentKey, "");

      /* Set topic */
      if(Topic && *Topic)
      {
        sprintf(buff, "TOPIC %s", Topic);
        rest = buff;

        free(cptr->Topic);
        if(strlen(Topic) > MAXTOPICLEN)
            Topic[MAXTOPICLEN] = '\0';
        if((cptr->Topic = (char *) malloc(strlen(Topic) + 1)))
	{
	    strcpy(cptr->Topic, Topic);
     	    send_to_user(uptr, "\002Default topic set to\002: %s", cptr->Topic);
	}
        else
	{
	    send_to_user(uptr, "\002Error allocating memory for new topic!");
   	    return(FALSE);
	}
      }
      WriteChanOpFile(cptr);
      send_to_user(uptr, "\002=-- \002%s\002 --=",Str);
      return(TRUE);
   }
   else
   {
     send_to_user(uptr, "\002=------\002 \026Registration Aborted\026 \002------=\002");
     return(FALSE);
   }
   return(TRUE);
}

int c_regchan(C_PARAM)
{
  FILE* ChanFile;
  char str[LINEBUFFSIZE + 6];
  ChannelListType ChanPtr;
  char* UpCaseChan;
  char topic[MAXLEN];
  char* chan;
  char* owneremail;
  int i;

  if(!rest || !*rest || !chanstr)
  {
    send_to_user(uptr, "\002Usage:\002 REGCHAN <channel> <owners@email.address>");
    return(FALSE);
  }
  chan = chanstr;
  owneremail = (char*) mysep(&rest, " ");
  if (!ChanOk(chan))
  {
    send_to_user(uptr, "\002Error:\002 Illegal channel name `%s'.\002", chan);
    return(FALSE);
  }
  if (owneremail == NULL)
  {
    send_to_user(uptr, "\002Error:\002 No e-mail address specified for `%s'!", chan);
    return(FALSE);
  }
  if (IsIllegalChannel(chan))
  {
    send_to_user(uptr, "\002That channel name is not acceptable.\002");
    return(FALSE);
  }
  ChanPtr = GetChannel(chan);
  if (ChanPtr && ChanPtr->Regged)
  {
    send_to_user(uptr, "\002Channel %s is already in memory!\002", chan);
    return(FALSE);
  }
  if(!ChanPtr)  /* A new channel, not currently in memory */
  {
    if(!AddChan(chan, FALSE, 0, NULL, NULL, NULL, NULL, NULL,NULL, NetTime(), 0))
    {
       send_to_user(uptr, "\002Unexpected error encountered while adding channel to memory.\002");
       return(FALSE);
    }
    if(!(ChanPtr = GetChannel(chan)))
    {
       send_to_user(uptr, "Could not find channel in memory; however, I have added it.");
       return(FALSE);
    }
  }
  strcpy(topic, "*");
/**** Here we convert the unreg'd chanptr into a regged chanptr. ***/
  /* Put the +nt into the chanptr */
  ConvertModeString("+nt", &ChanPtr->PlusModes, &ChanPtr->MinModes, &ChanPtr->Key, &ChanPtr->Limit);
  /* Topic.. */
  ChanPtr->Topic = (char *) malloc(2); /* ('*', '\0') */
  strcpy(ChanPtr->Topic, "*");
  ChanPtr->Email = (char *) malloc(strlen(owneremail)+1);
  strcpy(ChanPtr->Email, owneremail);
  /* Description.. */
  ChanPtr->Description = (char *) malloc(2);
  strcpy(ChanPtr->Description, "*");
  /* URL.. */
  ChanPtr->URL = (char *) malloc(2);
  strcpy(ChanPtr->URL, "*");
  /* ReggedBy */
  ChanPtr->Reggedby = (char *) malloc(strlen(uptr->Nick)+1);
  strcpy(ChanPtr->Reggedby, uptr->Nick);
  /* settings */
  for(i=0;i<NUMOFCHANSETTINGS;i++)
     ChanPtr->settings[i] = set_commands[i].def;
  /* CreationDate */
  ChanPtr->CreationDate = NetTime();
  /* LastAccess */
  ChanPtr->LastAccess = time(NULL);
  /* Regged status.. */
  ChanPtr->Regged = TRUE;
/**** Its now a regged channel ***/

  if (!(ChanFile = fopen(CHANFILENAME, "r")))
  {
    send_to_user(uptr, "\002Could not find %s; creating new file...\002", CHANFILENAME);
    ChanFile = fopen(CHANFILENAME, "w");
    fclose(ChanFile);
  }
  fclose(ChanFile);
  send_to_user(uptr, "\002Successfully registered %s.\002", ChanPtr->Name);
  ChanFile = fopen(CHANFILENAME, "a");
  fprintf(ChanFile, "%s\n", ChanPtr->Name);
  if(ferror(ChanFile))
  {  
     Debug(DBGERROR, "\002Error: Writing %s (out of disk space?).  %s data NOT saved.", CHANFILENAME, chan);
     fclose(ChanFile);
     return(FALSE);
  }
  fclose(ChanFile);
  UpCaseChan = UpCase(chan);
  sprintf(str, "./CHANS/%s", &UpCaseChan[1]);
  free(UpCaseChan);
  WriteChanOpFile(ChanPtr);
  send_join(ChanPtr);
  send_opchan(ChanPtr);
  send_chanmode(ChanPtr, "%s","+nt") ;
  return(TRUE);
}

int c_clvl(C_PARAM)
{
   char *handle;
   char *level;
   int numlevel;
   HandleListType Handle;
   
   handle = strtok(rest, " ");
   level = strtok(NULL, " ");

   if (!(level && handle))
   {
       send_to_user(uptr, "\002Usage:\002 CLVL <handle> <new level>");
       return(FALSE);
   }

   if(!strcasecmp(level, "peon"))
     numlevel = PE;
   else if (!strcasecmp(level, "op"))
     numlevel = OP;
   else if (!strcasecmp(level, "manager"))
     numlevel = MA;
   else if (!strcasecmp(level, "coowner"))
     numlevel = CO;
   else if (!strcasecmp(level, "co"))
     numlevel = CO; 
   else if (!strcasecmp(level, "owner"))
     numlevel = OW;
   else
   {
       send_to_user(uptr, "\002Error:\002 Incorrect level parameter.");
       send_to_user(uptr, "Usage: CLVL <handle> \002<new level>\002");
       return(FALSE);
   }
   if(!IsGodOper(uptr) && numlevel >= userlevel(hptr))
   {
       send_to_user(uptr, "You may only change users to levels lower than your own.");
       return(FALSE);
   }

   if (!(Handle = FindHandle(cptr, handle)))
   {
     send_to_user(uptr, "No user with handle %s found in %s userlist.", handle, cptr->Name);
     return(FALSE);
   }

   if(!IsGodOper(uptr) && userlevel(hptr) <= userlevelbychar(Handle->Level))
   {
       send_to_user(uptr, "\002Error for %s:\002 Cannot modify a user whose access is greater than or equal to your own.", 
                    Handle->Handle);   
       return(FALSE);
   }

   send_to_user(uptr, "Changing %s to %s in %s.", 
                Handle->Handle, levelnamefromint(numlevel), cptr->Name);
   switch(Handle->Level)
   {
       case 'P': PeonCount--; break;
       case 'O': OpCount--; break;
       case 'M': ManagerCount--; break;
       case 'C': CoOwnerCount--; break;
       case 'W': OwnerCount--; break;
       default: 
          Debug(DBGWARNING, "Address[0] is not POMCW in clvl!");
          return(FALSE);
   }
   Handle->Level = levelcharfromint(numlevel);
   switch(Handle->Level)
   {
       case 'P': PeonCount++; break;
       case 'O': OpCount++; break;
       case 'M': ManagerCount++; break;
       case 'C': CoOwnerCount++; break;
       case 'W': OwnerCount++; break;
       default: 
           Debug(DBGWARNING, "Address[0] is not POMCW in clvl!");
           return(FALSE);   
   }
   
   WriteChanOpFile(cptr);
   return(TRUE);
}

int c_addlamer(C_PARAM)
{
  char *Address;
  char str[LINEBUFFSIZE + 6];
  UserListType target = NULL;
  char buff[MAXLEN], rbuff[MAXLEN];
  char nuh[MAXLEN];
  char *Reason;
  LamerListType lptr;
  char *Handle;

  if (!(Address = mysep(&rest, " ")))
  {
    send_to_user(uptr, "\002Usage:\002 ADDLAMER <nick | hostmask> [reason]");
    return(FALSE);
  }
  Reason = mysep(&rest, "\0");
  if (!AddyOk(Address))
  {
    if((target = FindNick(Address, UsersN)))
    {
      if (IsModeK(target))
      {
	send_to_user(uptr, "\002You cannot kick/deop/lamer a channel service.\002");
	return(FALSE);
      }
      sprintf(nuh, "%s!%s@%s", target->Nick, ActiveUsername(target), ActiveHost(target));
      sprintf(buff, "%s", cluster(nuh));
      Address = buff;
    }
    else
    {
      send_to_user(uptr, "\002Error:\002 User not found or illegal hostmask.");
      return(FALSE);
    }
  }
  if((lptr = GetLamer(cptr, Address)))
  {
    send_to_user(uptr, "There is already a lamer with the given hostmask on the list.");
    return(FALSE);
  }
  if (!IsGodOper(uptr))
	  Handle = hptr->Handle;
  else
	  Handle = uptr->Nick;
  if (!(Reason))
    sprintf(rbuff, "\002<%s>\002", Handle);
  else
    sprintf(rbuff, "\002<%s>\002 %s", Handle, Reason);
  Reason = rbuff;
  sprintf(str, "%s", Address);

  AddLamer(cptr, time(NULL), 0, Address, Reason);

/* kick the user, if settings direct. */
  if(target && cptr->settings[INCHAN])
  {
      deopban(cptr, target, Address);
      send_kick(cptr, target, "%s", Reason);
  }
  send_to_user(uptr, "A lamer with hostmask [%s] successfully added to the %s lamerlist!", Address, cptr->Name);
  return(TRUE);
}

int c_subscan(C_PARAM)
{
  char *Global;

  if(!rest)
  {
     send_to_user(uptr, "USAGE: subscan <who>");
     send_to_user(uptr, "Try $*.org or $*.net for a network scan.");
     return(0);
  }
  Global = mysep(&rest, " ");
  send_privmsg(Global, "%s", "\001CLIENTINFO\001");
  return(1);
}

int c_raw(C_PARAM)
{
  if (!rest)
  {
    send_to_user(uptr, "\002Error:\002 No message to send.");
    return(FALSE);
  }
  send_wallops("%s issued RAW: %s", uptr->Nick, rest);
  sendtosock(Sock, "%s", rest);
  return(TRUE);
}

int c_backup(C_PARAM)
{
  char FileName[MAXLEN];
  char SysString[MAXLEN];

  sprintf(FileName, "x2datbackup-%lu.tgz", time(NULL)); 
  send_to_user(uptr, "Data files being backed up to %s...", FileName);
  sprintf(SysString, "tar -czf %s %s CHANS/", FileName, CHANFILENAME);
  send_to_user(uptr, "Sending [%s] to the system...", SysString);
  /* (KK) Technically, X2 should freeze while this happens, if I remember
   * correctly.  What this should do is spawn a child to do this.  I could
   * be wrong, though.
   */
  system(SysString);
  sleep(1); /* to keep from overwriting if this is run multiple times */
  return(TRUE);
}

int c_version(C_PARAM)
{
  send_to_user(uptr, "\002%s\002", VERSION);
  return(TRUE);
}

int c_dellamer(C_PARAM)
{
  char buff[MAXLEN];
  UserListType target;
  char* Address;
  LamerListType lptr;

  if (!(Address = strtok(rest, " ")))
  {
    send_to_user(uptr, "\002Error:\002 No hostmask provided.");
    return(FALSE);
  }
  if (!AddyOk(Address))
  {		/* if the address doesnt have address stuff in it (!@ etc) */
    if((target = FindNick(Address, UsersN)))
    {				/* if theres a user on with that nick */
      sprintf(buff, "%s!%s@%s", target->Nick, ActiveUsername(target), ActiveHost(target));
      Address = buff;
      send_to_user(uptr, "\002Searching for a mask matching: %s\002", Address);
      for(lptr = cptr->Lamers;lptr;lptr = lptr->Next)
      {
         if(!matches(lptr->Mask, buff))
         {
            send_to_user(uptr, "...removed %s from the lamer list.", lptr->Mask);
            DelLamer(cptr, lptr->Mask); /* note that lptr is now dirty. */
            return(TRUE);
         }
      }
      return(FALSE);
    }
    else
    {				/* Nick not found */
      send_to_user(uptr, "\002Sorry, %s is not currently on IRC.  Try using the hostmask.\002", Address);
      return(FALSE);
    }
  }				/* end if(!AddyOk(Address)) */
  if(DelLamer(cptr, Address))
  {
    send_to_user(uptr, "\002...deleted %s from %s lamer list.\002", Address, cptr->Name);
    return(TRUE);
  }
  else
  {
    send_to_user(uptr, "\002Sorry, couldn't find %s in the %s lamer list.", Address, cptr->Name);
    return(FALSE);
  }
}

int c_sethandle(C_PARAM)
{
   HandleListType Handle;

   char *OldHandle;
   char *NewHandle;

  OldHandle = (char *) strtok(rest, " ");
  NewHandle = (char *) strtok(NULL, " ");

   if (!(NewHandle && OldHandle))
   {
      send_to_user(uptr, "\002Usage: SETHANDLE <old handle> <new handle>\002");
      return(FALSE);
   }
   if (!HandleOk(NewHandle))
   {
    send_to_user(uptr, "\002Handle must not have control characters in it.\002");
    return(FALSE);
  }
  if(strlen(NewHandle) > 20) /* (KK) Magic numbers are bad. */
  {
    send_to_user(uptr, "\002A handle must not exceed 20 characters.\002");
    return(FALSE);
  }
  if (FindHandle(cptr, NewHandle)) 
  {
      send_to_user(uptr, "\002A user with handle `%s' is already on the list.\002", NewHandle);
      return(FALSE);
  }
  if (!(Handle = FindHandle(cptr, OldHandle)))
  {
      send_to_user(uptr, "\002I cannot find a user by the I can't find a user with handle %s on the list.\002", OldHandle);
      return(FALSE);
  }
  if (!IsGodOper(uptr) && userlevelbychar(hptr->Level) != OW && userlevelbychar(Handle->Level) >= userlevelbychar(hptr->Level))
  {
      send_to_user(uptr, "\002You cannot change the handle of a user with an access level greater than or equal to yours!\002");
      return(FALSE);
  }
  
  SetHandle(Handle, NewHandle);
  send_to_user(uptr, "\002User %s's handle changed to %s.\002", OldHandle, Handle->Handle);
  return(TRUE);
}

int c_ircops(C_PARAM)
{
  int i, count = 0;
  UserListType ListPtr;

  send_to_user(uptr, "=--\002IRC Operators currently online\002--=");
  for (i = 0; i < USERSEGSIZE; i++)
  {
    ListPtr = UsersN->Table[i];
    while (ListPtr)
    {
      if ((strcmp((ListPtr->Serv) + strlen(ListPtr->Serv) - strlen(".Services"), ".Services"))
	  && ((UMODEo & ListPtr->Modes) != 0))
      {
        if(!(ListPtr->Modes & UMODEk))
        {
          count++;
          if(ListPtr->Modes & UMODEaway)
	     send_to_user(uptr, "%16s : \002(Away)\002", 
                          ListPtr->Nick);
          else
             send_to_user(uptr, "\002%16s\002",
                          ListPtr->Nick);
        }
      }
      ListPtr = ListPtr->NextN;
    }
  }
  send_to_user(uptr, "=---------------- End ---------------=");
  if (!count)
    /* send_to_user(uptr, "Hmm, how odd. I don't see any IRC Ops at the moment."); *
     * (KK) Changed as a simple security measure -- how about _not_ telling people there's nobody to bust them for being bad?
     */
    send_to_user(uptr, "   There are no \002visible\002 IRC Operators at the moment.");
  return(TRUE);
}

int c_kick(C_PARAM)
{
  char *targetnick;
  char *reason;
  UserListType target;

  targetnick = (char *) strtok(rest, " ");
  reason = (char *) strtok(NULL, "\0");
  if(!(targetnick))
  {
    send_to_user(uptr, "\002Usage: KICK <nick>\002");
    return(FALSE);
  }
  if(!(target = FindNick(targetnick, UsersN)))
  {
    send_to_user(uptr, "\002Error:\002 Could not find user %s.", targetnick);
    return(FALSE);
  }
  if(!isinchan(cptr, target)) {
    send_to_user(uptr, "\002Error:\002 User %s is not in the channel.", targetnick);
    return(FALSE);
  }
  if (IsModeK(target))
  {
    send_to_user(uptr, "\002Sorry, you cannot kick or deop a Channel Service.\002");
    return(FALSE);
  }
  if (!kickok(cptr, uptr, target))
  {
    send_to_user(uptr, "\x2Sorry, that user is protected\x2");
    return(FALSE);
  }
  if (reason)
  {
    send_kick(cptr, target, "<%s> %s", uptr->Nick, reason);
    return(TRUE);
  }
  else
  {
      send_kick(cptr, target, "<%s> \002You're outta-here, %s!\002", uptr->Nick, target);
    return(TRUE);
  }
  return(TRUE);
}

int c_deopchan(C_PARAM)
{
  UserListType UserNode;
  char plusohs[MAXLEN], plusnicks[MAXLEN], plusnumnicks[MAXLEN], notfounds[MAXLEN], services[MAXLEN];
  int count = 0;
  int servcount = 0;
  int notfoundcount = 0;
  char *tmp;

  if (!rest || !(*rest))
  {
    send_to_user(uptr, "\002Usage:\002 DEOPCHAN [channel] <nickname> [nickname nickname ...]\002");
    return(FALSE);
  }
  strcpy(notfounds, "");
  strcpy(plusnicks, "");
  strcpy(plusnumnicks, "");
  strcpy(services, "");
  strcpy(plusohs, "");
  tmp = mysep(&rest, " ");
  for (; tmp; tmp = (char *) mysep(&rest, " "))
  {
    if (!(UserNode = FindNick(tmp, UsersN)))
    {
      strcat(notfounds, " ");
      strcat(notfounds, tmp);
      notfoundcount++;
    }
    else if (UserNode->Modes & UMODEk)
    {
      strcat(services, " ");
      strcat(services, tmp);
      servcount++;
    }
    else
    {
      strcat(plusohs, "o");
      strcat(plusnicks, " ");
      strcat(plusnicks, tmp);
      strcat(plusnumnicks, " ");
      strcat(plusnumnicks, NickToNum(tmp));
      count++;
    }
  }
  if (count > 0)
  {
    send_hackmode(cptr, "-%s %s", plusohs, plusnumnicks);
    send_wallops("\002%s used DEOPCHAN to deop %s/[%s ].\002", uptr->Nick, cptr->Name, plusnicks);
  }
  if ((servcount > 0) || (notfoundcount > 0))
  {
    send_to_user(uptr, "\002The following nicks had errors:\002");
    if (servcount > 0)
    {
      send_to_user(uptr, "\002...cannot deop a Channel Service: [ %s]\002", services);
    }
    if (notfoundcount > 0)
    {
      send_to_user(uptr, "\002...not found:                  [ %s]\002", notfounds);
    }
  }
  if(count > 0)
    return(TRUE);
  return(FALSE);
}

int c_opchan(C_PARAM)
{
  UserListType UserNode;
  char plusohs[MAXLEN], plusnicks[MAXLEN], plusnumnicks[MAXLEN], notfounds[MAXLEN];
  int count = 0;
  int notfoundcount = 0;
  char *tmp;

  if(!cptr)
  {
    send_to_user(uptr, "\002Usage:\002 OPCHAN [channel] <nickname> [nickname nickname ...]\002");
    return(FALSE);
  }
  if (!rest || !(*rest))
  {
    send_to_user(uptr, "Usage: OPCHAN [channel] \002<nickname>\002 [nickname nickname ...]\002");
    return(FALSE);
  }
  strcpy(notfounds, "");
  strcpy(plusnicks, "");
  strcpy(plusnumnicks, "");
  strcpy(plusohs, "");
  tmp = mysep(&rest, " ");
  for (; tmp; tmp = (char *) mysep(&rest, " "))
  {
    if (!(UserNode = FindNick(tmp, UsersN)))
    {
      strcat(notfounds, " ");
      strcat(notfounds, tmp);
      notfoundcount++;
    }
    else
    {
      strcat(plusohs, "o");
      strcat(plusnicks, " ");
      strcat(plusnicks, tmp);
      strcat(plusnumnicks, " ");
      strcat(plusnumnicks, NickToNum(tmp));
      count++;
    }
  }
  if (count > 0)
  {
    send_hackmode(cptr, "+%s %s", plusohs, plusnumnicks);
    send_wallops("\002%s used OPCHAN to op %s/[%s ]\002", uptr->Nick, cptr->Name, plusnicks);
  }
  if ((notfoundcount > 0))
  {
    send_to_user(uptr, "\002The following nicks had errors:\002");
    send_to_user(uptr, "\002...not found:               [ %s]\002", notfounds);
  }
  if(count >0)
    return(TRUE);
  return(FALSE);
}

int c_down(C_PARAM)
{
  if (!strcasecmp(uptr->Nick, IrcNick))	/* If somebody made me do it.. */
    return(FALSE);
  send_chanmode(cptr, "-ov %s %s", uptr->Numnick, uptr->Numnick);
  return(TRUE);
}

int c_up(C_PARAM)
{
  if(userlevel(hptr) < OP)
      if(cptr->settings[VOICE])
      {
        send_chanmode(cptr, "+v %s", uptr->Numnick);
      }
      else
        send_to_user(uptr, "Sorry, \002VOICE\002 is disabled.");
  else
  {
      send_op(cptr, uptr);
      return(TRUE);
  }
  return(FALSE);
}

int c_devoice(C_PARAM)
{
  UserListType UserNode;
  char plusvs[MAXLEN], plusnicks[MAXLEN], notfounds[MAXLEN];
  int count = 0;
  int notfoundcount = 0;
  char *tmp;

  if(!rest || !(*rest))
  {
     send_to_user(uptr, "\002Usage:\002 DEVOICE [channel] <nickname> [nickname nickname ...]");
     return(FALSE);
  }
  strcpy(notfounds, "");
  strcpy(plusnicks, "");
  strcpy(plusvs, "");
  tmp = mysep(&rest, " ");
  for (;tmp;tmp=(char *) mysep(&rest, " "))
  {
    if (!(UserNode = FindNick(tmp, UsersN)))
    {
      strcat(notfounds, " ");
      strcat(notfounds, tmp);         
      notfoundcount++;
    }
    else
    {
      strcat(plusvs, "v");
      strcat(plusnicks, " ");
      strcat(plusnicks, UserNode->Numnick);
      count++;
    }
  }
  if (count > 0)
  {
    send_chanmode(cptr, "-%s %s", plusvs, plusnicks);
  }
  if (notfoundcount > 0)
  {
    send_to_user(uptr, "\002No such user: [%s ]\002", notfounds);
  }
  if(count > 0)
    return(TRUE);
  return(FALSE);
}    

int c_voice(C_PARAM)
{
  UserListType UserNode;
  char plusvs[MAXLEN], plusnicks[MAXLEN], notfounds[MAXLEN];
  int count = 0;
  int notfoundcount = 0;
  char *tmp; 
  
  if(!rest || !(*rest))
  {
     send_to_user(uptr, "\002Usage:\002 VOICE [channel] <nickname> [nickname nickname ...]\002");
     return(FALSE);
  }
  strcpy(notfounds, "");
  strcpy(plusnicks, "");
  strcpy(plusvs, "");
  tmp = mysep(&rest, " ");
  for (;tmp;tmp=(char *) mysep(&rest, " "))
  {
    if (!(UserNode = FindNick(tmp, UsersN)))
    {
      strcat(notfounds, " ");
      strcat(notfounds, tmp);
      notfoundcount++;
    }
    else
    {
      strcat(plusvs, "v");
      strcat(plusnicks, " ");
      strcat(plusnicks, UserNode->Numnick);
      count++;
    }      
  }
  if (count > 0)
  {
    send_chanmode(cptr, "+%s %s", plusvs, plusnicks);
  }
  if (notfoundcount > 0)
  {
    send_to_user(uptr, "\002No such user: [%s ]\002", notfounds);
  }
  if(count > 0)
    return(TRUE);
  return(FALSE);
}

int c_op(C_PARAM)
{
  UserListType UserNode;
  char plusohs[MAXLEN], plusnicks[MAXLEN], plusnumnicks[MAXLEN], notfounds[MAXLEN],
       strictops[MAXLEN], notinchans[MAXLEN];
  int count = 0;
  int notfoundcount = 0, notinchancount = 0, strictopcount = 0;
  char *tmp;

  if (!rest || !(*rest))
  {
    send_to_user(uptr, "\002Usage:\002 OP <nickname> [nickname nickname ...]");
    return(FALSE);
  }
  strcpy(notfounds, "");
  strcpy(strictops, "");
  strcpy(plusnicks, "");
  strcpy(plusnumnicks, "");
  strcpy(plusohs, "");
  strcpy(notinchans, "");
  tmp = mysep(&rest, " ");
  for (; tmp; tmp = (char *) mysep(&rest, " "))
  {
    UserNode = FindNick(tmp, UsersN);
    if (!UserNode)
    {
      strcat(notfounds, " ");
      strcat(notfounds, tmp);
      notfoundcount++;
    }
    else if(!opok(cptr, uptr, UserNode))
    {
        strcat(strictops, " ");
        strcat(strictops, tmp);
        strictopcount++;
    }
    else if(!isinchan(cptr, UserNode))
    {
        strcat(notinchans, " ");
        strcat(notinchans, tmp);
        notinchancount++;
    }
    else
    {
      strcat(plusohs, "o");
      strcat(plusnicks, " ");
      strcat(plusnicks, tmp);
      strcat(plusnumnicks, " ");
      strcat(plusnumnicks, UserNode->Numnick);
      count++;
    }
  }
  if (count > 0)
  {
     send_chanmode(cptr, "+%s %s", plusohs, plusnumnicks);
  }
  if (notfoundcount > 0)
  {
    send_to_user(uptr, "\002No such user: [%s ]\002", notfounds);
  }
  if (notinchancount > 0)
  {
    send_to_user(uptr, "\002Not in channel %s: [%s]\002", cptr->Name, notinchans);
  }
  if (strictopcount > 0)
  {
    send_to_user(uptr, "\002Sorry, no authorization to op: [%s]\002", strictops);
  }
  if(count > 0)
    return(TRUE);
  return(FALSE);
}

int c_deop(C_PARAM)
{
  UserListType UserNode;
  char plusohs[MAXLEN], plusnicks[MAXLEN], plusnumnicks[MAXLEN], notfounds[MAXLEN], services[MAXLEN],
   noaccesses[MAXLEN], notinchans[MAXLEN];
  int count = 0;
  int servcount = 0;
  int notfoundcount = 0;
  int noaccesscount = 0;
  int notinchancount = 0;
  char *tmp;

  if (!rest || !(*rest))
  {
    send_to_user(uptr, "\002Usage:\002 DEOP <nickname> [nickname nickname ...]\002");
  }
  strcpy(notfounds, "");
  strcpy(plusnicks, "");
  strcpy(plusnumnicks, "");
  strcpy(services, "");
  strcpy(noaccesses, "");
  strcpy(plusohs, "");
  strcpy(notinchans, "");
  tmp = mysep(&rest, " ");
  for (; tmp; tmp = (char *) mysep(&rest, " "))
  {
    if (!(UserNode = FindNick(tmp, UsersN)))
    {
      strcat(notfounds, " ");
      strcat(notfounds, tmp);
      notfoundcount++;
    }
    else if (IsModeK(UserNode))
    {
      strcat(services, " ");
      strcat(services, tmp);
      servcount++;
    }
    else if (!deopok(cptr, uptr, UserNode))
    {
      strcat(noaccesses, " ");
      strcat(noaccesses, tmp);
      noaccesscount++;
    }
    else if (!isinchan(cptr, UserNode))
    {
      strcat(notinchans, " ");
      strcat(notinchans, tmp);
      notinchancount++;
    }
    else
    {
      strcat(plusohs, "o");
      strcat(plusnicks, " ");
      strcat(plusnicks, tmp);
      strcat(plusnumnicks, " ");
      strcat(plusnumnicks, UserNode->Numnick);
      count++;
    }
  }
  if (count > 0)
  {
    send_chanmode(cptr, "-%s %s", plusohs, plusnumnicks);
  }
  if ((servcount > 0) || (notfoundcount > 0) || (noaccesscount > 0) || (notinchancount > 0))
  {
    send_to_user(uptr, "\002The following nicks had errors:\002");
    if (servcount > 0)
    {
      send_to_user(uptr, "\002...cannot deop a Channel Service: [%s ]\002", services);
    }
    if (notfoundcount > 0)
    {
      send_to_user(uptr, "\002...not found:                    [%s ]\002", notfounds);
    }
    if (noaccesscount > 0)
    {
      send_to_user(uptr, "\002...access denied:                [%s ]\002", noaccesses);
    }
    if (notinchancount > 0)
    {
      send_to_user(uptr, "\002...not in channel:               [%s ]\002", notinchans);
    }
  }
  if(count > 0)
    return(TRUE);
  return(FALSE);
}

int c_spoofserv(C_PARAM)
{
  char* Serv;
  char* NumStr;
  char* Reason;
  int Numeric;
	
  Serv = (char *) strtok(rest, " ");
  NumStr = (char *) strtok(NULL, " ");
  Reason = (char *) strtok(NULL, "");
  if(!Reason)
  {
     send_to_user(uptr, "Syntax: SPOOFSERV <server name> <numeric> <reason>");
     return(FALSE);
  }
  if(!strchr(Serv, '.'))
  {
     send_to_user(uptr, "Invalid server name. Server must have at least one dot (.)");
     return(FALSE);
  }
  Numeric = atoi(NumStr);
  if (Numeric == 0)
  {
	  send_to_user(uptr, "Invalid numeric: %s", NumStr);
	  return(FALSE);
  }
  send_wallops("\002SPOOFSERV %s done by %s: %s.\002", Serv, uptr->Nick, Reason);
  send_spoofserv(Serv, Numeric, "JUPE - %s added by %s: %s", Serv, uptr->Nick, Reason);
  return(TRUE);
}

int c_help(C_PARAM)
{
  FILE *HelpFile;
  boolean Oper;
  char *UpCaseStr;
  char HelpFilePath[LINEBUFFSIZE];
  char* Comm = NULL;

  Oper = IsOper(uptr);
  if(rest && *rest)
    Comm = strtok(rest, " ");

  if (Comm && (strchr(Comm, '/') || strchr(Comm, '.')))
  {
    send_to_user(uptr, "\002Sorry, invalid help parameter.  Try using just HELP.\002");
    return(FALSE);
  }
  if (!Comm || !(*Comm))
  {
    if (Oper)
    {
      HelpFile = fopen(OPERHELPFILE, "r");
      if (HelpFile)
      {
	ShowOperHelp(uptr, HelpFile);
	fclose(HelpFile);
      }
      else
      {
	send_to_user(uptr, "The operator help file is not available at the moment (%s).", strerror(errno));
	return(FALSE);
      }
    }
    else
    {
      HelpFile = fopen(USERHELPFILE, "r");
      if (HelpFile)
      {
	ShowUserHelp(uptr, HelpFile);
	fclose(HelpFile);
      }
      else
      {
	send_to_user(uptr, "The user help file is not available at the moment (%s).", strerror(errno));
	return(FALSE);
      }
    }
  }
  else
  {
    UpCaseStr = UpCase(Comm);
    sprintf(HelpFilePath, "./HELP/%s", UpCaseStr);
    HelpFile = fopen(HelpFilePath, "r");
    if (HelpFile)
    {
      send_to_user(uptr, "=--- \002%s\002 ---=", UpCaseStr);
      ShowCommHelp(uptr, HelpFile);
      send_to_user(uptr, "=--- End of help ---=");
      fclose(HelpFile);
    }
    else
      send_to_user(uptr, "Sorry, there is no help on that topic.");
      send_to_user(uptr, "Use /msg %s HELP for a menu of commands.", IrcNick);
  }
  return(TRUE);
}

int c_motd(C_PARAM)
{
  if(!SendMOTD(uptr))
  {
     send_to_user(uptr, "There is no MOTD today.");
  }
  return(TRUE);
}

int c_authors(C_PARAM)
{
  send_to_user(uptr, "\002X2 was originally created by Robert Emmery in December 1996.\002");
  send_to_user(uptr, "\002...most functions have been updated and improved by Rubin.\002");
  send_to_user(uptr, "\002Special thanks to all those who have contributed code and bug\002");
  send_to_user(uptr, "\002reports over the years (too many to track)  Thanks to AfterNET\002");
  send_to_user(uptr, "\002for putting up with most of the development cycle, and the\002");
  send_to_user(uptr, "\002Asylum and Games IRC Networks who sent in lots of helpful bug\002");
  send_to_user(uptr, "\002reports and fixes!\002");
  send_to_user(uptr, "\002---\002");
  send_to_user(uptr, "\002X2 is now on Sourceforge and has numerous contributors.  See\002");
  send_to_user(uptr, "\002the following URL for more information:");
  send_to_user(uptr, "\002\002");
  send_to_user(uptr, "\002This GNU service is available with source code from:\002");
  send_to_user(uptr, "\002http://x2serv.sourceforge.net\002");
  send_to_user(uptr, "\002         O R\002");
  send_to_user(uptr, "\002ftp://agora.rdrop.com/pub/users/rubin/x2\002");
  return(TRUE);
}
