/*
 *    X2  IRC Operator and Channel services
 *    by Rubin (rubin@afternet.org)
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 1, or (at your option)
 * any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#include <signal.h>
#include <sys/param.h>
#include <sys/stat.h>
#include <errno.h>
#include <unistd.h>

#include "config.h"
#include "init.h"
#include "chan.h"
#include "user.h"
#include "parse.h"
#include "misc.h"
#include "sock.h"
#include "server.h"
#include "gate.h"
#include "send.h"
#include "route.h"

boolean DaemonStarted;
boolean AUTOROUTING, AUTOCONNECT, ENFORCEOPS, PINGOUTCONNECT, GATECHECK,
        FLOODCONTROL, NETLINK, NETTALK, PUBLIC, NETBURSTING;
unsigned char CLONERES=0, MAXCLONES, MAXCHAN, DEBUG, SUB7SCAN=0, LITMUS=0, SDBOT=0;
char ServiceName[MAXLEN] = "UNSET",
     Uplink[MAXLEN], ServAddress[16], Password[20], 
     IrcNick[MAXNICKLEN] = "X2", 
     CommandMark = '.', HomeDir[2048]="", AccessLog[2048]="";
int ServiceNumeric = 53;
char ServNum[SERVNUMLEN] = "A1";
char NumNick[NUMNICKLEN] = "A1AAA";

u_short Port, SERVERPORT;
int Sock, NewSock;
struct sockaddr_in ServAddr, CliAddr;
fd_set fdvar, wfdvar;
int MaxSock;

boolean SockConnected;
boolean ServerConnected;
boolean SentBurst;
boolean Authenticated;
boolean RoutsOk;
int GotPong;
boolean NOFORK = FALSE;

time_t LinkTimeStamp, ServerTimeStamp, ConnectTime, DiscTime, StartTime,
       PingStamp, PongStamp, FloodStamp;
long int NetOffset;
time_t bantimer;

char LastNickAdded[MAXNICKLEN];
char LastRequestNick[MAXNICKLEN];

ChanSegment Channels;
GlineListType *Glines;
GlineListType *SuperGlines;
ServSegment Servers, Routs;
UserSegment UsersN, UsersA;
UserSegment UsersU;

int ChannelCount, ServerCount, RoutCount, UserCount, OperCount,
    PeonCount, OpCount, ManagerCount, CoOwnerCount, OwnerCount, LamerCount;

NetMessageType NetMessage;
UpLinkList *CurUpLink;
UpLinkList *UpLinks = NULL;
extern int REMOTE_AUTHENTICATED;
char BindIP[80] = "";
IllegalChanListType IllegalChannels;

/* Start of the code section */

void StartDaemon(int Sock)
{
#ifndef debugon
  register int childpid;
#endif
  FILE *ProcFile;

#ifndef debugon
  if(!NOFORK)
  {
   childpid = fork();
   if (childpid < 0)
      exit(1);
   else if (childpid > 0)
      exit(0);
   close(0);			/* stdin */
   close(1);			/* stdout */
   close(2);			/* stderr */
   setsid();
  }
#endif
  if((ProcFile = fopen(ProcFilePath, "w")))
     fprintf(ProcFile, "%d", getpid());
  fclose(ProcFile);
  DaemonStarted = TRUE;
}

void AddUpLinkServer(char *ServIp, char *ServPass, char *ServName, int port)
{
   UpLinkList *UpLinkPtr;

   if(!(UpLinkPtr = malloc(sizeof(*UpLinkPtr))))
   {
      Debug(DBGSEVERE, "Out of memory adding server to uplink list");
      return;
   }
   strcpy(UpLinkPtr->ServName, ServName);
   strcpy(UpLinkPtr->ServIp, ServIp);
   strcpy(UpLinkPtr->ServPass, ServPass);
   UpLinkPtr->Port = port;
   UpLinkPtr->Next = UpLinks;
   UpLinks = UpLinkPtr;
}

int ReadConfig(char *Filename)
{

  FILE *ConfFile;
  char StrBuffr[MAXCONFSTRBUFF];
  char ch;
  UpLinks = NULL;

  if ((ConfFile = fopen(Filename, "r")) == NULL)
  {
    printf("Can't open file \"%s\"!\n", Filename);
    return (0);
  }
  while (fscanf(ConfFile, "%s", StrBuffr) != EOF)
  {
    if (StrBuffr[0] == '#')
    {
      while ((ch = fgetc(ConfFile)) != '\n')
	;
      continue;
    }
    if (!strcmp(StrBuffr, "SERVER"))
    {
      char *ServIp, *ServPass, *ServName, *ServPort;
      int IntPort = 0;

      fscanf(ConfFile, "%s", StrBuffr);
      ServIp = (char *) strtok(StrBuffr, ":");
      ServPass = (char *) strtok(NULL, ":");
      ServName = (char *) strtok(NULL, ":");
      ServPort = (char *) strtok(NULL, ":");
      if(!ServPort)
      {
         printf("Incorrect SERVER line syntax.\n");
         exit(1);
      }
      if((IntPort = atoi(ServPort)) >= 0)
      {
        AddUpLinkServer(ServIp, ServPass, ServName, IntPort);
        printf("Adding uplink server %s, IP %s on port %s.\n", ServName, ServIp, ServPort);
      }
      else
      {
        printf("Invalid port (%d) in SERVER line.\n", IntPort);
        exit(1);
      }
    }
    else if (!strcmp(StrBuffr, "SERVICE"))
    {
      fscanf(ConfFile, "%s", StrBuffr);
      strcpy(ServiceName, StrBuffr);
    }
    else if (!strcmp(StrBuffr, "NUMERIC"))
    {
    	int i;
    	
        fscanf(ConfFile, "%s", StrBuffr);
        
        /* Numeric must be 0-4095 */
        if ( (i = atol(StrBuffr)) & ~(4096-1) ) {
        	printf("Invalid numeric (%d) in NUMERIC line.  \n", i);
        	exit(1);
        }
        	
	inttobase64(ServNum, i, 2);
	strcpy(NumNick, ServNum);
	strcat(NumNick, "AAA");
	ServiceNumeric = i;
    }
    else if (!strcmp(StrBuffr, "PASSWORD"))
    {
      fscanf(ConfFile, "%s", StrBuffr);
      strcpy(Password, StrBuffr);
    }
    else if (!strcmp(StrBuffr, "IRCNICK"))
    {
      fscanf(ConfFile, "%s", StrBuffr);
      strcpy(IrcNick, StrBuffr);
      printf("Nick: %s\n", IrcNick);
    }
    else if (!strcmp(StrBuffr, "HOME"))
    {
      fscanf(ConfFile, "%s", StrBuffr);
      strcpy(HomeDir, StrBuffr);
      printf("Home Directory: %s\n", HomeDir);
    }
    else if (!strcmp(StrBuffr, "ACCESSLOG"))
    {
      fscanf(ConfFile, "%s", StrBuffr);
      strcpy(AccessLog, StrBuffr);
      printf("Logging network access to %s\n", AccessLog);
    }
    else if (!strcmp(StrBuffr, "PINGOUTCONNECT"))
    {
      fscanf(ConfFile, "%s", StrBuffr);
      PINGOUTCONNECT = atoi(StrBuffr);
      printf("AutoConnecting pingouts: %s\n", GetSetting(PINGOUTCONNECT));
    }
    else if (!strcmp(StrBuffr, "GATECHECK"))
    {
      fscanf(ConfFile, "%s", StrBuffr);
      GATECHECK = atoi(StrBuffr);
      printf("Gatecheck: %s\n", GetWinGateMet(GATECHECK));
    }
    else if (!strcmp(StrBuffr, "AUTOROUTING"))
    {
      fscanf(ConfFile, "%s", StrBuffr);
      AUTOROUTING = atoi(StrBuffr);
      printf("Automated Routing: %s\n", GetSetting(AUTOROUTING));
    }
    else if (!strcmp(StrBuffr, "AUTOCONNECT"))
    {
      fscanf(ConfFile, "%s", StrBuffr);
      AUTOCONNECT = atoi(StrBuffr);
      printf("Automated Reconnection: %s\n", GetSetting(AUTOCONNECT));
    }
    else if (!strcmp(StrBuffr, "ENFORCEOPS"))
    {
      fscanf(ConfFile, "%s", StrBuffr);
      ENFORCEOPS = atoi(StrBuffr);
      printf("Enforcing Ops: %s\n", GetSetting(ENFORCEOPS));
    }
    else if (!strcmp(StrBuffr, "MAXCLONES"))
    {
      fscanf(ConfFile, "%s", StrBuffr);
      MAXCLONES = atoi(StrBuffr);
    }
    else if (!strcmp(StrBuffr, "CLONERES"))
    {
      fscanf(ConfFile, "%s", StrBuffr);
      CLONERES = atoi(StrBuffr);
      printf("Clone Resolution Method: %s\n", GetCloneResMet(CLONERES));
      if (CLONERES)
	printf("Clone Limit: %d\n", MAXCLONES);
    }
    else if (!strcmp(StrBuffr, "LISTEN"))
    {
      fscanf(ConfFile, "%s", StrBuffr);
      SERVERPORT = atoi(StrBuffr);
      printf("I will listen on port %d if not connected.\n", SERVERPORT);
    }
    else if (!strcmp(StrBuffr, "CMDCHAR"))
    {
      fscanf(ConfFile, "%s", StrBuffr);
      CommandMark = *StrBuffr;
      printf("My command character is `%c'.\n", CommandMark);
    }
    else if(!strcmp(StrBuffr, "SUB7SCAN"))
    {
       fscanf(ConfFile, "%s", StrBuffr);
       SUB7SCAN = atoi(StrBuffr);
       printf("Sub7Scan: %s\n", GetWinGateMet(SUB7SCAN));
    }
     else if(!strcmp(StrBuffr, "LITMUS"))
    {
       fscanf(ConfFile, "%s", StrBuffr);
       LITMUS = atoi(StrBuffr);
       printf("Litmus Scan: %s\n", GetWinGateMet(LITMUS));
    }
    else if (!strcmp(StrBuffr, "SDBOT"))
    {
	fscanf(ConfFile, "%s", StrBuffr);
	SDBOT = atoi(StrBuffr);
	printf("SDBot Scan: %s\n", GetWinGateMet(SDBOT));
    }
    else if(!strcmp(StrBuffr, "BIND"))
    {
        fscanf(ConfFile, "%s", StrBuffr);
        strcpy(BindIP, StrBuffr);
        printf("I will bind my IP to %s.\n", BindIP);
    }
    else
    {
      printf("...unknown configuration setting \"%s\"\n", StrBuffr);
      fclose(ConfFile);
      exit(1);
      return (0);
    }
  }
  fclose(ConfFile);
  return (1);
}

int ReadGlineFile(char type, char *filename)
{
  FILE *TheFile;
  char address[MAXLEN], reason[MAXLEN];
  time_t gtime;
  int s;

  if (!(TheFile = fopen(filename, "r")))
  {
    Debug(DBGWARNING, "No gline file found.");
    return (1);
  }
  while ((s = fscanf(TheFile, "%s%lu %[^\n]\n", address, &gtime, reason)) != EOF)
  {
    if (s != 3)
    {
      Debug(DBGWARNING, "Corruption in GLINE file!");
      return (2);
    }
    if(type == 'G')
      ProcessAddGline(address, gtime - NetTime(), reason);
    else if(type == 'S')
      ProcessAddSGline(address, gtime, reason);
  }
  fclose(TheFile);
  return(0);
}

int ReadChanFile()
{
  FILE *TheFile;
  FILE *OpFile;
  char linebuff[LINEBUFFSIZE], *key;
  char *ptrbuff, *lb;
  char ChanName[LINEBUFFSIZE], ChanTopic[LINEBUFFSIZE], ModeString[LINEBUFFSIZE], *OpAddress;
  char ChanEmail[LINEBUFFSIZE];
  char URL[LINEBUFFSIZE], Description[LINEBUFFSIZE], Reggedby[MAXNICKLEN+1];
  int i, s, line = 0, line2, dupcount, fileversion=0;
  int UserAutoOp, UserPwdReq;
  int settings[NUMOFCHANSETTINGS+1];
  time_t LastAccess = 0, CreationDate = 0;
  time_t LastSeen = 0;
  time_t BanTime = 0;
  char ChanFileName[LINEBUFFSIZE + 6];
  char *UpCaseChan;
  char *Pass;
  char *Info;
  char FullChanName[LINEBUFFSIZE];
  char *Handle;
  char *Mask;
  char *Reason;
  char *Suspended;
  char Level;   /* WCMOPL */
  int AutoOp;
  HandleListType LastHandle = NULL, hptr;
  ChannelListType cptr = NULL;

  Debug(DBGNOTICE, "Processing chan.dat...\n");
  if (!(TheFile = fopen(CHANFILENAME, "r")))
  {
    printf("...no chan.dat...creating...");
    TheFile = fopen(CHANFILENAME, "w");
    fclose(TheFile);
    return (1);
  }
  while ((s = fscanf(TheFile, "%s\n", ChanName)) != EOF)
  {
    line++;
    if (s != 1)
    {
      printf("Parameter error in chan.dat, line %d. only %d nybbles read.\n", line, s);
      return (0);
    }
    if(GetChannel(ChanName))
    {
       printf("Channel %s listed TWICE!  Bailing...\n", ChanName);
       return(0);
    }
    UpCaseChan = UpCase(ChanName);
    sprintf(ChanFileName, "./CHANS/%s", &UpCaseChan[1]);
    free(UpCaseChan);
    if (!(OpFile = fopen(ChanFileName, "r")))
    {
      printf("Error opening OpFile for %s; chan.dat line %d.\n", ChanName, line);
      return (0);
    }
    line2 = 0;
    dupcount = 0;
    strcpy(ChanTopic, "*");
    strcpy(ModeString, "");
    strcpy(Description, "*");
    strcpy(URL, "*");
    strcpy(Reggedby, "UNKNOWN");
    Suspended = NULL;
    while (!feof(OpFile))
    {
      i++;
      s = fscanf(OpFile, "%[^\n]\n", linebuff);
      line2++;
      if (s == EOF)
	break;
      if (s != 1)
      {
         printf("Corruption in CHANS/%s, line %d.", ChanName, line2);
         return(0);
      }
      lb = linebuff;
      key = (char *) mysep(&lb, " ");
      if(!key || !*key)
      {
         printf("Cant really happen..\n");
         return(0);
      }
      switch(key[0])
      {
         case 'V': /* VERSION   current version = 4*/
                ptrbuff = (char *) mysep(&lb, " ");
		if(!ptrbuff)
                {
                    printf("Invalid VERSION syntax in CHANS/%s, line %d.\n", ChanName, line2);
                    return(0);
                }
                fileversion = strtol(ptrbuff, NULL, 10);
                if(fileversion > 4)
                {
                   printf("Warning!  Reading CHANS/%s which is file version %d.\n", ChanName, fileversion);
                }
                if(fileversion == 1)  /* Version 1 has no EMAIL field so fill something in. */
                {
                   strcpy(ChanEmail, "*");
                }
                continue;
         case 'C': /* Channel */
                ptrbuff = (char *) mysep(&lb, "\n");
		safecopy(FullChanName, ptrbuff, MAXLEN);
                if(strcasecmp(FullChanName, ChanName))
                {
                  printf("Warning!  CHANS/%s claims to be %s.  Quitting..\n", ChanName, ptrbuff);
                  return(0);
                }
                continue;
         case 'T': /* Topic */
                ptrbuff = (char *)mysep(&lb, "\n"); 
		safecopy(ChanTopic, ptrbuff, MAXTOPICLEN);
                continue;
         case 'E': /* Email */
                ptrbuff = (char *)mysep(&lb, "\n");
                if(ptrbuff[0] == '#')
                { /* Fix some corruption problems i had */
                   ptrbuff = "*";
                }
                safecopy(ChanEmail, ptrbuff, LINEBUFFSIZE);
                continue;
         case 'M': /* ModeString */
                ptrbuff = (char *)mysep(&lb, "\n");
		safecopy(ModeString, ptrbuff, MAXLEN);
                continue;
         case 'D': /* Description */
		ptrbuff = (char *)mysep(&lb, "\n");
		safecopy(Description, ptrbuff, MAXLEN);
                continue;
         case 'S': /* Settings */
                for(i=0;(ptrbuff = (char *) mysep(&lb, " "));i++)
                {
                   if((*ptrbuff))
                     settings[i] = strtol(ptrbuff, NULL, 10);
                }
		if(i < NUMOFCHANSETTINGS)
                {    /* If settings have been added,
			set them to the defined default
			values */
                   for(;i<NUMOFCHANSETTINGS;i++)
                      settings[set_commands[i].index] = set_commands[i].def;
                }
		/* If the suspended flag is used, convert it to the new system */
		/* This is temporary code, to be removed after all files are converted */
		if (settings[10])
		{
		  Suspended = malloc(strlen("Unknown reason") + 1);
		  strcpy(Suspended, "Unknown reason");
		  settings[10] = 0;
		}
                continue;
         case 'L': /* LastVisted */
                ptrbuff = (char *)mysep(&lb, " ");
                LastAccess = strtol(ptrbuff, NULL, 10);
                continue;
         case 'R': /* Regged by */
                ptrbuff = (char *)mysep(&lb, " ");
		safecopy(Reggedby, ptrbuff, MAXNICKLEN);
                ptrbuff = (char *)mysep(&lb, " ");
                CreationDate = strtol(ptrbuff, NULL, 10);

                /* This is the last required line, lets add the channel. */
                if (!AddChan(ChanName, TRUE, ModeString, ChanTopic, ChanEmail,
                   Description, URL, Reggedby,
                   settings, CreationDate, LastAccess))
                {
                    printf("Error with AddChan in %s line %d\n", ChanName, line2);
                    return (0);
                }
                cptr = GetChannel(ChanName);
                if(!cptr)
                {
                    printf("Error: Channel not added properly %s line %d", ChanName, line2);
                    return(0);
                }
		if (Suspended)
		    cptr->Suspended = Suspended;
                continue;
         case 'W': /* Website (URL) */
		ptrbuff = (char *)mysep(&lb, " ");
		safecopy(URL, ptrbuff, MAXLEN);
		continue;
         case 'H': /* HANDLE */
                if(cptr == NULL)
                {
                     printf("HANDLE before channel fully defined. %s line, %d", ChanName, line2);
                     return(0);
                }
                Handle = (char *) mysep(&lb, " ");
                ptrbuff = (char *) mysep(&lb, " ");
                Level = ptrbuff[0];
                ptrbuff = (char *) mysep(&lb, " ");
                AutoOp = strtol(ptrbuff, NULL, 10);
                ptrbuff = (char *) mysep(&lb, " ");
                LastSeen = strtol(ptrbuff, NULL, 10);
                Pass = (char *) mysep(&lb, " ");
                Info = (char *) mysep(&lb, "\n");
                /* ******************************************** */
                if(!Info)
                {
                    printf("Invalid HANDLE syntax in CHANS/%s, line %d\n", ChanName, line2);
                    return(0);
                }
                if(strlen(Pass) > MAXPASS)
                {
                    printf("Error in CHANS/%s, line %d. Password too long.\n", ChanName, line2);
                    return(0);
                }
                if(strlen(Handle) > 20)
                {
                    printf("Error in CHANS/%s, line %d.  Handle too long.\n", ChanName, line2);
                    return(0);
                }
                if(strlen(Info) > 200)
                {
                    printf("Error in CHANS/%s, line %d.  Infoline too long.\n", ChanName, line2);
                    return(0);
                }
                if(fileversion < 4)
                {
                     /* Convert to crypted. */
                     Pass = crypt(Pass, "$1$92737837$"); /* now using md5 for passwords to allow MUCH longer length */
                }
                LastHandle = AddHandle(cptr, Handle, Pass, Info, Level, AutoOp, LastSeen, 0);
                continue;
         case 'K': /* MASK */
                if(LastHandle == NULL)
                {
                    printf("Error in CHANS/%s, line %d.  MASK line does not follow HANDLE.\n", ChanName, line2);
                    return(0);
                }
                Mask = (char *) mysep(&lb, " ");
                if (AddyOk(Mask))
                    AddMask(LastHandle, Mask);
                continue;
         case 'B': /* BAN (lamer lits) */
                
                ptrbuff = (char *) mysep(&lb, " ");
                BanTime = strtol(ptrbuff, NULL, 10);
                ptrbuff = (char *) mysep(&lb, " ");
                LastSeen = strtol(ptrbuff, NULL, 10);
                Mask = (char *) mysep(&lb, " ");
                Reason = (char *) mysep(&lb, "\n\0");
                if(!AddyOk(Mask))
                {
                    printf("Ignoring lamer %s because its not valid.\n", Mask);
                    continue;
                }
                AddLamer(cptr, BanTime, LastSeen, Mask, Reason);
                continue;
         case 'U': /* USER */  /* Now defunct. convert to handle/mask only */
                if(cptr == NULL)
                {
                    printf("USER before channel fully defined (%s line %d).\n", ChanName, line2);
                    return(0);
                }
                OpAddress = (char *)mysep(&lb, " ");
                ptrbuff = (char *)mysep(&lb, " ");
                UserAutoOp = strtol(ptrbuff, NULL, 10);
                if(fileversion >= 2)
                {
                  ptrbuff = (char *)mysep(&lb, " ");
                  UserPwdReq = strtol(ptrbuff, NULL, 10);
                }
                else  /* Hack to be backward compatible w/ version 1 data file */
                {
                  UserPwdReq = 1;
                }
                ptrbuff = (char *)mysep(&lb, " ");
                LastSeen = strtol(ptrbuff, NULL, 10);
                Handle = (char *)mysep(&lb, " ");
                Pass = (char *)mysep(&lb, " ");
                Info = (char *)mysep(&lb, "\n");
                /* ******************************************** */
		if(!Info)
                {
                    printf("Invalid USER syntax in CHANS/%s, line %d\n", ChanName, line2);
                    return(0);
                }
                if(strlen(Pass) > MAXPASS)
                {
                    printf("Error in CHANS/%s, line %d. Password too long.\n", ChanName, line2);
                    return(0);
                }
                Pass = crypt(Pass, "$1$92737837$");
                if(strlen(Handle) > 20)
                {
                    printf("Error in CHANS/%s, line %d. Handle too long.\n", ChanName, line2);
                    return(0); 
                }
                if(strlen(Info) > 200)
                {
                    printf("Error in CHANS/%s, line %d. Infoline too long.\n", ChanName, line2);
                    return(0);
                }    
                if(OpAddress[0] == 'L')
                {
                  /* Add lamer list entry and return */
                  AddLamer(cptr, time(NULL), LastSeen, &OpAddress[1], Info);
                  continue;
                }
                /* A hack to enforce the new Handle length limit set in config.h */
                if(strlen(Handle) > MAXHANDLELEN)
                       Handle[MAXHANDLELEN] = '\0';
                if(!(hptr = FindHandle(cptr, Handle)))
                {
                      hptr = AddHandle(cptr, Handle, Pass, Info, OpAddress[0], UserAutoOp, LastSeen, 0);
                }
                if(!strcmp(crypt("*", "X2"), hptr->Pass))  /* If they didnt have a pass set, add all there masks.. */
                   AddMask(hptr, &OpAddress[1]);
		continue;
         case 'P': /* PSuspended (The P is added so that every option starts with a different char */
                if(cptr == NULL)
                {
                    printf("PSUSPENDED before channel fully defined (%s line %d).\n", ChanName, line2);
                    return(0);
                }
                ptrbuff = (char *)mysep(&lb, "\n"); 
		cptr->Suspended = malloc(strlen(ptrbuff) + 1);
		strcpy(cptr->Suspended, ptrbuff);
	        continue;
         case 'X': /* XSuspendedTime (see PSuspended) */
                if (cptr == NULL)
                {
                    printf("XSUSPENDEDTIME before channel fully defined (%s line %d).\n", ChanName, line2);
                    return (0);
                }
                ptrbuff = (char *) mysep(&lb, " ");
                cptr->SuspendedTime = strtol(ptrbuff, NULL, 10);
                continue;

         default: printf("Unknown key %s in CHANS/%s\n", key, ChanName);
                  return(0);
      }
    }
    fclose(OpFile);
  }
/*  fclose(OpFile); */
  fclose(TheFile);
  Debug(DBGNOTICE, ":Finished reading channel data files\n");
  return (1);
}

void DoTimerStuff()
{
  Debug(DBGINFO, "Doing time stuff");
}

int main(int argc, char *argv[])
{
  int c, s;
  int logfile;
  int FirstLoop = 1;
  char *cfgfile;
  char tmp[MAXLEN], buf[MAXLEN];
  static struct timeval polltime;
  time_t TopicTimer_4, TopicTimer_3, TopicTimer_2, TopicTimer_1, TopicTimer_0;
  struct stat StatStruct;
  char SocketBuffer[MAXRECVLEN+10];

  cfgfile = DefaultConfigFile;
 
  if (argc > 2) 
  {
     printf("try -h for help\n\n");
     exit(1);
  }
  if(argc == 2)
  {
    if( !strcmp(argv[1], "-h"))
    {
       printf("%s\n", VERSION);
       printf("By Rubin@afternet.org\n");
       printf("\nUsage: %s [-h|] [-r]\n\n", argv[0]);
       printf("-n = Don't fork\n");
       printf("-h = this help\n");
       exit(0);
    }
    else if (!strcmp(argv[1], "-n"))
    {
       NOFORK = TRUE;
    }
  }
  if (!ReadConfig(cfgfile))
  {
    printf(":configuration file error! cant open %s. Aborting.\n", cfgfile);
    exit(1);
  }
  if (chdir(HomeDir) < 0)
  {
    printf(":Couldn't cd to %s\nAborting!\n", HomeDir);
    exit(1);
  }
  if (!InitGlobals())
  {
    fprintf(stderr, ":initialization error! I need more memory! -- Aborting..\n");
    exit(1);
  }
  if (!ReadChanFile())
  {
    printf(":error encountered while reading channel file! - possibly missing a file in CHANS directory.\n");
    exit(1);
  }
  if (!(s = ReadGlineFile('G', GLINEFILENAME)))
  {
    if(s == 1)
    {
       printf(":gline.dat not found.");
    }
    else if(s == 2)
    {
      printf(":error encountered while reading gline data file!\n");
      printf(":Continuing anyway..\n");
    }
  }
  if (!(s = ReadGlineFile('S', SUPERGLINEFILENAME)))
  {
    if(s == 1)
    {  
       printf(":supergline.dat not found.");
    }
    else if(s == 2)
    { 
      printf(":error encountered while reading supergline data file!\n");
      printf(":Continuing anyway..\n");
    }                                                           
  } 
  ReadIllegalChannels();
  LoadRouteFile(Route, "routing.dat");
  stat(MESSAGEFILENAME, &StatStruct);
  NetMessage.Timestamp = StatStruct.st_mtime;
  Debug(DBGINFO, "NetMessage file %s has timestamp %lu", MESSAGEFILENAME, NetMessage.Timestamp);
  signal(SIGTTOU, SIG_IGN);
  signal(SIGTTIN, SIG_IGN);
  signal(SIGTSTP, SIG_IGN);
  signal(SIGPIPE, SIG_IGN);

  SockConnected = FALSE;
  ServerConnected = FALSE;
  DaemonStarted = FALSE;
  SentBurst = FALSE;
  Authenticated = FALSE;

  /* Prime us so we start with the first server this round.. 
   * This is kinda kludgy but makes the clode below have to do less tricks 
   * each loop..  This way it thinks it just had a bad connection (ConnectTime is less then 30 seconds ago)
   * and it will cycle the server and connect.. */
  if(!UpLinks)
  {
     printf("No servers defined. Edit x2.conf.\n");
     exit(1);
  }
  for(CurUpLink = UpLinks;CurUpLink->Next;CurUpLink = CurUpLink->Next);

  /* Initalize our timer system. crude but does the job */
  StartTime = bantimer = TopicTimer_4 = TopicTimer_3 = TopicTimer_2 = TopicTimer_1 = TopicTimer_0 = time(NULL);

/** WHY WHY WHY was this opening readonly?? it creates the file readonly! */
/* This i think is our famous file bug. */
/* should be fixed now.. (O_CREAT requires a 3rd parameter...) -relnev */
/*  logfile = open(AccountingLog, O_WRONLY | O_CREAT | O_APPEND, S_IRUSR | S_IWUSR); */
  logfile = open(AccountingLog, O_CREAT | O_APPEND, S_IRUSR | S_IWUSR);
  if (logfile != -1)
  {
    strcpy(tmp, ctime(&StartTime));
    tmp[strlen(tmp) - 1] = '\0';
    sprintf(buf, ":%s:%s Started\n", tmp, IrcNick);
    write(logfile, buf, strlen(buf));
    close(logfile);
  }
  printf("-=[*](%s Is Engaged)[*]=-\n", IrcNick);
  StartDaemon(Sock);
  for (;;)   /* Main loop */
  {
    if (!SockConnected)
    {
      int trycount;

      DiscTime = time(NULL);
      if(!FirstLoop)
        sleep(7); /* Dont want to flood us if theres an error.. */
      else
        FirstLoop = FALSE;
      close(Sock);
      ReadSocket(-1, NULL); /* Clear out the remainder buff */
      trycount = 0;
      Debug(DBGINFO, "Socket disconnected. (Re)Starting connection phase.");
      ServerConnected = FALSE;
      REMOTE_AUTHENTICATED = FALSE; /* for ProcessLine in parse.c */
      NETBURSTING = TRUE;
      SentBurst = FALSE;
      Authenticated = FALSE;
      KillServerList();
/* Are there any users if there are no servers?? */
/* YES! x2 itself. -Rubin */
      KillUserList(); 
      ClearAllChans();
      OperCount = 0;
      if(AUTOCONNECT)		/* If we should try to connect */
      {
        if(time(NULL) - ConnectTime < 30) /* If we get dumped in under 30 seconds, try a new server */
        {
           CurUpLink = CurUpLink->Next;
           if(!CurUpLink)
              CurUpLink = UpLinks;
        }
        while(!(TryToConnect(CurUpLink)))
        {
            CurUpLink = CurUpLink->Next;
            if(!CurUpLink)
            {
               CurUpLink = UpLinks;
               if(trycount++ > 2)
                 break; /* Try the loop twice, then just listen.. */
            }
        }
        if(SockConnected)
        {
          Authenticate(CurUpLink);
          Authenticated = TRUE;
          SentBurst = TRUE;
          /* Update rmap info for new point of view.. */
          ChangeRouteUplinks(Route);
          RankOutsideNess(Route);

        }
        else
	{
	  Debug(DBGINFO, "Unable to obtain connection. Listning on port %d", SERVERPORT);
	  Sock = my_init_server();
	}
      }				/* End if autoconnect */
      else
      {
	/* Stop and wait for a connection. */
        Debug(DBGINFO, "Listning for connection on port %d\n", SERVERPORT);
	Sock = my_init_server();
      }
    }
    if(!SockConnected)
       continue;
    SelectSockets();		/* Setup select() for our socket(s) */
    polltime.tv_sec = (unsigned long) PINGFREQ;
    polltime.tv_usec = (unsigned long) 1;
    c = select(MaxSock + 1, &fdvar, &wfdvar, NULL, &polltime);
      if(c == -1)
      {
#ifdef debugon
perror("SELECT");
#endif
          /*Debug(DBGSEVERE, "select returned -1. Errno:%d", errno);
           *exit(1);  profiling causes "Interrupted System Call" here.. 
          */
      }
      if(c > 0) /* Incomming data */
      {
        if(FD_ISSET(Sock, &fdvar) )
        {
           if(ReadSocket(Sock, SocketBuffer) == -1)
           {
              Debug(DBGWARNING, "Read from socket returned -1. errno: %d", errno);
              SockConnected = FALSE;
              continue;
           }
           else
              PongStamp = time(NULL);
        }
      }
      else
      {
          /* These things happen only if select times out */
          if(c == 0)
          {
             Debug(DBGINFO, "select() timeout.");
          }
          if (((time(NULL) - PongStamp) > PINGFREQ) && (GotPong))
          {
            send_serverping(Uplink);
            GotPong = false;
            PingStamp = time(NULL);
          }

      }
    /* These also check for timeout, so call them
     * even if select just timesout */
      GateCheck_CheckSend(wfdvar);
      GateCheck_CheckRead(fdvar);

#ifdef WEBSERV
    /* Handle the webserver */
      webserv_checkports();
#endif 

    /* Handle time stuff here: */
      if (ServerConnected)
      {
	int Now = time(NULL);
        /* Check for ping timeout */
	if (((Now - PingStamp) > PINGTIMEOUT) && (!GotPong))
	{
          Debug(DBGINFO, "No ping recieved. Killing link..");
	  GotPong = TRUE;
	  SockConnected = false;
	  continue;
	}
        /* Check for bans */
	if ((Now - bantimer) >= 120)
	{
          bantimer = Now;
	  CheckBanTimeouts();
          CheckGlineTimeouts();
          CheckSuspendedTimeouts();
	  NETBURSTING = FALSE;	/* after 2 min start parsing on_join completely */
	}
        /* Set topics and expire channels once a day */
	if ((Now - TopicTimer_1) >= 86400)
	{
	  TopicTimer_1 = Now;
	  SetTopics(1);		/* One a day and less */
          ExpireChannels(NULL); /* This seems like a good place */
	}
	if ((Now - TopicTimer_2) >= 43200)
	{
	  TopicTimer_2 = Now;
	  SetTopics(2);		/*  12 hours and less */
	}
	if ((Now - TopicTimer_3) >= 14400)
	{
	  TopicTimer_3 = Now;
	  SetTopics(3);		/*  4 hours and less */
          /* Every 4 hours we try to reconnect missing servers */
          Reroute(Route, NULL, 'C');
	}
	else if ((Now - TopicTimer_4) >= 3600)
	{			/* Once an hour fix hourly topics.. */
	  TopicTimer_4 = Now;
	  SetTopics(4);		/* 1 hour */
	}
      }
  }
}

/* EOF */
