/*
 *    X2  IRC Operator and Channel services
 *    by Rubin (rubin@afternet.org)
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 1, or (at your option)
 * any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */
#include <time.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <ctype.h>
#include "config.h"
#include "misc.h"
#include "user.h"
#include "sock.h"
#include "chan.h"
#include "send.h"
#include "match.h"
#include "server.h"
#include "time.h"

char *parmstr;
char DBTARGET[MAXLEN];

/*
   #define DBGNOTICE   10
   #define DBGWARNING  20
   #define DBGERROR    30
   #define DBGSEVERE   40
 */

const char convert2y[] =
{
  'A','B','C','D','E','F','G','H','I','J','K','L','M','N','O','P',
  'Q','R','S','T','U','V','W','X','Y','Z','a','b','c','d','e','f',
  'g','h','i','j','k','l','m','n','o','p','q','r','s','t','u','v',
  'w','x','y','z','0','1','2','3','4','5','6','7','8','9','[',']'
};

const unsigned int convert2n[] = {
   0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
   0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
   0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
  52,53,54,55,56,57,58,59,60,61, 0, 0, 0, 0, 0, 0,
   0, 0, 1, 2, 3, 4, 5, 6, 7, 8, 9,10,11,12,13,14,
  15,16,17,18,19,20,21,22,23,24,25,62, 0,63, 0, 0,
   0,26,27,28,29,30,31,32,33,34,35,36,37,38,39,40,
  41,42,43,44,45,46,47,48,49,50,51, 0, 0, 0, 0, 0,

   0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
   0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
   0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
   0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
   0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
   0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
   0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
   0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0
};

char *unsigned32ToDotQuad(unsigned long int ip)
{
 struct in_addr in;

 in.s_addr = ntohl(ip);
 return(inet_ntoa(in));
}

/* Contributed by the great sneep of afternet ;) */
void genpass(char *str, int len)
{
        int i = 0;
        char c = 0;

        for(i = 0; i < len; i++)
        {
                do
                {
                        c = (char)((float)rand() / (float)RAND_MAX * (float)256);
                } while(!((c >= 'a' && c <= 'z') || (c >= 'A' && c <= 'Z') || (c >= '0' && c <= '9')));
                str[i] = c;
        }
        str[i] = '\0';
        return;
}

unsigned long Hash(char *str)
{
	/* strtol is case insensitive, so don't worry about it */
	static char str2[STRTOLSIZE+1] = { 0 };
	int i = 0, j = 0;
	
	/* strncpy(str2, str, STRTOLSIZE); */
	
	if (str == NULL)
		return 0;
	
	/* but it will stop at the first non-alnum */	
	while((i < STRTOLSIZE) && str[j]) {
		if (isalnum(str[i])) {
			str2[i] = str[j];
			i++;
		}
		j++;
	}
	str2[i] = 0;
	
	/* some number from 0 to 36^6-1 */
	return strtol(str2, (char **)NULL, STRTOLBASE);
}

#define NUMNICKLOG 6;
unsigned int base64toint(const char *s)
{
  unsigned int i = convert2n[(unsigned char)*s++];
  while (*s)
  {
    i <<= NUMNICKLOG;
    i += convert2n[(unsigned char)*s++];
  }
  return i;
}

/* taken from ircu2.10.05 */
const char* inttobase64(char* buf, unsigned int v, unsigned int count)
{
  buf[count] = '\0';
  while (count > 0) {
    buf[--count] = convert2y[(v & NUMNICKMASK)];
    v >>= NUMNICKLOG;
  }
  return buf;
}


char *NickToNum(char *nick)
{
    UserListType uptr;
    if((uptr = FindNick(nick, UsersN)))
    {
       return(uptr->Numnick);
    }
    Debug(DBGWARNING, "NickToNum couldnt find nick %s", nick);
    return("NICKNOTFOUND");
}

char *NumToNick(char *num)
{
   UserListType uptr;
   if((uptr = FindNumnick(num, UsersU)))
   {
       return(uptr->Nick);
   }
   Debug(DBGWARNING, "NumToNick couldnt find num %s", num);
   return("NUMNICKNOTFOUND");
}

char *time2str(time_t thetime)
{
    char *buf, *tmp;

    buf = ctime(&thetime);
    tmp = (char *)strchr(buf, '\n');
    *tmp = '\0';
    return(buf);
}

/* mysep() is my answer to the strtok/strsep
 * issue. strsep is nice but doesn't skip
 * multiple dilimiters, which can really
 * offset tokens and cause huge corruption
 * so this function will use strsep but
 * act like strtok in that sence.
 */
char *mysep(char **sepstr, char *delim)
{
  static char *retstr;

  if(!*sepstr || !**sepstr)
    return(NULL);

  do
  {
      retstr = strsep(sepstr, delim);
  }while (retstr && !(*retstr));

  return(retstr);
}

void safecopy(char *dest, char *source, int size)
{
   int i;

   if(!source)
   {
     strcpy(dest, "(SAFENULL)");
     return;
   }
   for(i=0;i < size-1;i++)
   {
     if(source[i] == '\0')
     {
       dest[i] = '\0';
       break;
     }
     dest[i] = source[i];
   }
   dest[i] = '\0';
}

int SendMOTD(UserListType uptr)
{
  FILE *TheFile;
  char LineBuf[MAXLEN + 1];

  if(!uptr)
  {
    Debug(DBGWARNING, "SendMOTD called with NULL uptr");
    return(FALSE);
  }
  if((TheFile = fopen("x2.motd", "r")))
  {
    send_to_user(uptr, "=-------Message Of The Day-------=");
    while (fgets(LineBuf, MAXLEN, TheFile))
      send_to_user(uptr, "%s", LineBuf);
    send_to_user(uptr, "=--------- End of MOTD ----------=");
    fclose(TheFile);
    return(TRUE);
  }
  else
  {
    /* No file - thats ok, just means dont use this feature */
    return(FALSE);
  }
}

void WriteGlineDat()
{
  GlineListType *gptr;
  FILE *TheFile;

  TheFile = fopen(GLINEFILENAME, "w");

  if (!TheFile)
  {
    Debug(DBGERROR, "Cant open gline.dat file");
    fclose(TheFile);
    return;
  }
  if (!(gptr = Glines))
  {
    fclose(TheFile);
    return;	/* no glines */
  }
  while (gptr)
  {
    fprintf(TheFile, "%s %lu %s\n", gptr->address, gptr->time, gptr->reason);
    gptr = gptr->Next;
  }
  fclose(TheFile);
}

void WriteSGlineDat()
{
  GlineListType *gptr;
  FILE *TheFile;

  TheFile = fopen(SUPERGLINEFILENAME, "w");

  if (!TheFile)
  {
    Debug(DBGERROR, "Cant open superline.dat file");
    fclose(TheFile);
    return;
  }
  if (!(gptr = SuperGlines))
  {
    fclose(TheFile);
    return;			/* no glines */
  }
  while (gptr)
  {
    fprintf(TheFile, "%s %lu %s\n", gptr->address, gptr->time, gptr->reason);
    gptr = gptr->Next;
  }
  fclose(TheFile);
}

char *time2apache(char *timestr)
{
   char b[MAXLEN];
   char *buff = b;
   char* m,*d,*t,*y,*w;
   static char str[MAXLEN];


   strcpy(buff, timestr);

   w = (char *)mysep(&buff, " ");
   m = (char *)mysep(&buff, " ");
   d = (char *)mysep(&buff, " ");
   t = (char *)mysep(&buff, " ");
   y = (char *)mysep(&buff, " ");

   if(!(m&&d&&t&&y))
   {
      return "TIME ERROR";
   }
   sprintf(str, "%s/%s/%s:%s -0700",
          d,m,y,t);
   return str;
}

/* Makes an apache style log of connections and channel joins
 * for things like awstats to analyze
 */
void LogAccess(UserListType uptr, char* item, int size)
{
   FILE* LOGFILE;

   if(strcmp(AccessLog, "none"))
   {
      if((LOGFILE = fopen(AccessLog, "a+")) != NULL)
      {
          fprintf(LOGFILE, "%s - - [%s] \"GET /%s HTTP/1.1\" 200 %d \"%s\" \"IRC\"\n",
               uptr->Address, time2apache(time2str(time(NULL))), item, size, uptr->RealName );
          fclose(LOGFILE);
      }
   }
}

void LogCommand(ChannelListType cptr, UserListType uptr, HandleListType hptr, 
                char* command, char* rest, int ChanSecurity, int HelpOverride)
{
   FILE* LOGFILE;
   char buff[MAXLEN*4];
/* global nick!user@host command rest [time] */
/* #channel nick!user@host GOD command rest [time] */
   if( ! (strcasecmp(command, "LOGIN") && 
      strcasecmp(command, "RESETPASS") &&
      strcasecmp(command, "MLOGIN") ) )
   {
       rest = strtok(rest, " "); 
   }
   if(!(strcasecmp(command, "PASS")))
   {
       rest = "*********";
   }
   if(ChanSecurity)
   {
       sprintf(buff, "[%s] %s %s(%s!%s@%s) %s %s %s",  time2str(time(NULL)),
                      cptr->Name, hptr?hptr->Handle:"*", uptr->Nick, ActiveUsername(uptr), ActiveHost(uptr),
                      HelpOverride?"+h":"-h", command, rest);
   }
   else
   {
       sprintf(buff, "[%s] global %s!%s@%s - %s %s",  time2str(time(NULL)),
                      uptr->Nick, ActiveUsername(uptr), ActiveHost(uptr),
                      command, rest);
   }
   if((LOGFILE = fopen(AccountingLog, "a+")) != NULL)
   {
       fprintf(LOGFILE, "%s\n", buff);
       fclose(LOGFILE);
   }
   else
      Debug(DBGWARNING, "%s: Failed to open!", AccountingLog);
}

void stripctrlchars(char* buff, char replacement)
{
   for(;*buff;buff++)
   {
       if(iscntrl(*buff))
          *buff = replacement;
   }
}

void Debug(int Level, char *fmt,...)
{
  va_list args;
  char buff[MAXLEN*4];
  char *tmp;
  FILE *LogFile;
#ifdef LOGALLDEBUG
  static FILE *DbgFile = NULL;
#endif
  static int severecount = 0;

  va_start(args, fmt);
  vsnprintf(buff, MAXLEN, fmt, args);
  buff[MAXLEN - 1] = '\0';
  va_end(args);

  for (tmp = buff; *tmp; tmp++)
  {
    if ((*tmp == '\n') || (*tmp == '\r'))
      *tmp = '\0';
    else if (*tmp == '\001')
      *tmp = ' ';
  }
  stripctrlchars(buff, '?');

#ifdef debugon
  if (Level >= DBGNOTICE)
    printf("DEBUG(%d): %s\n", Level, buff);
#endif

  if ((Level >= DEBUG) && *DBTARGET)
    if (SockConnected)
      sendtosockDEBUG(Sock, ":%s PRIVMSG %s :DEBUG(%d): %s", IrcNick, DBTARGET, Level, buff);

  if(Level >= DBGSEVERE)
  {
      if(severecount++ > MAXDBGSEVERE)
      {
          system("echo Terminating due to exceeded severe errors.>>x2.errors");
          exit(2);
      }
  }

  if (Level >= DBGWARNING)
  {
    sendtosockDEBUG(Sock, ":%s WALLOPS :DEBUG(%d): %s", IrcNick, Level, buff);
    if((LogFile = fopen("x2.errors", "a")))
    {
      fprintf(LogFile, "%s:DEBUG(%d): %s\n", time2str(time(NULL)), Level, buff);
      fclose(LogFile);
    }
  }

#ifdef LOGALLDEBUG
    if(!DbgFile)
    {
       DbgFile = fopen("x2.debug", "w");
    }
    if(DbgFile)
    {
      fprintf(DbgFile, "%lu:DEBUG(%d): %s\n", time(NULL), Level, buff);
    }
#endif
}

time_t NetTime()
{
  return (time(NULL) + NetOffset);
}

char *t2s(char *s, time_t stamp)
{
  strcpy(s, ctime(&stamp));
  s[strlen(s) - 1] = '\0';
  return (s);
}

void SendNetMessage(UserListType uptr)
{
  FILE *TheFile;
  char LineBuf[MAXLEN];

  if((TheFile = fopen(MESSAGEFILENAME, "r")))
  {
    send_to_user(uptr, "\002News from %s CService:\002", NETWORK);
    send_to_user(uptr, "=--------------------------------=");
    while (fgets(LineBuf, MAXLEN, TheFile))
      send_to_user(uptr, "%s", LineBuf);
    send_to_user(uptr, "=-------------- End -------------=");
    fclose(TheFile);
  }
}

int isinchan(ChannelListType cptr, UserListType uptr)
{
  /* look though cptr->Joined for uptr.. */
  ChannelUsersListType ChanUser;

  for(ChanUser = cptr->Joined; ChanUser; ChanUser = ChanUser->Next)
      if (uptr == ChanUser->User)
          return TRUE;
  return(FALSE);
}

int opok(ChannelListType ChanPtr, UserListType uptr, UserListType optr)
{
  /* Check if its ok for user from to op user to given settings */
  if(uptr == NULL)
     return(TRUE); /* No user.. must be a server */
  switch (ChanPtr->settings[STRICTOP])
  {
  case 0:			/* anyone */
    return (TRUE);
  case 1:			/* Ops Managers & (CO)Owners only */
    if (userlevel2(optr, FindBestUserMatch(ChanPtr, optr)) >= OP)
      return (TRUE);
    if (userlevel2(uptr, FindBestUserMatch(ChanPtr, uptr)) >= OP)
      return (TRUE);
    return (FALSE);
  case 2:			/* Managers & (co)owners */
    if (userlevel2(optr, FindBestUserMatch(ChanPtr, optr)) >= OP)
      return (TRUE);
    if (userlevel2(uptr, FindBestUserMatch(ChanPtr, uptr)) >= MA)
      return (TRUE);
    return (FALSE);
  default:
    return (TRUE);
  }
}

int halfopok(ChannelListType ChanPtr, UserListType uptr, UserListType optr)
{
  /* Check if its ok for user from to op user to given settings */
  if(uptr == NULL)
     return(TRUE); /* No user.. must be a server */
  switch (ChanPtr->settings[STRICTOP])
  {
  case 0:			/* anyone */
    return (TRUE);
  case 1:			/* Ops Managers & (CO)Owners only */
    if (userlevel2(optr, FindBestUserMatch(ChanPtr, optr)) >= OP)
      return (TRUE);
    if (userlevel2(uptr, FindBestUserMatch(ChanPtr, uptr)) >= OP)
      return (TRUE);
    return (FALSE);
  case 2:			/* Managers & (co)owners */
    if (userlevel2(optr, FindBestUserMatch(ChanPtr, optr)) >= OP)
      return (TRUE);
    if (userlevel2(uptr, FindBestUserMatch(ChanPtr, uptr)) >= MA)
      return (TRUE);
    return (FALSE);
  default:
    return (TRUE);
  }
}

int deopok(ChannelListType ChanPtr, UserListType from, UserListType to)
{
  int fromlev, tolev;

  if(from == NULL)
      return(TRUE); /* no from, must be server */
  if(from ==  to)  /* Mode done to self. */
  {
      return(TRUE);
  }
  fromlev = userlevel2(from, FindBestUserMatch(ChanPtr, from));
  tolev = userlevel2(to, FindBestUserMatch(ChanPtr, to));
  /* Check if its ok for user from to deop user to given settings */
  switch (ChanPtr->settings[PROTECT])
  {
  case 0:			/* Noone from anyone */
    return (TRUE);
  case 1:			/* Higher rank from lower rank */
    if (tolev < OP)
      return (TRUE);
    if (fromlev < tolev)
      return (FALSE);
    return (TRUE);
  case 2:			/* Equal or higher rank */
    if (tolev < OP)
      return (TRUE);
    if (fromlev <= tolev)
      return (FALSE);
    return (TRUE);
  case 3:			/* Managers & owners from everyone */
    if (tolev < MA)
      return (TRUE);
    if (fromlev <= tolev)
      return (FALSE);
    return (TRUE);
  case 4:			/* Owners from everyone */
    if (tolev < CO)
      return (TRUE);
    if (fromlev < tolev)
      return (FALSE);
    return (TRUE);
  default:
    return (TRUE);
  }
}

int dehalfopok(ChannelListType ChanPtr, UserListType from, UserListType to)
{
  int fromlev, tolev;

  if(from == NULL)
      return(TRUE); /* no from, must be server */
  if(from ==  to)  /* Mode done to self. */
  {
      return(TRUE);
  }
  fromlev = userlevel2(from, FindBestUserMatch(ChanPtr, from));
  tolev = userlevel2(to, FindBestUserMatch(ChanPtr, to));
  /* Check if its ok for user from to deop user to given settings */
  switch (ChanPtr->settings[PROTECT])
  {
  case 0:			/* Noone from anyone */
    return (TRUE);
  case 1:			/* Higher rank from lower rank */
    if (tolev < OP)
      return (TRUE);
    if (fromlev < tolev)
      return (FALSE);
    return (TRUE);
  case 2:			/* Equal or higher rank */
    if (tolev < OP)
      return (TRUE);
    if (fromlev <= tolev)
      return (FALSE);
    return (TRUE);
  case 3:			/* Managers & owners from everyone */
    if (tolev < MA)
      return (TRUE);
    if (fromlev <= tolev)
      return (FALSE);
    return (TRUE);
  case 4:			/* Owners from everyone */
    if (tolev < CO)
      return (TRUE);
    if (fromlev < tolev)
      return (FALSE);
    return (TRUE);
  default:
    return (TRUE);
  }
}

int kickok(ChannelListType ChanPtr, UserListType from, UserListType to)
{
  int fromlev, tolev;

  if(from == NULL)
      return(TRUE); /* no from, must be server */
  fromlev = userlevel2(from, FindBestUserMatch(ChanPtr, from));
  tolev = userlevel2(to, FindBestUserMatch(ChanPtr, to));
  /* Check if its ok for user from to kick user to given settings */
  switch (ChanPtr->settings[PROTECT])
  {
  case 0:			/* Noone from anyone */
    return (TRUE);
  case 1:			/* Higher rank from lower rank */
    if (tolev < OP)
      return (TRUE);
    if (fromlev < tolev)
      return (FALSE);
    return (TRUE);
  case 2:			/* Equal or higher rank */
    if (tolev < OP)
      return (TRUE);
    if (fromlev <= tolev)
      return (FALSE);
    return (TRUE);
  case 3:			/* Managers & owners from everyone */
    if (tolev < MA)
      return (TRUE);
    if (fromlev <= tolev)
      return (FALSE);
    return (TRUE);
  case 4:			/* (CO)Owners from everyone */
    if (tolev < CO)
      return (TRUE);
    if (fromlev < tolev)
      return (FALSE);
    return (TRUE);
  default:
    return (TRUE);
  }
}

int modeok(ChannelListType ChanPtr, UserListType from)
{
  if(from == NULL)
     return(TRUE); /* must be server */
  /* Check if its ok for user from to change modes from def */
  switch (ChanPtr->settings[ENFMODES])
  {
  case 0:
    return (TRUE);		/* everyone */
  case 1:			/* Only OPs, MANAGERs, and (CO)OWNERs */
    if (userlevel2(from, FindBestUserMatch(ChanPtr, from)) >= OP)
      return (TRUE);
    return (FALSE);
  case 2:			/* Only H and W */
    if (userlevel2(from, FindBestUserMatch(ChanPtr, from)) >= MA)
      return (TRUE);
    return (FALSE);
  default:
    return (TRUE);
  }
}

int banok(ChannelListType ChanPtr, UserListType from, char* mask)
{
  int tolev, fromlev;

  if(from == NULL)
     return(TRUE);
  if (!matches(mask, "blasdewrwaasdvcyuzysafew!asdfasdzxcvasretyyas@dsasdffw.asdfasdf.asdfasdfwettyyy.casdfasgdg"))
    if (ChanPtr->settings[PROTECT] > 0)
      return (FALSE);		/* if it matches this it must be a global (*!*@*) ban.. */
  tolev = IsProt(ChanPtr, mask);
  fromlev = userlevel2(from, FindBestUserMatch(ChanPtr, from));
  /* Check if the mask is protected */
  switch (ChanPtr->settings[PROTECT])
  {
  case 0:			/* Noone from anyone */
    return (TRUE);
  case 1:			/* Higher rank from lower rank */
    if (tolev < OP)
      return (TRUE);
    if (fromlev < tolev)
      return (FALSE);
    return (TRUE);
  case 2:			/* Equal or higher rank */
    if (tolev < OP)
      return (TRUE);
    if (fromlev <= tolev)
      return (FALSE);
    return (TRUE);
  case 3:			/* Managers & owners from everyone */
    if (tolev < MA)
      return (TRUE);
    if (fromlev <= tolev)
      return (FALSE);
    return (TRUE);
  case 4:			/* Owners from everyone */
    if (tolev < CO)
      return (TRUE);
    if (fromlev < tolev)
      return (FALSE);
    return (TRUE);
  default:
    return (TRUE);
  }
  return (TRUE);
}

void punish(ChannelListType ChanPtr, UserListType from, UserListType to, char type)
{
  int fromlev, tolev;

  if(from == NULL)
     return;
  /* If to is protected from from, apply protlev punishment on from. */
  fromlev = userlevel2(from, FindBestUserMatch(ChanPtr, from));
  tolev = userlevel2(to, FindBestUserMatch(ChanPtr, to));

  if (IsModeK(from))
    return;
  if (fromlev >= OP)
  {
    /* From is a user too, so dont hurt them too bad.. */
    send_to_user(from, "\002%s is protected.\002", to->Nick);
    return;
  }
  send_kick(ChanPtr, from, "\002%s is protected, get out!\002", to->Nick);
  send_to_user(from, "\002%s is Protected!\002", to->Nick);
  return;
}
/* A hack to make a string escaped for use with system().
 * Not really used anymore.
 */
void SysPrep(char *str)
{
  char buf[255];
  int b = 0;
  int s = 0;

  safecopy(buf, str, 255);

  while (buf[b] != '\0')
  {
    if (ispunct(buf[b]))
    {
      str[s] = '\\';
      s++;
    }
    str[s] = buf[b];
    b++;
    s++;
  }
  str[s] = '\0';
}

int isiphost(char *hostname)
{
  while (*hostname)		/* assuming null terminated */
  {
    if ((!isdigit(*hostname)) && (hostname[0] != '.'))
      return 0;			/* false */
    hostname++;
  }
  return 1;			/* true */
}

int IsAllDigits(char *string)
{
  while (*string)		/* assuming null terminated */
  {
    if (!isdigit(*string))
      return 0;			/* false */
    string++;
  }
  return 1;			/* true */
}

#ifdef checkspoof

void CheckSpoof(char *stat)
{
  UserListType UserNode;
  struct hostent *HostPtr;
  struct in_addr addr;
  unsigned long int longip;
  char tmp[MAXLEN];
  char *nick = NULL, *user = NULL, *host = NULL;
  char *ptr;
  int done = 0;

  /* This mess here is to parse nick!user@host from the userip
   * reply which is :Rubin*=+rubin@199.2.210.241 this got a
   * lot simpler after userip was put into ircd.
   * We now have the ip stored so this needs to be redone..
   * course spoofing is not really possible anymore.
   */
  stat++;			/* get rid of leading : */
  nick = (char *) strtok(stat, "=");
  if (nick[strlen(nick) - 1] == '*')
    nick[strlen(nick) - 1] = '\0';
  user = (char *) strtok(NULL, "@");
  if (user[0] == '+')
    user++;
  host = (char *) strtok(NULL, "\n \r\0");
  if (!(nick && user && host))
  {
    Debug(DBGINFO, "nick, user, or host was NULL returning from checkspoof()");
    return;
  }
  if (!strcmp(host, "127.0.0.1"))	/* users on local to servers */
    return;
  if (!strcmp(host, "0.0.0.0"))	/* services */
    return;
  safecopy(tmp, host, MAXLEN);
  if (!(UserNode = FindNick(nick, UsersN)))
  {
    Debug(DBGINFO, "User not found %s", nick);
    return;
  }
  if (isiphost(UserNode->Address))
  {
    if (strcmp(tmp, UserNode->Address))
    {
      send_wallops("\002Posible spoof! [\002%s!%s@%s\002] - IP mismatch - Real IP=[\002%s\002]\002",
	 UserNode->Nick, UserNode->UserID, UserNode->Address, tmp);
      return;
    }
    return;
  }
  longip = inet_network(host);
  if (longip == INADDR_NONE)
    return;
  addr.s_addr = htonl(longip);
  HostPtr = gethostbyaddr((char *) &addr, sizeof(struct in_addr), AF_INET);

  if (!HostPtr)
    return;
  if (HostPtr->h_name)
  {
    if (strcasecmp(HostPtr->h_name, UserNode->Address))		/* If they are different */
    {
      send_wallops("\002Posible spoof! [\002%s!%s@%s\002] - Host Mismatch - Real host=[\002%s\002]\002",
		 UserNode->Nick, UserNode->UserID, UserNode->Address, HostPtr->h_name);
    }
    return;
  }
}
#endif /* checkspoof */

void ban_user(ChannelListType cptr, UserListType uptr)
{
  char buff[MAXLEN];

  sprintf(buff, "%s!%s@%s", uptr->Nick, ActiveUsername(uptr), ActiveHost(uptr));
  send_chanmode(cptr, "-ov+b %s %s %s", uptr->Numnick, uptr->Numnick, cluster(buff));
  add_ban(cptr, cluster(buff));
}

void deopban(ChannelListType cptr, UserListType uptr, char *mask)
{
  send_chanmode(cptr, "-o+b %s %s", uptr->Numnick, mask);
  add_ban(cptr, mask);
}

void ban_mask(ChannelListType cptr, char *mask)
{
  send_chanmode(cptr, "+b %s", mask);
  add_ban(cptr, mask);
}

/* So far as i can tell this does nothing.
 * Whats it used for?
 */
char *strip(char *temp, char c)
{
  if (temp)
    for (; (temp != NULL) && (*temp == c); temp++);
  return temp;
}

/* returns the right (num) charactors of the string */
char *right(char *string, int num)
{
  if (strlen(string) < num)
    return string;
  return (string + strlen(string) - num);
}

/* returns the number of c in string */
int numchar(char *string, char c)
{
  int num = 0;

  while (*string)
  {
    if (tolower(*string) == tolower(c))
      num++;
    string++;
  }
  return num;
}

/* Returns a good guess at a hostmask for a given userhost string.
 * eg nick!user@host.domain.com becomes *!*user@*.domain.com
 * Attempting to add support for .nl etc type countrys.. but not finished.
 */
char *cluster(char *userhost)
{
  char *nick, *user;
  char nuh[255];
  char fullhost[255];
  static char buffer[255];
  char *temp, *tptr;
  int numofdots, x, isnd = 0;
  int didtruncate = 0;

  numofdots = 1;
  safecopy(nuh, userhost, 255);
  nick = (char *) strtok(nuh, "!");
  if (!nick || !*nick)
    nick = "*";
  user = (char *) strtok(NULL, "@");
  if (user)
    user = (char *) strip(strip(user, '~'), '#');
  if (!user || !*user)
    user = "*";
  temp = (char *) strtok(NULL, "\n\0");
  if (temp && *temp)
    safecopy(fullhost, temp, 255);
  else
    strcpy(fullhost, "*");
  temp = (char *) right(fullhost, 2);
/* This will have to be reworked when the new domains are deploid */
  tptr = &temp[strlen(temp) - 4];
/* Try to guess how many dots we need (minus one) */
 if (      !strcasecmp(tptr, ".com") ||
           !strcasecmp(tptr, ".edu") ||
           !strcasecmp(tptr, ".org") ||
           !strcasecmp(tptr, ".net") ||
           !strcasecmp(tptr, ".gov") ||
           !strcasecmp(tptr, ".mil") ||
           !strcasecmp(tptr, ".orb") ||
           !strcasecmp(tptr, ".my")  ||
           !strcasecmp(tptr, ".kz")  ||
           !strcasecmp(tptr, ".es")  ||
           !strcasecmp(tptr, ".no")  ||
           !strcasecmp(tptr, ".ca")  ||
           !strcasecmp(tptr, ".nu")  ||
           !strcasecmp(tptr, ".fr")
    )
  {
    numofdots = 1;
  }
  else
  {
    numofdots = 2;
  }
  if(!matches("*.AfterNET.Org", fullhost))
      numofdots = 4;
  temp = fullhost;
  for (x = 0; x <= (strlen(fullhost) - 1); x++)
  {
    if (!(fullhost[x] == '.') && !isdigit(fullhost[x]))
      isnd++;
  }

  if (isnd == 0)		/* If its digits */
  {
    while (temp && *temp && (numchar(temp, '.') > 1))
    {
      temp = (char *) strchr(temp, '.');
      if (temp)
	temp++;
      else
	return "";
    }
    if (temp)
      *temp = '\0';
    strcat(fullhost, "*");
    temp = fullhost;
  }
  else				/* If its a hostname */
    while (temp && *temp && (numchar(temp, '.') > numofdots))
    {
      didtruncate++;
      temp = (char *) strchr(temp, '.');
      if (temp)
	temp++;
      else
	return "";
    }
  /* We must omit the * after ! if the username is 10, because in
   * "standard" ircd's you canot ban 11 char usernames and it counts
   * the "*" as a char. so *!*1234567890@ will result in the 0 being
   * dropped and the ban being uneffective. -Rubin */
  if (strlen(user) > 9)
    strcpy(buffer, "*!");
  else
    strcpy(buffer, "*!*");
  strcat(buffer, user);
  if (isnd == 0)
    strcat(buffer, "@");
  else
  {
    if (didtruncate)
      strcat(buffer, "@*.");
    else
      strcat(buffer, "@");
  }
  if(isnd != 0)
    for(tptr=temp;*tptr;tptr++)
    {
       if(isdigit(*tptr))
         *tptr= '*';
    }
  strcat(buffer, temp);
  return buffer;
}

time_t getthetime(void)
{
  time_t t;

  t = time((time_t *) NULL);
  return t;
}

/* Return true if channel begins with # or &
 * Note: x2 would never see a & channel... */
int ischannel(char *channel)
{
  return (*channel == '#' || *channel == '&');
}

int userlevel2(UserListType uptr, HandleListType hptr)
{
  int level = 0;

  if(!uptr || !hptr)
  {
     return(0);
  }
  if (!strcasecmp(IrcNick, uptr->Nick))
  {
    return (700); /* Its me. */
  }
  if (IsModeK(uptr))
    level = 600;
  else if (IsChanOwner(hptr))
    level = OW;
  else if (IsChanCoOwner(hptr))
    level = CO;  
  else if (IsChanManager(hptr))
    level = MA;
  else if (IsChanOp(hptr))
    level = OP;
  else if (IsChanPeon(hptr))
    level = PE;    
  return (level);
}

char levelcharfromint(int theint)
{
   if(theint >= OW)
      return('W');
   else if(theint >= CO)
      return('C');
   else if(theint >= MA)
      return('M');
   else if(theint >= OP)
      return('O');
   else if(theint >= PE)
      return('P');
   else
   {
      Debug(DBGSEVERE, "levelcharfromint() called with argument less then 10 (PE)!.");
      return('E'); 
   }
}

char levelcharfromname(char *level)
{
   if(!strcasecmp(level, "OWNER"))
        return('W');
   if(!strcasecmp(level, "COOWNER") || !strcasecmp(level, "CO"))
        return('C');
   if(!strcasecmp(level, "MANAGER"))
        return('M');
   if(!strcasecmp(level, "OP"))
        return('O');
   if(!strcasecmp(level, "PEON"))
        return('P');
   if(!strcasecmp(level, "LAMER"))
        return('L');
   return('\0');
}

char *levelnamefromchar(char thechar)
{
   switch(thechar)
   {
     case 'L': return("LAMER");
     case 'P': return("PEON");
     case 'O': return("OP");
     case 'M': return("MANAGER");
     case 'C': return("COOWNER");
     case 'W': return("OWNER");
     default: return("UNKNOWN/INVALID");
   }
}

char *levelnamefromint(int theint)
{
   if(theint >= OW)
      return("OWNER");
   else if(theint >= CO)
      return("COOWNER");
   else if(theint >= MA)
      return("MANAGER");
   else if(theint >= OP)
      return("OP");
   else if(theint >= PE)
      return("PEON");
   else return("NONE");
}

int userlevelbychar(char thechar)
{
   switch(thechar)
   {
      case 'L': return(0);
      case 'P': return(PE);
      case 'O': return(OP);
      case 'M': return(MA);
      case 'C': return(CO);
      case 'W': return(OW);
      default: 
         Debug(DBGWARNING, "userlevelbychar() error: Not LPOHC or W!");
         return(0);
   }
}

/* Greatly simplified this by taking a hptr instead of a user and a channel.
 * now it is pretty simple, but the calling params changed so everywhere
 * needs updating.  */
int userlevel(HandleListType hptr)
{
  if(hptr == NULL)
     return(0); 
  return(userlevelbychar(hptr->Level));
}

/* I'd prefer I think that functions use IsOper IsGodOper etc rather then this,
 * so we have more readable code.  When everyones stopped using this, remove it. */
int userleveloper(UserListType uptr, HandleListType hptr)
{
  int level = 0;

  if(hptr == NULL)
    return (0);
  else if (IsChanOwner(hptr))
    level = OW;
  else if (IsChanCoOwner(hptr))
    level = CO;
  else if (IsChanManager(hptr))
    level = MA;
  else if (IsChanOp(hptr))
    level = OP;
  else if (IsChanPeon(hptr))
    level = PE;  
  if(level == 0)
  {
     if(IsGodOper(uptr))
     {
        level = 600;
        send_to_user(uptr, "\002WARNING:\002 You are using GOD mode to override channel security!");
     }
  }
  return (level);
}

char *GetSetting(boolean Setting)
{
  switch (Setting)
  {
  case 0:
    return ("DISABLED");
  case 1:
    return ("ENABLED");
  }
  return ("ERROR");
}

char *GetCloneResMet(int Setting)
{

  switch (Setting)
  {
  case 0:
    return ("ALLOW");
  case 1:
    return ("WARN");
  case 2:
    return ("KILL");
  case 3:
    return ("GLINE");
  default: return("ERROR");
  }
}

/* ALSO USED FOR SUB7 */
char *GetWinGateMet(int Setting)
{
   switch(Setting)
   {
     case 0:
       return("DISABLED");
     case 1:
       return("WALLOP");
     case 2:
       return("GLINE");
     default: return("ERROR");
   }
}

char *MakeUpper(char *Str)
{
   int i,Len;
   Len = strlen(Str);
   for(i=0;i<Len;i++)
   {
      Str[i] = toupper(Str[i]); 
   }
   return(Str);
}

/* Trying to phase this out.. its annoying
 * to use.  Use The above on a new string instead.
 */
char *UpCase(char *Str)
{
  char *UpCaseStr;
  int Len, i;

  Len = strlen(Str);
  if((UpCaseStr = (char *) malloc(Len + 1)))
  {
  for (i = 0; i < Len; i++)
    UpCaseStr[i] = toupper(Str[i]);
  UpCaseStr[i] = '\0';
  return (UpCaseStr);
  }
  else
  {
     Debug(DBGWARNING, "Malloc failed in UpCase()");
     exit(2);
  }
}

boolean NickOk(char *Nick)
{

/* not finished... */
  if (strlen(Nick) >= MAXNICKLEN)
    return (false);

  return (true);
}
/* This is the function regchan uses to check 
 * if the name is ok.  
 */
boolean ChanOk(char *Chan)
{
/* not finished ... (i think) */

  int len, i;

  if (!Chan)
    return (0);
  len = strlen(Chan) - 1;
  if (len > LINEBUFFSIZE)
    return (false);
  if (Chan[0] != '#')
    return (false);
  if(strchr(Chan, '/')) 
    return(false);
  if ((!isalnum(Chan[1])) && (Chan[1] != '~') && (Chan[1] != '_') && (Chan[1] != '@') && (Chan[1] != '#') && (Chan[1] != '%') && (Chan[1] != '^'))
    return (false);
  for (i = 2; i < len; i++)
    if ((!isalnum(Chan[i])) && (!ispunct(Chan[i])) && (Chan[i] != '-') && (Chan[i] != '_') && (Chan[i] != '@') && (Chan[i] != '#') && (Chan[i] != '%') && (Chan[i] != '^') && (Chan[i] != '.'))
      return (false);
  return (true);
}

boolean ModeOk(char *Mode)
{

/* not done.. */
  if ((Mode[0] != '+') && (Mode[0] != '-'))
    return (0);
  return (1);
}

int HandleOk(char *Handle)
{
  char *tmp;

  if (!Handle)
    return (TRUE);
  for (tmp = Handle; *tmp; tmp++)
  {
    if (iscntrl(*tmp))
      return (FALSE);
  }
  return (TRUE);
}

int PassOk(char *Pass)
{
   char *tmp;

   if(!Pass)
      return(FALSE);
   for(tmp = Pass; *tmp; tmp++)
   {
     if(iscntrl(*tmp))
        return(FALSE);
   }
   if(strlen(Pass) > PASSWORDLEN)
       return(FALSE);
   return(TRUE);
}

int KeyOk(char *Key)
{
  if (!Key)
    return (TRUE);
  if(strlen(Key) > 56)
    return(FALSE);
  {
    u_char *s1, *s2;

    for (s1 = s2 = (u_char *)Key; *s2; s2++)
      if ((*s1 = *s2 & 0x7f) > (u_char) 32 && *s1 != ':')
	s1++;
    *s1 = '\0';
  }
  if (!*Key)
    return (FALSE);
  return (TRUE);
}

int AddyOk(char *Address)
{
  int i, len;
  boolean gotexcl = false;
  boolean gotat = false;

  len = strlen(Address);

  for (i = 0; i < len; i++)
  {
    if (isalnum(Address[i]))
      continue;
    switch (Address[i]) {
        case '*': case '?': case '+': case '-': case '.': case '\\':
        case '^': case '_': case '{': case '[': case '}': case ']': case '~':
        case '\'': case '`': case '|':
            continue;
        case '!': {
            if (gotexcl || gotat)
	        return (false);
            else
          	gotexcl = true;
            break;
        }
        case '@': {
            if (gotat)
	        return (false);
            else
   	        gotat = true;
            break;
        }
        default:
            return (false);
    }
  }
  if (!gotat)
    return (false);
  return (true);
}

int GlineAddyOk(char *Address)
{
  int i = 0, len;
  boolean gotat = false;
  boolean gotstar = false;

  len = strlen(Address);

/*  if ((!isalpha(Address[0])) && (Address[0] != '*') && (Address[0] != '?'))
 *    return (0);
 */
  if (Address[0] == '*')
  {
    gotstar = true;
    i++;
  }
  /** is this supposed to be i=0? was just i and 
   * -Wall cleanup found it. */
  for(i=0; i < len; i++)
  {
    if ((isalnum(Address[i])))
      continue;
    else if (Address[i] == '*')
    {
      if ((!gotstar))
	gotstar = true;
      else
      {
/* Why are we not allowing * before @???
 *	if (!gotat)
 *	  return (false);
 */
      }
    }
    else if (Address[i] == '!')
    {
      return (0);
    }
    else if (Address[i] == '@')
    {
      if (gotat)
	return (0);  /* Second @ */
      else
	gotat = true;
    }
    else if(ispunct(Address[i]))
        continue;
  }
  if (!gotat)
    return (false);
  if (matches(Address, "fercoasde@apdic.yfa.dsfsdaffsdasfdasfd.asdfasdffdsafsdasfda.asdfasfdfsdsfdasfda.ydfbe"))		/* if it doesnt match this, its ok.. if it does its prolly *@*. */
    return (true);
  else
    return (false);
}

int GetModeNum(char *Mode)
{

  int Len, i;
  int ModeNum = 0;


  Len = strlen(Mode);
  for (i = 0; i < Len; i++)
  {
    switch (Mode[i])
    {
    case 'n':
      ModeNum = ModeNum | CMODEn;
      break;
    case 't':
      ModeNum = ModeNum | CMODEt;
      break;
    case 's':
      ModeNum = ModeNum | CMODEs;
      break;
    case 'p':
      ModeNum = ModeNum | CMODEp;
      break;
    case 'i':
      ModeNum = ModeNum | CMODEi;
      break;
    case 'm':
      ModeNum = ModeNum | CMODEm;
      break;
    case 'r':
      ModeNum = ModeNum | CMODEr;
      break;
    case 'u':
      ModeNum = ModeNum | CMODEu;
      break;
    case 'c':
      ModeNum = ModeNum | CMODEc;
      break;
    case 'C':
      ModeNum = ModeNum | CMODEC;
      break;
    default:
      return (-1);
    }
  }
  return (ModeNum);

}

char *reconstruct(int ac, char **av, int start)
{
  static char tmp[MAXLEN];
  int I;

  strcpy(tmp, "");

  for (I = start; I < ac; I++)
  {
    if (I != start)
      strcat(tmp, " ");
    strcat(tmp, av[I]);
  }

  return (tmp);
}


int calcmaskweight(char *mask)
{
   int count;
   char *tmp;

   tmp = mask;
   for(count = 0;*mask != '\0';mask++)
   {
      if(isalnum(*mask))
         count++;
   }
   return(count);
}

char OpCh(int Aop)
{
  if (Aop)
    return ('@');
  else
    return (' ');
}
char *NameCh(char *Name)
{
  if (Name[0] == '%')
    return ("");
  else
    return (Name);
}

char *ActiveHost(UserListType uptr)
{
    static char buff[MAXLEN];
    char *buffptr = buff;
    if(uptr->Modes & UMODEh)
    {
        char *uname;
        char *host;
        strcpy(buff, uptr->Sethost);
        uname = mysep(&buffptr, "@");
        host = mysep(&buffptr, "@");

        return(host);
    }
    else if(uptr->Modes & UMODEf)
        return(uptr->Fakehost);
    else if(uptr->Modes & UMODEr)
    {
        if(uptr->Modes & UMODEo)
            sprintf(buff, "%s.%s", uptr->Account, HIDDEN_HOST_STAFF);
        else
            sprintf(buff, "%s.%s", uptr->Account, HIDDEN_HOST);
        return(buff);
    }
    else
        return(uptr->Address);
}

char *ActiveUsername(UserListType uptr)
{
    static char buff[MAXLEN];
    char *buffptr = buff;
    if(uptr->Modes & UMODEh)
    {
        char *uname;
        strcpy(buff, uptr->Sethost);
        uname = mysep(&buffptr, "@");
        return(uname);
    }
    else
        return(uptr->UserID);
} 



